/*============================================================================*/
/*                        L a b W i n d o w s / C V I                         */
/*----------------------------------------------------------------------------*/
/*    Copyright (c) National Instruments 1987-1996.  All Rights Reserved.     */
/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Title:       daq_num.c                                                     */
/* Purpose:     implements analog input/output numeric controls               */
/*                                                                            */
/*============================================================================*/

#include <ansi_c.h>
#include <utility.h>
#include <userint.h>
#include "easyio.h"
#include "custsupp.h"
#include "toolbox.h"
#include "easytab.h"
#include "daqctrl.h"
#include "daq_num.h"


#define ATTR_DAQ_NUMERIC_CHANNEL_STRING_DEFAULT             ATTR_DAQ_CTRL_CHANNEL_STRING_DEFAULT
#define ATTR_DAQ_NUMERIC_HIGH_LIMIT_DEFAULT                 ATTR_DAQ_CTRL_HIGH_LIMIT_DEFAULT
#define ATTR_DAQ_NUMERIC_LOW_LIMIT_DEFAULT                  ATTR_DAQ_CTRL_LOW_LIMIT_DEFAULT
#define ATTR_DAQ_NUMERIC_SCAN_FREQUENCY_DEFAULT             10.0
#define ATTR_DAQ_NUMERIC_ENABLED_DEFAULT                    ATTR_DAQ_CTRL_ENABLED_DEFAULT
#define ATTR_DAQ_NUMERIC_VISIBLE_DEFAULT                    ATTR_DAQ_CTRL_VISIBLE_DEFAULT
#define ATTR_DAQ_NUMERIC_ALLOW_RCLICK_EDIT_DEFAULT          ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT_DEFAULT
#define ATTR_DAQ_NUMERIC_EDIT_DIALOG_TITLE_DEFAULT          "Edit DAQ Numeric Properties"
#define ATTR_DAQ_NUMERIC_EDIT_DIALOG_DISPLAYED_DEFAULT      ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED_DEFAULT
#define ATTR_DAQ_NUMERIC_EDIT_DIALOG_MODAL_DEFAULT          ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL_DEFAULT
#define ATTR_DAQ_NUMERIC_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT   ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT
#define ATTR_DAQ_NUMERIC_ERROR_NOTIFICATION_DEFAULT         ATTR_DAQ_CTRL_ERROR_NOTIFICATION_DEFAULT

#define DAQ_NUMERIC_SETTINGS_PREFIX     "DAQNUMERIC_"   /* section prefix in configuration file */
#define DAQ_NUMERIC_TYPE_NAME           "DAQ Numeric"
#define DAQ_NUMERIC_BETA_VERSION        1
#define DAQ_NUMERIC_VERSION             2

#define CTRL_VALUE_TAG                  "Control Value"
#define BETA_REFRESH_FREQUENCY_TAG      "Refresh Frequency"
#define SCAN_FREQUENCY_TAG              "Scan Frequency"

#define DAQ_NUMERIC_TYPE                "DAQNUMERIC"

typedef struct DAQ_NumericInfo_Struct
    {
    COMMON_DAQ_CTRL_FIELDS      /* macro from daqctrl.h containing fields common to all daq ctrls */

    /* daq numeric fields */
    int         timerCtrl;
    int         outputInvalidated;
    double      lastValueOutput;
    double      savedVoltageVal;
    double      savedScanFrequency;
    int         valueNumeric;
    int         scanFrequencyNumeric;
    int         actualFrequencyNumeric;
    double      lastFreqDisplayTime;
    int         numTicksSinceLastFreqDisplay;

    } *DAQ_NumericInfo;


static int Internal_DAQ_Numeric_Create(int panel, int style, char *label, int ioDirection, int top, int left, char *configurationName, CtrlCallbackPtr callback, void *callbackData);
static int Internal_DAQ_Numeric_ConvertFromNumeric(int panel, int baseCtrl, char *configurationName);

static int CVIFUNC_C Internal_DAQ_Numeric_SetAttribute(int panel, int daqNumeric, int attribute, ...);
static int CVIFUNC Internal_DAQ_Numeric_GetAttribute(int panel, int daqNumeric, int attribute, void *value);


static int  CVICALLBACK DAQ_NumericTimerCallback(int panel, int timerConrol, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK DAQ_NumericCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CreateDAQ_NumericInfo(int panel, int ctrl, DAQ_NumericInfo *daqNum);
static void DiscardDAQ_NumericInfo(DAQ_NumericInfo daqNum);

static int  DAQ_NumericHasReasonableRangeForSlavingToLimits(int panel, int daqNumeric);

static int CVIFUNC RecallSettings(DAQ_NumericInfo daqNum, char *settingsName, char *errorDescription);
static int CVIFUNC SaveSettings(DAQ_NumericInfo daqNum, char *settingsName);

    /*  edit dialog support functions */
static int  CVIFUNC DAQ_Numeric_BuildPopup(DAQ_NumericInfo daqNum);
static int  CVIFUNC CopyAttributesToDialog(DAQ_NumericInfo daqNum);
static int  CVIFUNC ApplyDialogSettings(DAQ_NumericInfo daqNum);
static int  CVIFUNC RememberAttributes(DAQ_NumericInfo daqNum);
static int  CVIFUNC RevertAttributes(DAQ_NumericInfo daqNum);

static int  DAQ_Numeric_FormatPopup(DAQ_NumericInfo daqNum);
static void UpdateOutput(DAQ_NumericInfo daqNum);
static void UpdateInput(DAQ_NumericInfo daqNum);
static int  CVIFUNC UpdateDialogDimming(DAQ_NumericInfo daqNum);
static void CVICALLBACK DeferredOutputFunction (void *daqNum);
static int  InvalidateOutput(DAQ_NumericInfo daqNum);
static int  CreateDummyEditingDAQNumeric(int *panel, int *dummyNumeric, int modal);

/*****************************************************************************************/
#pragma warning(disable: 4761)
static int CreateDAQ_NumericInfo(int panel, int ctrl, DAQ_NumericInfo *newDaqNum)
{
    int             error = 0;
    DAQ_NumericInfo daqNum = 0;

    nullChk( daqNum = (DAQ_NumericInfo)calloc(sizeof(*daqNum), 1));

    daqNum->setAttribute = Internal_DAQ_Numeric_SetAttribute;
    daqNum->getAttribute = Internal_DAQ_Numeric_GetAttribute;
    daqNum->buildDialog = (BuildDialogFunc)DAQ_Numeric_BuildPopup;
    daqNum->saveSettings = (SaveSettingsFunc)SaveSettings;
    daqNum->recallSettings = (RecallSettingsFunc)RecallSettings;
    daqNum->updateDialogDimming = (UpdateDialogDimmingFunc)UpdateDialogDimming;
    daqNum->applyDialogSettings = (ApplyDialogSettingsFunc)ApplyDialogSettings;
    daqNum->revertAttributes = (RevertAttributesFunc)RevertAttributes;
    daqNum->rememberAttributes = (RememberAttributesFunc)RememberAttributes;
    daqNum->copyAttributesToDialog = (CopyAttributesToDialogFunc)CopyAttributesToDialog;
    daqNum->settingsPrefix = DAQ_NUMERIC_SETTINGS_PREFIX;
    daqNum->ctrlTypeName = DAQ_NUMERIC_TYPE_NAME;
    daqNum->version = DAQ_NUMERIC_VERSION;
    daqNum->baseCtrl = ctrl;
    daqNum->panel = panel;

    errChk( InitDaqCtrl ((DaqCtrl)daqNum));

    errChk( daqNum->timerCtrl = NewCtrl(panel, CTRL_TIMER, 0, -100, -100));
    errChk( InstallCtrlCallback(panel, daqNum->timerCtrl, DAQ_NumericTimerCallback, daqNum));

    errChk( ChainCtrlCallback(panel, ctrl,  DAQ_NumericCallback, (void *)daqNum, DAQ_NUMERIC_TYPE));


    if (newDaqNum)
        *newDaqNum = daqNum;

Error:
    if (error < 0)
        {
        if (newDaqNum)
            *newDaqNum = 0;
        return error;
        }
    else
        return 0;
}

/*****************************************************************************************/

static void DiscardDAQ_NumericInfo(DAQ_NumericInfo daqNum)
{
    if (daqNum)
        {
        DiscardCtrlNoBOLE(daqNum->panel, daqNum->timerCtrl, TRUE);
        if (daqNum->panel > 0 && daqNum->baseCtrl > 0)
            UnchainCtrlCallback(daqNum->panel, daqNum->baseCtrl, DAQ_NUMERIC_TYPE);
        
        DaqCtrl_DisposeDaqCtrl((DaqCtrl)daqNum);
        }
}

/*****************************************************************************************/

int CVIFUNC DAQ_Numeric_Create(int panel, int style, char *label, int ioDirection, int top, int left, char *configurationName, CtrlCallbackPtr callback, void *callbackData)
{
    return Internal_DAQ_Numeric_Create(panel, style, label, ioDirection, top,
                                       left, configurationName, callback,
                                       callbackData);
}

static int Internal_DAQ_Numeric_Create(int panel, int style, char *label, int ioDirection, int top, int left, char *configurationName, CtrlCallbackPtr callback, void *callbackData)
{
    int baseCtrl = 0;
    int error  = 0;

    errChk( baseCtrl = NewCtrl (panel, style, label, top, left));
    errChk( InstallCtrlCallback(panel, baseCtrl, callback, callbackData));

    errChk( SetCtrlModeToIODirection(panel, baseCtrl, ioDirection));

    SetCtrlAttribute(panel, baseCtrl, ATTR_MAX_VALUE, ATTR_DAQ_NUMERIC_HIGH_LIMIT_DEFAULT);
    SetCtrlAttribute(panel, baseCtrl, ATTR_MIN_VALUE, ATTR_DAQ_NUMERIC_LOW_LIMIT_DEFAULT);

    return Internal_DAQ_Numeric_ConvertFromNumeric(panel, baseCtrl, configurationName);

Error:
    if (baseCtrl > 0)
        DiscardCtrlNoBOLE(panel, baseCtrl, TRUE);

    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Numeric_ConvertFromNumeric(int panel, int baseCtrl, char *configurationName)
{
    return Internal_DAQ_Numeric_ConvertFromNumeric(panel, baseCtrl, configurationName);
}

static int Internal_DAQ_Numeric_ConvertFromNumeric(int panel, int baseCtrl, char *configurationName)
{
    int             error = 0;
    int             style;
    int             dataType;
    int             ioDirection;
    int             ctrlMode;
    double          voltage;
    DAQ_NumericInfo daqNum = 0;

    errChk( GetCtrlAttribute(panel, baseCtrl, ATTR_CTRL_STYLE, &style));

    switch (style)
        {
        case CTRL_NUMERIC:
        case CTRL_NUMERIC_THERMOMETER:
        case CTRL_NUMERIC_TANK:
        case CTRL_NUMERIC_GAUGE:
        case CTRL_NUMERIC_METER:
        case CTRL_NUMERIC_KNOB:
        case CTRL_NUMERIC_DIAL:
        case CTRL_NUMERIC_VSLIDE:
        case CTRL_NUMERIC_HSLIDE:
        case CTRL_NUMERIC_FLAT_VSLIDE:
        case CTRL_NUMERIC_FLAT_HSLIDE:
        case CTRL_NUMERIC_LEVEL_VSLIDE:
        case CTRL_NUMERIC_LEVEL_HSLIDE:
        case CTRL_NUMERIC_POINTER_VSLIDE:
        case CTRL_NUMERIC_POINTER_HSLIDE:
            break;
        default:
            errChk(UIEInvalidControlType);
        }

    errChk( GetCtrlAttribute(panel, baseCtrl, ATTR_DATA_TYPE, &dataType));
        /* change type to double, but try and preserve min, max, and increment values */
    if (dataType != VAL_DOUBLE)
        {
        double  min, max, incr, value;

        errChk( GetNumericAttributeAsType(panel, baseCtrl, ATTR_MAX_VALUE, VAL_DOUBLE, &max));
        errChk( GetNumericAttributeAsType(panel, baseCtrl, ATTR_MIN_VALUE, VAL_DOUBLE, &min));
        errChk( GetNumericAttributeAsType(panel, baseCtrl, ATTR_INCR_VALUE, VAL_DOUBLE, &incr));
        errChk( GetNumericAttributeAsType(panel, baseCtrl, ATTR_CTRL_VAL, VAL_DOUBLE, &value));

        errChk( SetCtrlAttribute(panel, baseCtrl, ATTR_DATA_TYPE, VAL_DOUBLE));
        errChk( SetCtrlAttribute(panel, baseCtrl, ATTR_MAX_VALUE, max));
        errChk( SetCtrlAttribute(panel, baseCtrl, ATTR_MIN_VALUE, min));
        errChk( SetCtrlAttribute(panel, baseCtrl, ATTR_INCR_VALUE, incr));
        errChk( SetCtrlAttribute(panel, baseCtrl, ATTR_CTRL_VAL, value));
        }

    errChk( GetCtrlAttribute(panel, baseCtrl, ATTR_CTRL_MODE, &ctrlMode));
    if (ctrlMode != VAL_INDICATOR)
        ioDirection = ANALOG_OUTPUT;
    else
        ioDirection = ANALOG_INPUT;
        
    errChk( CreateDAQ_NumericInfo(panel, baseCtrl, &daqNum));

    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_DEVICE, FindDefaultDevice(ioDirection, NULL)));
    error = Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_CHANNEL_STRING, ATTR_DAQ_NUMERIC_CHANNEL_STRING_DEFAULT);
    if (error < 0)
        {
        /*  added by johnm - this would be a DAQ error, let it slip through in this case 
            so the DAQ control will still be created */
        daqNum->channelString = StrDup(ATTR_DAQ_NUMERIC_CHANNEL_STRING_DEFAULT);
        error = 0;
        }
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_HIGH_LIMIT, ATTR_DAQ_NUMERIC_HIGH_LIMIT_DEFAULT));

    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_LOW_LIMIT, ATTR_DAQ_NUMERIC_LOW_LIMIT_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_ENABLED, ATTR_DAQ_NUMERIC_ENABLED_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_VISIBLE, ATTR_DAQ_NUMERIC_VISIBLE_DEFAULT));

    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_ALLOW_RCLICK_EDIT, ATTR_DAQ_NUMERIC_ALLOW_RCLICK_EDIT_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_EDIT_DIALOG_TITLE, ATTR_DAQ_NUMERIC_EDIT_DIALOG_TITLE_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_EDIT_DIALOG_DISPLAYED, ATTR_DAQ_NUMERIC_EDIT_DIALOG_DISPLAYED_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_EDIT_DIALOG_MODAL, ATTR_DAQ_NUMERIC_EDIT_DIALOG_MODAL_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_ERROR_NOTIFICATION, ATTR_DAQ_NUMERIC_ERROR_NOTIFICATION_DEFAULT));

#if ATTR_DAQ_NUMERIC_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT
    if (DAQ_NumericHasReasonableRangeForSlavingToLimits(panel, baseCtrl))
        errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_TIE_CTRL_RANGE_TO_LIMITS, TRUE));
#endif

    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_IO_DIRECTION, ioDirection));

        /*  update the voltage output if needed */
    errChk( GetCtrlVal(panel, baseCtrl, &voltage));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, baseCtrl, ATTR_DAQ_NUMERIC_CTRL_VAL, voltage));

    if (configurationName && *configurationName)
        DAQ_Numeric_RecallConfig(panel, baseCtrl, configurationName);   /* don't report config recall errors when creating controls, they can call DAQ_Numeric_RecallConfig() if it matters */

    return baseCtrl;

Error:
    DiscardDAQ_NumericInfo(daqNum);
    return error;
}

/*****************************************************************************************/

#if ATTR_DAQ_NUMERIC_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT
static int  DAQ_NumericHasReasonableRangeForSlavingToLimits(int panel, int daqNumeric)
{
    double  max, min;

    GetCtrlAttribute(panel, daqNumeric, ATTR_MAX_VALUE, &max);
    GetCtrlAttribute(panel, daqNumeric, ATTR_MIN_VALUE, &min);

    if (max <= TIE_CTRL_RANGE_TO_LIMITS_MAX && min >= TIE_CTRL_RANGE_TO_LIMITS_MIN)
        return TRUE;
    else
        return FALSE;
}
#endif
/*****************************************************************************************/

static void CVICALLBACK DeferredOutputFunction (void *daqNum)
{
    ((DAQ_NumericInfo)daqNum)->outputInvalidated = FALSE;
    UpdateOutput((DAQ_NumericInfo)daqNum);
}

/*****************************************************************************************/

    /*  This function is used to cause the output (if the control is in ouput mode) to
        be refreshed at a later time.  This is called when the channel, or device is
        changed.  We could update the output immediately, but then it would be impossible
        to move to both a new channel and device without causing an output at either the wrong
        channel or the wrong device.
    */
static int InvalidateOutput(DAQ_NumericInfo daqNum)
{
    int error = 0;

    if (!daqNum->outputInvalidated)
        {
        errChk(PostDeferredCall(DeferredOutputFunction, (void *)daqNum));
        daqNum->outputInvalidated = TRUE;
        }

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC_C DAQ_Numeric_SetAttribute(int panel, int daqNumeric, int attribute, ...)
{
    int             error = 0;
    va_list         parmInfo;

    va_start(parmInfo, attribute);
    error = DAQ_Numeric_SetAttributeFromParmInfo(panel, daqNumeric, attribute, parmInfo);
    va_end(parmInfo);

    return error;
}

static int CVIFUNC_C Internal_DAQ_Numeric_SetAttribute(int panel, int daqNumeric, int attribute, ...)
{
    int             error = 0;
    va_list         parmInfo;

    va_start(parmInfo, attribute);
    error = DAQ_Numeric_SetAttributeFromParmInfo(panel, daqNumeric, attribute, parmInfo);
    va_end(parmInfo);

    return error;
}

int CVIFUNC DAQ_Numeric_SetAttributeFromParmInfo(int panel, int daqNumeric, int attribute, va_list parmInfo)
{
    int             error = 0;
    int             newIntValue;
    double          newDoubleValue;
    DAQ_NumericInfo daqNum = 0;

    errChk( GetDAQCtrlInfo(panel, daqNumeric, DAQ_NUMERIC_TYPE, (DaqCtrl *)&daqNum));

    switch (attribute)
        {
        case ATTR_DAQ_NUMERIC_DEVICE:
        case ATTR_DAQ_NUMERIC_CHANNEL_STRING:
        case ATTR_DAQ_NUMERIC_IO_DIRECTION:
            errChk( DaqCtrl_SetAttributeFromParmInfo((DaqCtrl)daqNum, attribute, parmInfo));
            errChk( InvalidateOutput(daqNum));
            daqNum->autoErrorClearingNeeded = daqNum->ioDirection == ANALOG_OUTPUT || !daqNum->enabled;
            break;
        case ATTR_DAQ_NUMERIC_SCAN_FREQUENCY:
            newDoubleValue = (double)va_arg(parmInfo, double);
            if (newDoubleValue != Pin(newDoubleValue, DAQ_NUMERIC_MIN_SCAN_FREQUENCY, DAQ_NUMERIC_MAX_SCAN_FREQUENCY))
                errChk( UIEValueIsInvalidOrOutOfRange);

            errChk( SetCtrlAttribute(panel, daqNum->timerCtrl, ATTR_INTERVAL, (double) 1.0 / newDoubleValue));
            break;
        case ATTR_DAQ_NUMERIC_ENABLED:
            newIntValue = va_arg(parmInfo, int);
            errChk( SetCtrlAttribute(panel, daqNum->timerCtrl, ATTR_ENABLED, newIntValue));
            daqNum->enabled = newIntValue;
            if (!daqNum->enabled)
                DaqCtrl_ClearDisplayedError((DaqCtrl)daqNum);
            daqNum->autoErrorClearingNeeded = daqNum->ioDirection == ANALOG_OUTPUT || !daqNum->enabled;
            UpdateOutput(daqNum);
            break;
        case ATTR_DAQ_NUMERIC_VISIBLE:
        case ATTR_DAQ_NUMERIC_ALLOW_RCLICK_EDIT:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_TITLE:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_DISPLAYED:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_MODAL:
        case ATTR_DAQ_NUMERIC_VERSION:
        case ATTR_DAQ_NUMERIC_ERROR_NOTIFICATION:
            errChk( DaqCtrl_SetAttributeFromParmInfo((DaqCtrl)daqNum, attribute, parmInfo));
            break;
        case ATTR_DAQ_NUMERIC_TIE_CTRL_RANGE_TO_LIMITS:
            newIntValue = va_arg(parmInfo, int);

            if (newIntValue != daqNum->tieCtrlRangeToLimits)
                {
                daqNum->tieCtrlRangeToLimits = newIntValue;

                if (daqNum->tieCtrlRangeToLimits)
                    {
                    double  max, min;

                    Internal_DAQ_Numeric_GetAttribute(panel, daqNumeric, ATTR_DAQ_CTRL_LOW_LIMIT, &min);
                    Internal_DAQ_Numeric_GetAttribute(panel, daqNumeric, ATTR_DAQ_CTRL_HIGH_LIMIT, &max);
                    SetCtrlAttribute(panel, daqNumeric, ATTR_MIN_VALUE, min);
                    SetCtrlAttribute(panel, daqNumeric, ATTR_MAX_VALUE, max);
                    }
                }
            break;
        case ATTR_DAQ_NUMERIC_HIGH_LIMIT:
            newDoubleValue = (double)va_arg(parmInfo, double);
            if (newDoubleValue <= daqNum->lowLimit)
                errChk( UIEAxisRangeError);
            daqNum->highLimit = newDoubleValue;

            if (daqNum->tieCtrlRangeToLimits)
                {
                SetCtrlAttribute(panel, daqNumeric, ATTR_MAX_VALUE, daqNum->highLimit);
                if (daqNum->dialog > 0)
                    SetCtrlAttribute(daqNum->dialog, daqNum->valueNumeric, ATTR_MAX_VALUE, daqNum->highLimit);
                }
            break;
        case ATTR_DAQ_NUMERIC_LOW_LIMIT:
            newDoubleValue = (double)va_arg(parmInfo, double);
            if (newDoubleValue >= daqNum->highLimit)
                errChk( UIEAxisRangeError);
            daqNum->lowLimit = newDoubleValue;

            if (daqNum->tieCtrlRangeToLimits)
                {
                SetCtrlAttribute(panel, daqNumeric, ATTR_MIN_VALUE, daqNum->lowLimit);
                if (daqNum->dialog > 0)
                    SetCtrlAttribute(daqNum->dialog, daqNum->valueNumeric, ATTR_MIN_VALUE, daqNum->lowLimit);
                }
            break;
        case ATTR_DAQ_NUMERIC_CTRL_VAL:
            newDoubleValue = (double)va_arg(parmInfo, double);
            errChk( SetCtrlVal(panel, daqNumeric, newDoubleValue));
            UpdateOutput(daqNum);
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }

        /* make sure the editing dialog displays the latest changes */
    if (!daqNum->updatingCtrlFromDialog && daqNum->dialog && daqNum->dialogDisplayed)
        {
        errChk(CopyAttributesToDialog(daqNum));
        errChk( UpdateDialogDimming(daqNum));
        }

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Numeric_GetAttribute(int panel, int daqNumeric, int attribute, void *value)
{
    return Internal_DAQ_Numeric_GetAttribute(panel, daqNumeric, attribute, value);
}

static int CVIFUNC Internal_DAQ_Numeric_GetAttribute(int panel, int daqNumeric, int attribute, void *value)
{
    int             error = 0;
    DAQ_NumericInfo daqNum = 0;

    if (!value)
        errChk( UIENullPointerPassed);

    errChk( GetDAQCtrlInfo(panel, daqNumeric, DAQ_NUMERIC_TYPE, (DaqCtrl *)&daqNum));

    switch (attribute)
        {
        case ATTR_DAQ_NUMERIC_DEVICE:
        case ATTR_DAQ_NUMERIC_CHANNEL_STRING:
        case ATTR_DAQ_NUMERIC_CHANNEL_STRING_COPY:
        case ATTR_DAQ_NUMERIC_CHANNEL_STRING_LENGTH:
        case ATTR_DAQ_NUMERIC_VISIBLE:
        case ATTR_DAQ_NUMERIC_ALLOW_RCLICK_EDIT:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_TITLE:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_TITLE_COPY:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_TITLE_LENGTH:
        case ATTR_DAQ_NUMERIC_IO_DIRECTION:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_DISPLAYED:
        case ATTR_DAQ_NUMERIC_EDIT_DIALOG_MODAL:
        case ATTR_DAQ_NUMERIC_VERSION:
        case ATTR_DAQ_NUMERIC_ERROR_NOTIFICATION:
        case ATTR_DAQ_NUMERIC_TIE_CTRL_RANGE_TO_LIMITS:
        case ATTR_DAQ_NUMERIC_HIGH_LIMIT:
        case ATTR_DAQ_NUMERIC_LOW_LIMIT:
            errChk( DaqCtrl_GetAttribute((DaqCtrl)daqNum, attribute, value));
            break;
        case ATTR_DAQ_NUMERIC_SCAN_FREQUENCY:
            {
            double  interval;

            errChk( GetCtrlAttribute(panel, daqNum->timerCtrl, ATTR_INTERVAL, &interval));
            *(double *)value = 1.0 / interval;
            }
            break;
        case ATTR_DAQ_NUMERIC_ENABLED:
            errChk( GetCtrlAttribute(panel, daqNum->timerCtrl, ATTR_ENABLED, value));
            break;
        case ATTR_DAQ_NUMERIC_CTRL_VAL:
            errChk( GetCtrlVal(panel, daqNumeric, value));
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }

Error:
    return error;
}

/*****************************************************************************************/

    /* the callback data is the DAQ_NumericInfo */
static int CVICALLBACK DAQ_NumericTimerCallback(int panel, int timerControl, int event, void *callbackData, int eventData1, int eventData2)
{
    DAQ_NumericInfo daqNum;

    UNUSED(panel);
    UNUSED(timerControl);
    UNUSED(eventData1);
    UNUSED(eventData2);

    daqNum = (DAQ_NumericInfo)callbackData;

    switch (event)
        {
        case EVENT_TIMER_TICK:
            if (daqNum->ioDirection == ANALOG_OUTPUT)
                { /*UpdateOutput(daqNum);*/}
            else
                UpdateInput(daqNum);
            break;
        case EVENT_DISCARD:
            daqNum->timerCtrl = 0;  /* in case the panel is discarded, which might automatically discard the timer before the DaqNum can discard it */
            break;
        }

    return 0;
}

/*****************************************************************************************/

static void UpdateInput(DAQ_NumericInfo daqNum)
{
    int             daqError = 0;
    int             channelStringPresent;
    double          voltage = 0;
    int             breakOnLibraryErrors = SetBOLE (0);
    double          now, elapsedTime;

    if (daqNum->ioDirection != ANALOG_OUTPUT)
        {
        channelStringPresent = daqNum->channelString && *daqNum->channelString;

        if (channelStringPresent)
            daqError  = AISampleChannel (daqNum->device, daqNum->channelString, daqNum->highLimit, daqNum->lowLimit, &voltage);


        DaqCtrl_DisplayError((DaqCtrl)daqNum, daqError, 0.0, 0.0);

        CallCtrlCallback(daqNum->panel, daqNum->baseCtrl, EVENT_DAQ_NUMERIC_DATA_READY, (int)&voltage, daqError, 0);

        if (channelStringPresent && daqError >= 0)
            {
            SetCtrlVal(daqNum->panel, daqNum->baseCtrl, voltage);
            if (daqNum->dialog && daqNum->dialogDisplayed)
                SetCtrlVal(daqNum->dialog, daqNum->valueNumeric, voltage);

            now = Timer();
            elapsedTime = now - daqNum->lastFreqDisplayTime;
            daqNum->numTicksSinceLastFreqDisplay++;

            if (elapsedTime > DAQ_CTRL_MIN_ACTUAL_FREQ_UPDATE_INTERVAL)
                {
                if (daqNum->lastFreqDisplayTime > 0.0)      /* if not the first update */
                    SetCtrlVal(daqNum->dialog, daqNum->actualFrequencyNumeric, daqNum->numTicksSinceLastFreqDisplay / elapsedTime);

                daqNum->lastFreqDisplayTime = now;
                daqNum->numTicksSinceLastFreqDisplay = 0;
                }
            }
        }

    SetBOLE(breakOnLibraryErrors);
}

/*****************************************************************************************/

static void  UpdateOutput(DAQ_NumericInfo daqNum)
{
    int     daqError = 0;
    int     channelStringPresent;
    double  voltage = 0;
    int     breakOnLibraryErrors = SetBOLE(0);

    if (daqNum->ioDirection == ANALOG_OUTPUT && daqNum->enabled)
        {
        GetCtrlVal(daqNum->panel, daqNum->baseCtrl, &voltage);

        channelStringPresent = daqNum->channelString && *daqNum->channelString;

        if(channelStringPresent)
            {
            daqError = AOUpdateChannel(daqNum->device, daqNum->channelString, voltage);
            daqNum->lastValueOutput = voltage;
            }

        DaqCtrl_DisplayError((DaqCtrl)daqNum, daqError, 0.0, MIN_DAQ_ERROR_DISPLAY_TIME_TILL_CLEARED);
        }

    SetBOLE(breakOnLibraryErrors);
}

/*****************************************************************************************/

static int CVICALLBACK DAQ_NumericCallback(int panel, int daqNumeric, int event, void *callbackData, int eventData1, int eventData2)
{
    int             error = 0;
    DAQ_NumericInfo daqNum = (DAQ_NumericInfo)callbackData;
    double          voltage;

    UNUSED(eventData1);
    UNUSED(eventData2);

    switch (event)
        {
        case EVENT_DISCARD:
            DiscardDAQ_NumericInfo(daqNum);
            break;
        case EVENT_RIGHT_CLICK:
            {
            int allowRightClickEdits;
            int result;

            errChk( Internal_DAQ_Numeric_GetAttribute(panel, daqNumeric, ATTR_DAQ_NUMERIC_ALLOW_RCLICK_EDIT, &allowRightClickEdits));

            if (allowRightClickEdits)
                {
                errChk( result = DaqCtrl_RunRightClickPopupMenu(panel, MakePoint(eventData2, eventData1)));
                if (result)
                    errChk( Internal_DAQ_Numeric_SetAttribute(panel, daqNumeric, ATTR_DAQ_NUMERIC_EDIT_DIALOG_DISPLAYED, 1));
                }
            }
            break;
        case EVENT_VAL_CHANGED:
        case EVENT_COMMIT:
            UpdateOutput(daqNum);
            if (daqNum->dialog && daqNum->dialogDisplayed)
                {           /* update voltage control in dialog */
                errChk( Internal_DAQ_Numeric_GetAttribute(panel, daqNumeric, ATTR_DAQ_NUMERIC_CTRL_VAL, &voltage));
                errChk( SetCtrlVal(daqNum->dialog, daqNum->valueNumeric, voltage));
                }
            break;
        }

Error:
    return 0;
}

/*********************************************************************************/

static int CreateDummyEditingDAQNumeric(int *panel, int *dummyNumeric, int modal)
{
    int error = 0;
    int width, height;
    int labelTop;
    int margin = 3 * DLG_MARGIN;

    *panel = 0;
    errChk( *panel = NewPanel (0, "DAQ Numeric For Quick Configuration Editing", 0, 0, 10, 10));
    errChk( *dummyNumeric = Internal_DAQ_Numeric_Create (*panel, CTRL_NUMERIC_VSLIDE, "Temporary DAQ Numeric For Configuration Editing", ANALOG_OUTPUT, margin, margin, 0, 0, 0));

    errChk( GetCtrlAttribute(*panel, *dummyNumeric, ATTR_LABEL_TOP, &labelTop));
    labelTop -= 10;
    errChk( SetCtrlAttribute(*panel, *dummyNumeric, ATTR_LABEL_TOP, labelTop));

    errChk( SetCtrlAttribute(*panel, *dummyNumeric, ATTR_CHECK_RANGE, VAL_IGNORE));
    errChk( GetCtrlBoundingRect(*panel, *dummyNumeric, 0, 0, &height, &width));

    width += 2 * margin + 75;   /* extra 75 to give the title a chance to show */
    height += 2 * margin;

    errChk( CenterCtrl(*panel, *dummyNumeric, 0, 0, height, width));

    errChk( SetPanelAttribute(*panel, ATTR_WIDTH, width));
    errChk( SetPanelAttribute(*panel, ATTR_HEIGHT, height));

    errChk( SetPanelAttribute (*panel, ATTR_CLOSE_ITEM_VISIBLE, FALSE));

    errChk( SetPanelPos(*panel, VAL_AUTO_CENTER, VAL_AUTO_CENTER));
    errChk( Internal_DAQ_Numeric_SetAttribute (*panel, *dummyNumeric, ATTR_DAQ_NUMERIC_HIGH_LIMIT, ATTR_DAQ_NUMERIC_HIGH_LIMIT_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute (*panel, *dummyNumeric, ATTR_DAQ_NUMERIC_LOW_LIMIT, ATTR_DAQ_NUMERIC_LOW_LIMIT_DEFAULT));
    errChk( Internal_DAQ_Numeric_SetAttribute (*panel, *dummyNumeric, ATTR_DAQ_NUMERIC_TIE_CTRL_RANGE_TO_LIMITS, TRUE));

    if (modal)
        errChk( InstallPopup(*panel));
    else
        errChk( DisplayPanel(*panel));

Error:
    if (error < 0 && *panel > 0)
        {
        DiscardPanel(*panel);
        *panel = 0;
        }

    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Numeric_EditPopup(int panel, int daqNumeric, char *dialogTitle, int modal)
{
    int     error = 0;
    int     deletePanel = FALSE;
    DaqCtrl daqCtrl = 0;

    if (panel == 0 && daqNumeric == 0)
        {
        errChk( CreateDummyEditingDAQNumeric(&panel, &daqNumeric, modal));
        deletePanel = TRUE;
        }

    errChk( GetDAQCtrlInfo(panel, daqNumeric, DAQ_NUMERIC_TYPE, &daqCtrl));
    errChk( DaqCtrl_EditPopup(daqCtrl, dialogTitle, modal));

Error:
    if (deletePanel && panel > 0)
        DiscardPanel(panel);

    return error;
}

/*****************************************************************************************/

    /*  copy the control's attributes to the dialog fields */
static int CVIFUNC CopyAttributesToDialog(DAQ_NumericInfo daqNum)
{
    int     error = 0;
    int     panel, ctrl, dialog;
    double  scanFrequency;
    double  value;

    dialog = daqNum->dialog;
    panel = daqNum->panel;
    ctrl = daqNum->baseCtrl;

    errChk( DaqCtrl_CopyAttributesToDialog((DaqCtrl)daqNum));

    errChk( Internal_DAQ_Numeric_GetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY, &scanFrequency));
    errChk( Internal_DAQ_Numeric_GetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_CTRL_VAL, &value));

    errChk( SetCtrlVal(daqNum->dialog, daqNum->scanFrequencyNumeric, scanFrequency));
    errChk( SetCtrlVal(daqNum->dialog, daqNum->valueNumeric, value));

    UpdateDialogDimming(daqNum);

Error:
    return error;
}

/*****************************************************************************************/

    /* save the current attribute values of the control */
static int CVIFUNC RememberAttributes(DAQ_NumericInfo daqNum)
{
    int error = 0;
    int panel = daqNum->panel;
    int ctrl  = daqNum->baseCtrl;

    errChk( DaqCtrl_RememberAttributes((DaqCtrl)daqNum));   /* base class functionality */

    errChk( Internal_DAQ_Numeric_GetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_CTRL_VAL, &daqNum->savedVoltageVal));
    errChk( Internal_DAQ_Numeric_GetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY, &daqNum->savedScanFrequency));

Error:
    return error;
}

/*****************************************************************************************/

    /* copy the saved values of the attributes back to the control */
static int CVIFUNC RevertAttributes(DAQ_NumericInfo daqNum)
{
    int error = 0;
    int panel = daqNum->panel;
    int ctrl = daqNum->baseCtrl;

    daqNum->updatingCtrlFromDialog = TRUE;
    daqNum->revertingSettings = TRUE;

    errChk( DaqCtrl_RevertAttributes((DaqCtrl)daqNum));

    errChk( Internal_DAQ_Numeric_SetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_CTRL_VAL, daqNum->savedVoltageVal));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY, daqNum->savedScanFrequency));

Error:
    daqNum->updatingCtrlFromDialog = FALSE;
    daqNum->revertingSettings = FALSE;
    return error;
}

/*****************************************************************************************/

    /* copy current dialog control settings to the control's attributes */
static int CVIFUNC ApplyDialogSettings(DAQ_NumericInfo daqNum)
{
    int     error = 0;
    int     dialog;
    int     panel;
    int     ctrl;
    double  scanFrequency;
    double  value;

    dialog = daqNum->dialog;
    panel = daqNum->panel;
    ctrl = daqNum->baseCtrl;

    daqNum->updatingCtrlFromDialog = TRUE;

    errChk( DaqCtrl_ApplyDialogSettings((DaqCtrl)daqNum));

    errChk( GetCtrlVal(dialog, daqNum->valueNumeric, &value));
    errChk( GetCtrlVal(dialog, daqNum->scanFrequencyNumeric, &scanFrequency));

    errChk( Internal_DAQ_Numeric_SetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY, scanFrequency));
    errChk( Internal_DAQ_Numeric_SetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_CTRL_VAL, value));

Error:
    daqNum->updatingCtrlFromDialog = FALSE;
    return error;
}

/*****************************************************************************************/

static int CVIFUNC UpdateDialogDimming(DAQ_NumericInfo daqNum)
{
    int     error = 0;
    int     ioDirection;
    int     dialog;
    int     labelLeft;

    dialog = daqNum->dialog;

    if (dialog > 0)
        {
        errChk( DaqCtrl_UpdateDialogDimming((DaqCtrl)daqNum));

        errChk( GetCtrlVal(dialog, daqNum->ioDirectionRing, &ioDirection));
        errChk( GetCtrlAttribute(dialog, daqNum->valueNumeric, ATTR_LABEL_LEFT, &labelLeft));   /* keep label from moving */
        errChk( SetCtrlAttribute(dialog, daqNum->valueNumeric, ATTR_CTRL_MODE, ioDirection == ANALOG_INPUT ? VAL_INDICATOR : VAL_HOT));
        errChk( SetCtrlAttribute(dialog, daqNum->valueNumeric, ATTR_LABEL_LEFT, labelLeft));
        errChk( SetCtrlAttribute(dialog, daqNum->scanFrequencyNumeric, ATTR_DIMMED, ioDirection != ANALOG_INPUT));
        errChk( SetCtrlAttribute(dialog, daqNum->actualFrequencyNumeric, ATTR_DIMMED, ioDirection != ANALOG_INPUT));
#if 0 /*removed by request of John M. Its more convenient to alway be able to access the limits */
        errChk( SetCtrlAttribute(dialog, daqNum->highLimitNumeric, ATTR_DIMMED, ioDirection != ANALOG_INPUT));
        errChk( SetCtrlAttribute(dialog, daqNum->lowLimitNumeric, ATTR_DIMMED, ioDirection != ANALOG_INPUT));
#endif
        }

Error:
    return error;
}

/*****************************************************************************************/

static int DAQ_Numeric_FormatPopup(DAQ_NumericInfo daqNum)
{
    int error = 0;

    errChk( DaqCtrl_CalcMinDialogWidth((DaqCtrl)daqNum));
    errChk( DaqCtrl_FormatDialog((DaqCtrl)daqNum));
    errChk( DaqCtrl_FinishDialogSizing((DaqCtrl)daqNum));
    ExpandCtrlToMeetLabel(daqNum->dialog, daqNum->channelStringCtrl, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);

Error:
    return error;
}

/*****************************************************************************************/

static int CVIFUNC DAQ_Numeric_BuildPopup(DAQ_NumericInfo daqNum)
{
    int error = 0;
    int savedWaitCursor;
    int dialog = 0;

    savedWaitCursor = GetWaitCursorState();
    SetWaitCursor(1);

    errChk( CreateDaqDlg((DaqCtrl)daqNum));

    dialog = daqNum->dialog;        /* initialized by CreateDaqDlg() */

    errChk( CreateDlgLists((DaqCtrl)daqNum));
    errChk( CreateErrorControls((DaqCtrl)daqNum));
    errChk( CreateConfigCtrls((DaqCtrl)daqNum));
    errChk( CreateGeneralCtrls((DaqCtrl)daqNum));
    errChk( CreateDAQCtrls((DaqCtrl)daqNum));
    errChk( CreateStdButtons((DaqCtrl)daqNum));
    errChk( CreateDlgSheetGroup((DaqCtrl)daqNum));

        /* Create the attribute ctrls */
    errChk( daqNum->valueNumeric =      NewCtrl (dialog, CTRL_NUMERIC, "Voltage:", 0, 0));
    errChk( daqNum->scanFrequencyNumeric =  NewCtrl (dialog, CTRL_NUMERIC, "Requested Scan Frequency (Hz):", 0, 0));
    errChk( daqNum->actualFrequencyNumeric =    NewCtrl (dialog, CTRL_NUMERIC, "Actual Scan Frequency (Hz):", 0, 0));

    errChk( MakeCtrlGrayRecessedIndicator(dialog, daqNum->actualFrequencyNumeric));
        /*** Build lists of the controls on the dialog to make formatting easier later */

    nullChk( AppendToIntegerList(daqNum->ctrlsWithLabelsOnLeft, 0, daqNum->valueNumeric, daqNum->scanFrequencyNumeric, daqNum->actualFrequencyNumeric, 0));

    nullChk( ListInsertBeforeItem(daqNum->daqTabCtrls, &daqNum->valueNumeric, &daqNum->deviceNumeric, 0));

    nullChk( ListInsertAfterItem(daqNum->daqTabCtrls, &daqNum->scanFrequencyNumeric, &daqNum->lowLimitNumeric, 0));
    nullChk( ListInsertAfterItem(daqNum->daqTabCtrls, &daqNum->actualFrequencyNumeric, &daqNum->scanFrequencyNumeric, 0));

    nullChk( AppendToIntegerList(daqNum->ctrlsOnLeft, 0, 0));   /* none so far */

    nullChk( daqNum->ctrlListsToDistribute = BuildIntegerList(0, daqNum->daqTabCtrls, daqNum->generalTabCtrlsToDistribute,
        daqNum->configTabCtrls, 0));

    errChk( EasyTab_AddSheet(dialog, daqNum->sheetGroupCtrl, "DAQ", NULL, 1, VAL_TAB_SHEET_FROM_LIST, daqNum->daqTabCtrls)); 
    errChk( EasyTab_AddSheet(dialog, daqNum->sheetGroupCtrl, "General", NULL, 1, VAL_TAB_SHEET_FROM_LIST, daqNum->generalTabCtrls)); 
    errChk( EasyTab_AddSheet(dialog, daqNum->sheetGroupCtrl, "Save/Recall", NULL, 1, VAL_TAB_SHEET_FROM_LIST, daqNum->configTabCtrls)); 

        /* Set data types */
    errChk( SetCtrlAttribute(dialog, daqNum->valueNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dialog, daqNum->scanFrequencyNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dialog, daqNum->actualFrequencyNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));

        /* Set Min and Max Values, etc.. */
    {           /* duplicate the numeric attributes of the base control on the value control */
    int     checkRange;
    double  min, max, inc;

    errChk( GetCtrlAttribute (daqNum->panel, daqNum->baseCtrl, ATTR_CHECK_RANGE, &checkRange));
    errChk( GetCtrlAttribute (daqNum->panel, daqNum->baseCtrl, ATTR_INCR_VALUE, &inc));
    errChk( GetCtrlAttribute (daqNum->panel, daqNum->baseCtrl, ATTR_MIN_VALUE, &min));
    errChk( GetCtrlAttribute (daqNum->panel, daqNum->baseCtrl, ATTR_MAX_VALUE, &max));

    errChk( SetCtrlAttribute (dialog, daqNum->valueNumeric, ATTR_CHECK_RANGE, checkRange));
    errChk( SetCtrlAttribute (dialog, daqNum->valueNumeric, ATTR_INCR_VALUE, inc));
    errChk( SetCtrlAttribute (dialog, daqNum->valueNumeric, ATTR_MIN_VALUE, min));
    errChk( SetCtrlAttribute (dialog, daqNum->valueNumeric, ATTR_MAX_VALUE, max));
    errChk( SetCtrlAttribute (dialog, daqNum->valueNumeric, ATTR_PRECISION, 5));
    }

    errChk( SetCtrlAttribute (dialog, daqNum->scanFrequencyNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( SetCtrlAttribute (dialog, daqNum->scanFrequencyNumeric, ATTR_INCR_VALUE, 0.1));
    errChk( SetCtrlAttribute (dialog, daqNum->scanFrequencyNumeric, ATTR_MIN_VALUE, DAQ_NUMERIC_MIN_SCAN_FREQUENCY));
    errChk( SetCtrlAttribute (dialog, daqNum->scanFrequencyNumeric, ATTR_MAX_VALUE, DAQ_NUMERIC_MAX_SCAN_FREQUENCY));

        /* Install Control Callbacks */

    errChk( InstallCtrlCallback (dialog, daqNum->valueNumeric, DAQ_CtrlDialogCallback, (void *)daqNum));
    errChk( InstallCtrlCallback (dialog, daqNum->scanFrequencyNumeric, DAQ_CtrlDialogCallback, (void *)daqNum));

        /* Arrange the controls and size the popup */
    errChk( DAQ_Numeric_FormatPopup(daqNum));

    errChk( SetPanelPos(dialog, VAL_AUTO_CENTER, VAL_AUTO_CENTER));

Error:
    if (error < 0)
        {
        if (dialog > 0)
            DiscardPanel(dialog);
        daqNum->dialog = 0;
        }

    SetWaitCursor(savedWaitCursor);
    return error;
}

/*****************************************************************************************/

static int CVIFUNC SaveSettings(DAQ_NumericInfo daqNum, char *settingsName)
{
    int     error = 0;
    IniText iniText = 0;
    char    section[MAX_SETTING_NAME_LENGTH + 128];     /* control prefix should never be anywhere near 128 characters */
    int     panel = daqNum->panel;
    int     ctrl = daqNum->baseCtrl;
    double  value;
    double  scanFrequency;

    errChk( ReadConfigIni(&iniText));

    sprintf(section, "%s%s", DAQ_NUMERIC_SETTINGS_PREFIX, settingsName);
    Ini_RemoveSection (iniText, section);

    errChk( DaqCtrl_WriteSettings((DaqCtrl)daqNum, iniText, section));

    errChk( Internal_DAQ_Numeric_GetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY, &scanFrequency));
    errChk( Internal_DAQ_Numeric_GetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_CTRL_VAL, &value));


    errChk( Ini_PutDouble   (iniText, section, SCAN_FREQUENCY_TAG, scanFrequency));
    errChk( Ini_PutDouble   (iniText, section, CTRL_VALUE_TAG, value));

    errChk( WriteConfigIni(iniText));
    strcpy(daqNum->settingsName, settingsName);     /* if success, this is the new settings name */

Error:
    Ini_Dispose (iniText);
    return error;
}

/****************************************************************************************/

    /*  errorDescription should be a pointer to a buffer initialized to an empty string, but large
        enough for a short error message description.

        All attributes are optional.
    */
static int CVIFUNC RecallSettings(DAQ_NumericInfo daqNum, char *settingsName, char *errorDescription)
{
    int     error = 0;
    IniText iniText = 0;
    char    section[MAX_SETTING_NAME_LENGTH + 256];
    int     panel = daqNum->panel;
    int     ctrl = daqNum->baseCtrl;
    double  scanFrequency;
    double  value;
    int     doingGets = FALSE;
    int     foundScanFrequency;
    int     foundValue;
    int     result, warning = 0;

    daqNum->recallingSettings = TRUE;

    errChk(ReadConfigIni(&iniText));
    sprintf(section, "%s%s", DAQ_NUMERIC_SETTINGS_PREFIX, settingsName);

    if (!Ini_SectionExists (iniText, section))
        {
        if (errorDescription)
            sprintf(errorDescription, "Couldn't find section '%s'", section);
        errChk( ToolErr_MissingSection);
        }

    errChk( result = DaqCtrl_ReadSettings((DaqCtrl)daqNum, iniText, section, errorDescription, &doingGets));
    if (!warning) warning = result;

    doingGets = TRUE;
    if (daqNum->lastRecalledVersion == DAQ_NUMERIC_BETA_VERSION)
        errChk( foundScanFrequency  = Ini_GetDouble (iniText, section, BETA_REFRESH_FREQUENCY_TAG, &scanFrequency));
    else
        errChk( foundScanFrequency  = Ini_GetDouble (iniText, section, SCAN_FREQUENCY_TAG, &scanFrequency));

    errChk( foundValue              = Ini_GetDouble (iniText, section, CTRL_VALUE_TAG, &value));
    doingGets = FALSE;

    if (foundScanFrequency)
        errChk( Internal_DAQ_Numeric_SetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_SCAN_FREQUENCY, (double)scanFrequency));

    if (foundValue)
        errChk( Internal_DAQ_Numeric_SetAttribute(panel, ctrl, ATTR_DAQ_NUMERIC_CTRL_VAL, (double)value));

    strcpy(daqNum->settingsName, settingsName);     /* if success, this is the new settings name */

Error:
    daqNum->recallingSettings = FALSE;

    if (error < 0)
        {
        if (doingGets && errorDescription && !*errorDescription)
            sprintf(errorDescription, "Error reading from line %d", Ini_LineOfLastAccess (iniText));    
        }
    else
        error = warning;

    Ini_Dispose (iniText);
    return error;
}

/**********************************************************/

int CVIFUNC DAQ_Numeric_RecallConfig(int panel, int daqNumeric, char *configurationName)
{
    int             error = 0;
    DAQ_NumericInfo daqNum = 0;

    errChk( GetDAQCtrlInfo(panel, daqNumeric, DAQ_NUMERIC_TYPE, (DaqCtrl *)&daqNum));

    if (configurationName && *configurationName)
        errChk( RecallSettings(daqNum, configurationName, 0));

    errChk( DaqCtrl_UpdateConfigDlgCtrl((DaqCtrl)daqNum));    /* update the config name in the dialog */

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/**********************************************************/

int CVIFUNC DAQ_Numeric_SaveConfig(int panel, int daqNumeric, char *configurationName)
{
    int             error = 0;
    DAQ_NumericInfo daqNum = 0;

    errChk( GetDAQCtrlInfo(panel, daqNumeric, DAQ_NUMERIC_TYPE, (DaqCtrl *)&daqNum));

    if (!configurationName)
        errChk( UIENullPointerPassed);

    if (*configurationName)
        errChk( SaveSettings(daqNum, configurationName));

    errChk( DaqCtrl_UpdateConfigDlgCtrl((DaqCtrl)daqNum));    /* update the config name in the dialog */

Error:
    if (error < 0)
        return error;
    else
        return 0;
}
