/*============================================================================*/
/*                        L a b W i n d o w s / C V I                         */
/*----------------------------------------------------------------------------*/
/*    Copyright (c) National Instruments 1987-1996.  All Rights Reserved.     */
/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Title:       daqchart.c                                                    */
/* Purpose:     implements DAQ stripchart controls                            */
/*                                                                            */
/*============================================================================*/

#include <ansi_c.h>
#include <utility.h>
#include <userint.h>
#include "easyio.h"
#include "custsupp.h"
#include "toolbox.h"
#include "easytab.h"
#include "daqctrl.h"
#include "datalog.h"
#include "daqchart.h"


#define ATTR_DAQ_CHART_CHANNEL_STRING_DEFAULT               ATTR_DAQ_CTRL_CHANNEL_STRING_DEFAULT
#define DAQ_CHART_DEFAULT_NUMBER_OF_CHANNELS                1
#define ATTR_DAQ_CHART_UPDATE_HIGH_LIMIT_DEFAULT            ATTR_DAQ_CTRL_HIGH_LIMIT_DEFAULT
#define ATTR_DAQ_CHART_UPDATE_LOW_LIMIT_DEFAULT             ATTR_DAQ_CTRL_LOW_LIMIT_DEFAULT
#define ATTR_DAQ_CHART_REFRESH_FREQUENCY_DEFAULT            10.0
#define ATTR_DAQ_CHART_SCAN_FREQUENCY_DEFAULT               10.0
#define ATTR_DAQ_CHART_ENABLED_DEFAULT                      ATTR_DAQ_CTRL_ENABLED_DEFAULT
#define ATTR_DAQ_CHART_VISIBLE_DEFAULT                      ATTR_DAQ_CTRL_VISIBLE_DEFAULT
#define ATTR_DAQ_CHART_ALLOW_RCLICK_EDIT_DEFAULT            ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT_DEFAULT
#define ATTR_DAQ_CHART_EDIT_DIALOG_TITLE_DEFAULT            "Edit DAQ Chart Properties"
#define ATTR_DAQ_CHART_EDIT_DIALOG_DISPLAYED_DEFAULT        ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED_DEFAULT
#define ATTR_DAQ_CHART_EDIT_DIALOG_MODAL_DEFAULT            ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL_DEFAULT
#define ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE_DEFAULT          500
#define ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG_DEFAULT          FALSE
#define ATTR_DAQ_CHART_WARN_ON_HW_TIMING_RESTARTS_DEFAULT   TRUE
#define ATTR_DAQ_CHART_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT     ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT
#define ATTR_DAQ_CHART_ERROR_NOTIFICATION_DEFAULT           ATTR_DAQ_CTRL_ERROR_NOTIFICATION_DEFAULT

#define TIME_COLUMN_HEADING                                 "Time (sec),"
#define MIN_DAQ_CHART_HISTORY_BUFFER_SIZE                   10

#define DAQ_CHART_WARNING_HW_TIMED_BUF_OVERRUN              -9000
#define DAQ_CHART_WARNING_HW_TIMED_TRANSFER_ABORTED         -9001
#define DAQ_CHART_WARNING_INVALID_CHANNEL                   -9002
#define DAQ_CHART_WARNING_DATALOGGING_TERMINATED            -9003
#define DAQ_CHART_ERROR_MAX_64_CHANNELS_AND_ALARMS          ToolErr_TooManyItems

#define MIN_HW_TIMED_BUFFER_SIZE            6
#define HW_TIMED_SCAN_BUF_LENGTH_IN_SECONDS 10

#define SCAN_CATCHUP_TIME               0.15    /* we try to read all pending hw timed scans at once, unless is takes longer than this time (we will take twice this amount if we are behind) */
#define SCAN_BACKLOG_UPDATE_INTERVAL    0.5     /* update the scan backlog numeric no more than twice a second (to avoid wasting time) */

#define DAQ_CHART_SETTINGS_PREFIX       "DAQCHART_" /* section prefix in configuration file */
#define DAQ_CHART_TYPE_NAME             "DAQ Chart"
#define DAQ_CHART_BETA_VERSION          1
#define DAQ_CHART_VERSION               2

#define DAQ_CHART_TYPE                  "DAQCHART"
#define MIN_DATALOG_PROGRESS_UPDATE_INTERVAL    0.5

#define HISTORY_SIZE_TAG                "History Buffer Size"
#define SCAN_FREQUENCY_TAG              "Scan Frequency"
#define REFRESH_FREQUENCY_TAG           "Refresh Frequency"
#define WARN_ON_RESTARTS_TAG            "Warn On Hardware Timed Acquisition Restarts"
#define HARDWARE_TIMING_TAG             "Use Hardware Timing"
#define NUM_ALARMS_TAG                  "Number Of Alarms"
#define ALARM_NAME_TAG                  "Alarm%d Name"
#define ALARM_CHANNEL_TAG               "Alarm%d Channel"
#define ALARM_LIMIT_TAG                 "Alarm%d Limit"
#define ALARM_LOWER_LIMIT_TAG           "Alarm%d Lower Limit"
#define ALARM_DEADBAND_TAG              "Alarm%d DeadBand"
#define ALARM_MODE_TAG                  "Alarm%d Mode"
#define ALARM_COLOR_TAG                 "Alarm%d Color"
#define ALARM_LOG_WHILE_ON_TAG          "Alarm%d Log While On"
#define ALARM_LOG_WHILE_OFF_TAG         "Alarm%d Log While Off"
#define ALARM_PLOT_TAG                  "Alarm%d PlotLimit"
#define NUM_TRACES_TAG                  "Number Of Traces"
#define TRACE_COLOR_TAG                 "Trace %d Color"

#define ALARM_TRIGGER_ABOVE_TAG_VAL         "Trigger Above Limit"
#define ALARM_TRIGGER_BELOW_TAG_VAL         "Trigger Below Limit"
#define ALARM_TRIGGER_IN_RANGE_TAG_VAL      "Trigger In Range"
#define ALARM_TRIGGER_OUT_OF_RANGE_TAG_VAL  "Trigger Out Of Range"

#define ALARMS_CONTROL_LOGGING_TAG          "Logging Controlled By Alarms"

#define ALARM_LOG_WHILE_ON                  (1L << 0)
#define ALARM_LOG_WHILE_OFF                 (1L << 1)

#define ASYNC_READ_CACHE_SIZE               500     /* size in scans */
#define MAX_SECONDS_OF_CACHED_SCANS         2.0

#define DATALOGFIELDS   \
    int     nameLogFileButton;          \
    int     dataLogFileString;          \
    int     eventLogFileString;         \
    int     logEventsCheckBox;          \
    int     fileFormatRing;             \
    int     textFormatButton;           \
    int     logControlTextMsg;          \
    int     logControlDeco;             \
    int     startLoggingToggle;         \
    int     pauseLoggingToggle;         \
    int     loggingStateString;         \
    int     startOnLoadCheckBox;        \
    int     advancedLoggingButton;      \
    int     convertLogFileButton;       \
    int     savedLoggingOn;             \
    int     savedLoggingPaused;         \
    int     savedLogEvents;             \
    int     savedFileFormat;            \
    int     savedStartOnLoad;           \
    char    savedDataLogFilePath[MAX_PATHNAME_LEN]; \
    int     savedPromptOnOverwrite;     \
    int     savedFieldWidth;            \
    int     savedWriteHeader;           \
    char    *savedColumnHeadings;       \
    char    *savedComment;              \
    int     savedLogMode;               \
    int     savedNumScans;              \
    int     savedLogConditionally;      \
    DataLog dataLog;                    \


typedef struct DAQ_ChartInfo_Struct
    {
    COMMON_DAQ_CTRL_FIELDS      /* macro from daqctrl.h containing operational fields common to all daq ctrls */

    /* daq chart fields */
    int         groupSize;
    int         numChannels;
    int         scanSize;
    double      *dataHistory;
    double      *timeHistory;
    int         nextScanIndex;
    int         nextPlotIndex;
    int         historyBufferSize;
    int         historyLength;
    ListType    alarmList;
    int         numAlarmsToPlot;
    int         numAlarmLimitsToPlot;
    int         useHardwareTiming;
    double      hwTimedAcqStartTime;
    int         hwTimedAcqBufferSize;
    int         hwTimedAcqIsFallingBehind;  /* TRUE if we are not keeping up with the hardware scan rate */
    unsigned long   hwTimedScanNum;
    unsigned long   numScansLeft;
    unsigned long   asyncTaskId;
    int         warnOnHWTimedAcqRestarts;
    int         refreshTimerCtrl;
    int         scanTimerCtrl;
    double      scanFrequency;
    int         alarmsControlLogging;
    double      *scanCache;
    int         readCacheIndex;
    int         readCacheNumScans;
    double      savedRefreshFrequency;
    double      savedScanFrequency;
    int         savedUseHardwareTiming;
    int         savedHistorySize;
    ListType    savedTraceColorList;
    ListType    savedAlarmList;
    ListType    alarmTabCtrls;
    ListType    dataLogTabCtrls;

    int         frequencyNumeric;
    int         actualFrequencyNumeric;
    int         refreshNumeric;
    int         refreshRing;
    int         refreshPopupMenuBar;
    int         refreshPopupMenu;
    int         historySizeNumeric;
    int         useHardwareTimingCheckBox;
    int         scanBackLogNumeric;
    int         editColorsButton;
    int         editAlarmButton;
    int         alarmListBox;
    int         newAlarmButton;
    int         deleteAlarmButton;
    double      nextScanBackLogUpdateTime;
    double      lastActualFreqDisplayTime;
    int         numScansSinceLastActualFreqDisplayTime;
    double      lastDatalogStateDisplayTime;
    DATALOGFIELDS
    } * DAQ_ChartInfo;

typedef struct
    {
    char    *name;
    char    *channel;
    double  limit;
    double  lowerLimit;
    double  deadband;
    int     mode;
    int     plot;
    int     channelIndex;
    int     color;
    int     on;
    int     valid;
    int     logAction;
    int     traceIndex;     /* index of the alarm's first limit trace.  Only valid if the alarm is in the chart's alarm list and its limits are being plotted */
    } Alarm;

Alarm gDefaultAlarm =
    {
    "Untitled Alarm",               /*  name */
    "0",                            /*  channel */
    5.0,                            /*  limit */
    -5.0,                           /*  lower limit */
    0.01,                           /*  dead band */
    VAL_ALARM_ABOVE_LIMIT,      /*  mode */
    TRUE,                           /*  plot */
    0,                              /*  channel index */
    VAL_RED,                        /*  color */
    FALSE,                          /*  on (default not used)*/
    TRUE,                           /*  valid (default not used) */
    0                               /*  log action */
    };

static int Internal_DAQ_Chart_Create(int panel, char *label, int top, int left, char *configurationName,  CtrlCallbackPtr callback, void *callbackData);
static int CVIFUNC_C Internal_DAQ_Chart_SetAttribute(int panel, int daqChart, int attribute, ...);
static int Internal_DAQ_Chart_GetHistory(int panel, int daqChart, int *numPoints, int *numScans,
        int *channelsPerScan, int groupDataByChannels, double samples[], double sampleTimes[], int remove);

static int Internal_DAQ_Chart_GetHistoryCopy(int panel, int daqChart, int *numPoints, int *numScans,
        int *channelsPerScan, int groupDataByChannels, double **samples,
        double **sampleTimes, int remove);

static int Internal_DAQ_Chart_CreateAlarm(int panel, int daqChart, char *alarmName, char *channel,
            double limit, double lowerLimit, double deadband, int mode, int plotTheLimit, int traceColor);
static int Internal_DAQ_Chart_RecallConfig(int panel, int daqChart, char *configurationName);

static int  CVICALLBACK DAQ_ChartRefreshTimerCallback(int panel, int timerControl, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK DAQ_ChartScanTimerCallback(int panel, int timerControl, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK DAQ_ChartCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CreateDAQ_ChartInfo(int panel, int ctrl, DAQ_ChartInfo *chartInfo);
static void DiscardDAQ_ChartInfo(DAQ_ChartInfo chartInfo);

static int  ConfigureChartHWTimingMode(DAQ_ChartInfo chartInfo);
static int  GetSingleScan(DAQ_ChartInfo chartInfo, double *scan, double *time, int *scanReady, int *moreScansLeft);
static int  SetChartHWTimingMode(DAQ_ChartInfo chartInfo, int useHardwareTiming);
static void IncrementScanPosition(DAQ_ChartInfo chartInfo);
static void GetPlottedAlarmLimits(DAQ_ChartInfo chartInfo, double *dest);
static int  CopyAlarm(Alarm *src, Alarm  *dest);
static int  AllocateAlarmResources(Alarm *alarm, char *alarmName, char *channel, double limit, double lowerLimit,
            double deadband, int mode, int plotTheLimit, int traceColor, int logAction);
static void FreeAlarmResources(Alarm *alarm);
static void FreeAlarmList(ListType alarmList);
static int  WriteOutAlarms(DAQ_ChartInfo chartInfo, IniText iniText, char *section);
static int  NumPendingScansToPlot(DAQ_ChartInfo chartInfo);
static int  PlotPendingScans(DAQ_ChartInfo chartInfo);
static int  Safe_AIClearAcquisition(DAQ_ChartInfo chartInfo);
static int  UpdateChartHWTimingMode(DAQ_ChartInfo chartInfo);
static int  HandleHWTimedAcqOverrun(DAQ_ChartInfo chartInfo);
static void DoHWTimedTransferRestartWarning(DAQ_ChartInfo chartInfo, int errorCode);
static void CVICALLBACK FinishHWTimedTransferRestartWarning(void *callbackData);
static int  GetSingleHWTimedScan(DAQ_ChartInfo chartInfo, double *scan, double *time, int *scanReady, int *moreScansLeft);
static int  GetIndexFromChannel(DAQ_ChartInfo chartInfo, char *channel);
static int  WriteOutTraceColors(DAQ_ChartInfo chartInfo, IniText iniText, char *section);
static int  ReplaceAllChartAlarms(DAQ_ChartInfo chartInfo, ListType newAlarmList);
static int  ReadInTraceColors(DAQ_ChartInfo chartInfo, IniText iniText, char *section, char *errorDescription, int *doingGets);
static int  IsRangeMode(int limitMode);
static int  ReadInAlarms(DAQ_ChartInfo chartInfo, IniText iniText, char *section, char *errorDescription, int *doingGets);
static int  BetaReadInAlarms(DAQ_ChartInfo chartInfo, IniText iniText, char *section, char *errorDescription, int *doingGets);
static Alarm *AlarmPtr(DAQ_ChartInfo chartInfo, char *alarmName);
static int  Internal_GetAlarmAttribute(DAQ_ChartInfo chartInfo, char *alarmName, int attribute, void *value);
static int  Internal_SetAlarmAttributeFromParmInfo(DAQ_ChartInfo chartInfo, char *alarmName, int attribute, va_list parmInfo);
static int  MakeAlarmNameUnique(DAQ_ChartInfo chartInfo, Alarm *alarm);

    /*  edit dialog support functions */
static int  CVIFUNC BuildDaqChartPopup(DAQ_ChartInfo chartInfo);
static int  CVIFUNC CopyAttributesToDialog(DAQ_ChartInfo chartInfo);
static int  CVIFUNC ApplyDialogSettings(DAQ_ChartInfo chartInfo);
static int  CVIFUNC RememberAttributes(DAQ_ChartInfo chartInfo);
static int  CVIFUNC RevertAttributes(DAQ_ChartInfo chartInfo);

static int  DAQ_Chart_FormatPopup(DAQ_ChartInfo chartInfo);
static int  UpdateInput(DAQ_ChartInfo chartInfo);
static int  CVIFUNC UpdateDialogDimming(DAQ_ChartInfo chartInfo);
static int  SetDaqChartChannelString(DAQ_ChartInfo chartInfo, char *newValue);
static void DisposeHistory(DAQ_ChartInfo chartInfo);
static int  CreateHistory(DAQ_ChartInfo chartInfo, int newSize, int numChannels);
static void DisposeAllChartAlarms(DAQ_ChartInfo chartInfo);
static void DisposeAlarmByIndex(DAQ_ChartInfo chartInfo, int index);
static int  UpdateNumTraces(DAQ_ChartInfo chartInfo);
static int  PlotScan(DAQ_ChartInfo chartInfo, double *scan);
static void CheckAlarms(DAQ_ChartInfo chartInfo, double *scan, double *time);
static void SendAlarmEvent(DAQ_ChartInfo chartInfo, Alarm *alarm, int event, double value, double *time);
static void GetTraceAttributes(int panel, int stripChart, int trace, int *color,
                int *pointStyle, int *plotStyle, int *lineStyle);
static int  CVICALLBACK RefreshRingCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CVIFUNC GetDAQChartErrorString(DAQ_ChartInfo chartInfo, int errorCode, char **errorTitle, char **errorMessage);
static void CVICALLBACK DAQ_Chart_ApplyColorChanges(ListType colorList, void *callbackData, int colorListIndex);
static int  SelectChartColors(DAQ_ChartInfo chartInfo);
static int  CVICALLBACK EditColorsCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK EditAlarmsCtrlCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static ListType CreateChartColorList(DAQ_ChartInfo chartInfo);
static void DisposeChartColorList(ListType colorList);
static int  DAQ_ChartHasReasonableRangeForSlavingToLimits(int panel, int daqChart);
static int  UpdateAlarmDlgCtrls(DAQ_ChartInfo chartInfo);
static int  EditAlarmDialog(DAQ_ChartInfo chartInfo, char *dlgTitle, Alarm *alarm);
static int  ReplaceAlarmByIndex(DAQ_ChartInfo chartInfo, int indexOfAlarmToReplace, Alarm *replacementAlarm);
static int  RememberTraceColors(DAQ_ChartInfo chartInfo);
static int  RevertTraceColors(DAQ_ChartInfo chartInfo);
static void DisposeSavedAlarmList(DAQ_ChartInfo chartInfo);
static int  RevertAlarms(DAQ_ChartInfo chartInfo);
static int  RememberAlarms(DAQ_ChartInfo chartInfo);
static int  CreateDummyEditingDAQChart(int *panel, int *dummyChart, int modal);
static int  CVICALLBACK EditConfigDlgCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);

static int  CVIFUNC SaveSettings(DAQ_ChartInfo chartInfo, char *settingsName);
static int  CVIFUNC RecallSettings(DAQ_ChartInfo chartInfo, char *settingsName, char *errorDescription);
static void UpdateActualFrequencyDisplay(DAQ_ChartInfo chartInfo, double time);
static void UpdateScanBackLogDisplay(DAQ_ChartInfo chartInfo, double time);
static int  AdjustAlarmDlgForLimitMode(int dlg, int limitRing, int limitNumeric, int lowerLimitNumeric);
static int  CheckAlarmDlgLimitValidity(int dlg, int limitRing, int limitNumeric, int lowerLimitNumeric, int deadBandNumeric);
static void WarnIfAlarmsHaveInvalidChannels(DAQ_ChartInfo chartInfo);
static void CVIFUNC HandleErrorNotification(DAQ_ChartInfo chartInfo, int errorCode, char *errorMsg);
static int  CVIFUNC ConfirmRevert(DAQ_ChartInfo chartInfo, RevertReply *reply);

/* data logging functions */
static int  CreateLoggingControls(DAQ_ChartInfo chartInfo);
static int  FormatDataLogTab(DAQ_ChartInfo chartInfo);
static int  CalcDataLogTabMinDialogWidth(DAQ_ChartInfo chartInfo);
static int  CVICALLBACK NameLogFileBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK AdvancedLoggingBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK ConvertLogFileBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK TextFormatBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  ApplyDataLogDialogSettings(DAQ_ChartInfo chartInfo);
static int  UpdateDataLogDialogDimming(DAQ_ChartInfo chartInfo);
static int  RememberDataLogAttributes(DAQ_ChartInfo chartInfo);
static int  CopyDataLogAttrToDialog(DAQ_ChartInfo chartInfo);
static void CancelLoggingDueToAttributeChange(DAQ_ChartInfo chartInfo);
static void CVICALLBACK HandleDatalogNotification(void *callbackData, DataLogNotification event, void *eventData);
static int  UpdateLoggingStateIndicator(DAQ_ChartInfo chartInfo);
static void UpdateDataLogStateDisplay(DAQ_ChartInfo chartInfo, double time);
static int  SetDataLogHeaderChannelString(DAQ_ChartInfo chartInfo);
static int  RevertDataLogSettings(DAQ_ChartInfo chartInfo);
static int  AdvancedLoggingOptionsDlg(DAQ_ChartInfo chartInfo);
static int  ReviseColumnHeaders(DAQ_ChartInfo chartInfo, char **columnHeaders);
static int  GetDefaultHeadings(DAQ_ChartInfo chartInfo, char **defaultHeadings);
static int  FillAlarmLogActionListBox(DAQ_ChartInfo chartInfo, int panel, int listBox);
static int  ApplyAlarmLogActionSelections(DAQ_ChartInfo chartInfo, int panel, int listBox);
static void HandleAlarmControlledConditionalLogging(DAQ_ChartInfo chartInfo);
static int  GetCachedScan(DAQ_ChartInfo chartInfo, double *scan, double *time, int *scanReady);
static int  FillReadCache(DAQ_ChartInfo chartInfo);
static void FlushReadCache(DAQ_ChartInfo chartInfo);

/*****************************************************************************************/
#pragma warning(disable: 4761)
static int CreateDAQ_ChartInfo(int panel, int ctrl, DAQ_ChartInfo *newChartInfo)
{
    int             error = 0;
    DAQ_ChartInfo chartInfo = 0;

    nullChk( chartInfo = (DAQ_ChartInfo)calloc(sizeof(*chartInfo), 1));

    chartInfo->setAttribute = Internal_DAQ_Chart_SetAttribute;
    chartInfo->getAttribute = DAQ_Chart_GetAttribute;
    chartInfo->buildDialog = (BuildDialogFunc)BuildDaqChartPopup;
    chartInfo->saveSettings = (SaveSettingsFunc)SaveSettings;
    chartInfo->recallSettings = (RecallSettingsFunc)RecallSettings;
    chartInfo->updateDialogDimming = (UpdateDialogDimmingFunc)UpdateDialogDimming;
    chartInfo->applyDialogSettings = (ApplyDialogSettingsFunc)ApplyDialogSettings;
    chartInfo->revertAttributes = (RevertAttributesFunc)RevertAttributes;
    chartInfo->errorNotifyFunc = (ErrorNotifyFunc)HandleErrorNotification;
    chartInfo->getErrorString = (GetErrorStringFunc)GetDAQChartErrorString;
    chartInfo->confirmRevertFunc = (ConfirmRevertFunc)ConfirmRevert;
    chartInfo->rememberAttributes = (RememberAttributesFunc)RememberAttributes;
    chartInfo->copyAttributesToDialog = (CopyAttributesToDialogFunc)CopyAttributesToDialog;
    chartInfo->settingsPrefix = DAQ_CHART_SETTINGS_PREFIX;
    chartInfo->ctrlTypeName = DAQ_CHART_TYPE_NAME;
    chartInfo->version = DAQ_CHART_VERSION;
    chartInfo->baseCtrl = ctrl;
    chartInfo->panel = panel;

    errChk( InitDaqCtrl ((DaqCtrl)chartInfo));

    errChk( chartInfo->refreshTimerCtrl = NewCtrl(panel, CTRL_TIMER, 0, -100, -100));
    errChk( InstallCtrlCallback(panel, chartInfo->refreshTimerCtrl, DAQ_ChartRefreshTimerCallback, chartInfo));

    errChk( chartInfo->scanTimerCtrl = NewCtrl(panel, CTRL_TIMER, 0, -100, -100));
    errChk( InstallCtrlCallback(panel, chartInfo->scanTimerCtrl, DAQ_ChartScanTimerCallback, chartInfo));

    nullChk( chartInfo->alarmList = ListCreate(sizeof(Alarm)));

    errChk( ChainCtrlCallback(panel, ctrl,  DAQ_ChartCallback, (void *)chartInfo, DAQ_CHART_TYPE));

    errChk( CreateHistory(chartInfo, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE_DEFAULT, DAQ_CHART_DEFAULT_NUMBER_OF_CHANNELS));

    nullChk( chartInfo->dataLog = DataLog_Create(HandleDatalogNotification, (void *)chartInfo));

    if (newChartInfo)
        *newChartInfo = chartInfo;

Error:
    if (error < 0)
        {
        if (newChartInfo)
            *newChartInfo = 0;
        return error;
        }
    else
        return 0;
}

/*****************************************************************************************/

static void DiscardDAQ_ChartInfo(DAQ_ChartInfo chartInfo)
{
    if (chartInfo)
        {
        UnchainCtrlCallback(chartInfo->panel, chartInfo->baseCtrl, DAQ_CHART_TYPE);
        Safe_AIClearAcquisition(chartInfo);

        DiscardCtrlNoBOLE(chartInfo->panel, chartInfo->refreshTimerCtrl, TRUE);
        DiscardCtrlNoBOLE(chartInfo->panel, chartInfo->scanTimerCtrl, TRUE);

        DataLog_Dispose(chartInfo->dataLog);
        DisposeHistory(chartInfo);

        ListDispose(chartInfo->alarmTabCtrls);
        ListDispose(chartInfo->dataLogTabCtrls);
        ListDispose(chartInfo->savedTraceColorList);

        DisposeSavedAlarmList(chartInfo);           /* disposes list as well */
        DisposeAllChartAlarms(chartInfo);           /* doesn't dispose list */
        ListDispose(chartInfo->alarmList);

        free(chartInfo->scanCache);
        free(chartInfo->savedColumnHeadings);
        free(chartInfo->savedComment);

            /*  Save these resources for last because other dispose functions need them */
        DaqCtrl_DisposeDaqCtrl((DaqCtrl)chartInfo);
        }
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_Create(int panel, char *label, int top, int left, char *configurationName,  CtrlCallbackPtr callback, void *callbackData)
{
    return Internal_DAQ_Chart_Create(panel, label, top, left, configurationName,
                            callback, callbackData);
}

static int Internal_DAQ_Chart_Create(int panel, char *label, int top, int left, char *configurationName,  CtrlCallbackPtr callback, void *callbackData)
{
    int baseCtrl = 0;
    int error = 0;

    errChk( baseCtrl = NewCtrl (panel, CTRL_STRIP_CHART, label, top, left));
    errChk( InstallCtrlCallback(panel, baseCtrl, callback, callbackData));

#ifdef INCLUDE_IO_DIRECTION
    errChk(SetCtrlModeToIODirection(panel, baseCtrl, ioDirection));
#endif

    return DAQ_Chart_ConvertFromStripChart(panel, baseCtrl, configurationName);

Error:
    if (baseCtrl > 0)
        DiscardCtrlNoBOLE(panel, baseCtrl, TRUE);
    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_ConvertFromStripChart(int panel, int baseCtrl, char *configurationName)
{
    int             error = 0;
    int             style;
    int             ioDirection;
    int             ctrlMode;
    DAQ_ChartInfo   chartInfo = 0;

    errChk( GetCtrlAttribute(panel, baseCtrl, ATTR_CTRL_STYLE, &style));

    switch (style)
        {
        case CTRL_STRIP_CHART:
            break;
        default:
            errChk( UIEInvalidControlType);
        }

    errChk( CreateDAQ_ChartInfo(panel, baseCtrl, &chartInfo));

    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_DEVICE, FindDefaultDevice(ANALOG_INPUT, NULL)));

    error = Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_CHANNEL_STRING, ATTR_DAQ_CHART_CHANNEL_STRING_DEFAULT);  /* don't errChk() this, it can fail due to poorly configured DAQ */
    if (error < 0)
        {
        /*  make sure we have a default, because setting the channel string attribute can fail for chart because of DAQ config problems (no memory check needed, NULL is not disastrous)  */
        chartInfo->channelString = StrDup(ATTR_DAQ_CHART_CHANNEL_STRING_DEFAULT);
        chartInfo->numChannels = 1;
        error = 0;
        }

    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_REFRESH_FREQUENCY, ATTR_DAQ_CHART_REFRESH_FREQUENCY_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_SCAN_FREQUENCY, ATTR_DAQ_CHART_SCAN_FREQUENCY_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_HIGH_LIMIT, ATTR_DAQ_CHART_UPDATE_HIGH_LIMIT_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_LOW_LIMIT, ATTR_DAQ_CHART_UPDATE_LOW_LIMIT_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_ENABLED, ATTR_DAQ_CHART_ENABLED_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_VISIBLE, ATTR_DAQ_CHART_VISIBLE_DEFAULT));

    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_ALLOW_RCLICK_EDIT, ATTR_DAQ_CHART_ALLOW_RCLICK_EDIT_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_EDIT_DIALOG_TITLE, ATTR_DAQ_CHART_EDIT_DIALOG_TITLE_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_EDIT_DIALOG_DISPLAYED, ATTR_DAQ_CHART_EDIT_DIALOG_DISPLAYED_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_EDIT_DIALOG_MODAL, ATTR_DAQ_CHART_EDIT_DIALOG_MODAL_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_ERROR_NOTIFICATION, ATTR_DAQ_CHART_ERROR_NOTIFICATION_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_WARN_ON_HW_TIMING_RESTARTS, ATTR_DAQ_CHART_WARN_ON_HW_TIMING_RESTARTS_DEFAULT));

#if ATTR_DAQ_CHART_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT
        {
        int reasonable;

        errChk( reasonable = DAQ_ChartHasReasonableRangeForSlavingToLimits(panel, baseCtrl));
        if (reasonable)
            errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_TIE_CTRL_RANGE_TO_LIMITS, TRUE));
        }
#endif

    errChk( GetCtrlAttribute(panel, baseCtrl, ATTR_CTRL_MODE, &ctrlMode));
    if (ctrlMode != VAL_INDICATOR)
        ioDirection = ANALOG_OUTPUT;
    else
        ioDirection = ANALOG_INPUT;
    errChk( Internal_DAQ_Chart_SetAttribute(panel, baseCtrl, ATTR_DAQ_CHART_IO_DIRECTION, ioDirection));

    if (configurationName && *configurationName)
        Internal_DAQ_Chart_RecallConfig(panel, baseCtrl, configurationName); /* don't report config recall errors when creating controls, they can call DAQ_Chart_RecallConfig() if it matters */

    return baseCtrl;

Error:
    DiscardDAQ_ChartInfo(chartInfo);
    return error;
}

/*****************************************************************************************/

    /* returns TRUE, FALSE, or negative error code */
#if ATTR_DAQ_CHART_TIE_CTRL_RANGE_TO_LIMITS_DEFAULT
static int  DAQ_ChartHasReasonableRangeForSlavingToLimits(int panel, int daqChart)
{
    double  max, min;
    int     error = 0;

    errChk( GetAxisRange (panel, daqChart, 0, 0, 0, 0, &min, &max));

    if (max <= TIE_CTRL_RANGE_TO_LIMITS_MAX && min >= TIE_CTRL_RANGE_TO_LIMITS_MIN)
        return TRUE;
    else
        return FALSE;

Error:
    return error;
}
#endif

/*****************************************************************************************/

static void DisposeHistory(DAQ_ChartInfo chartInfo)
{
    free(chartInfo->dataHistory);
    free(chartInfo->timeHistory);

    chartInfo->timeHistory = 0;
    chartInfo->dataHistory = 0;
    chartInfo->historyBufferSize = 0;
}

/*****************************************************************************************/

static int CreateHistory(DAQ_ChartInfo chartInfo, int newSize, int numChannels)
{
    int     error = 0;
    double  *timeHistory = 0;
    double  *dataHistory = 0;

    Assert(newSize > 0);

    if (numChannels > 0)
        {
        nullChk( timeHistory = malloc(sizeof(double) * newSize));
        nullChk( dataHistory = malloc(sizeof(double) * numChannels * newSize));
        }

    DisposeHistory(chartInfo);

    chartInfo->historyLength = 0;
    chartInfo->nextPlotIndex = 0;
    chartInfo->nextScanIndex = 0;
    chartInfo->historyBufferSize = newSize;
    chartInfo->timeHistory = timeHistory;
    timeHistory = 0;
    chartInfo->dataHistory = dataHistory;
    dataHistory = 0;

Error:
    if (error < 0)
        {
        free(timeHistory);
        free(dataHistory);
        }

    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_GetHistoryCopy(int panel, int daqChart, int *numPoints, int *numScans,
        int *channelsPerScan, int groupDataByChannels, double **samples,
        double **sampleTimes, int remove)
{
    return Internal_DAQ_Chart_GetHistoryCopy(panel, daqChart, numPoints, numScans,
                                             channelsPerScan, groupDataByChannels,
                                             samples, sampleTimes, remove);
}

static int Internal_DAQ_Chart_GetHistoryCopy(int panel, int daqChart, int *numPoints, int *numScans,
        int *channelsPerScan, int groupDataByChannels, double **samples,
        double **sampleTimes, int remove)
{
    int     error = 0;
    int     numChannels;
    int     length;
    double  *samplesCopy = 0;
    double  *sampleTimesCopy = 0;

    errChk( DAQ_Chart_GetAttribute(panel, daqChart, ATTR_DAQ_CHART_HISTORY_LENGTH, &length));
    errChk( DAQ_Chart_GetAttribute(panel, daqChart, ATTR_DAQ_CHART_NUM_CHANNELS, &numChannels));

    if (length > 0 && numChannels > 0)
        {
        if (samples)
            nullChk( samplesCopy = (double *)malloc(length * numChannels * sizeof(double)));
        if (sampleTimes)
            nullChk( sampleTimesCopy = (double *)malloc(length * sizeof(double)));

        errChk( Internal_DAQ_Chart_GetHistory(panel, daqChart, numPoints, numScans, channelsPerScan,
            groupDataByChannels, (samples ? samplesCopy : 0), (sampleTimes ? sampleTimesCopy : 0), remove));
         }
    else
        {
        if (numPoints)
            *numPoints = 0;

        if (numScans)
            *numScans = 0;

        if (channelsPerScan)
            *channelsPerScan = 0;
        }
Error:
    if (error < 0)
        {
        free(samplesCopy);
        free(sampleTimesCopy);
        samplesCopy = 0;
        sampleTimesCopy = 0;
        }

    if (samples)
        *samples = samplesCopy;
    if (sampleTimes)
        *sampleTimes = sampleTimesCopy;

    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_GetHistory(int panel, int daqChart, int *numPoints, int *numScans, int *channelsPerScan,
                    int groupDataByChannels, double samples[], double sampleTimes[], int remove)
{
    return Internal_DAQ_Chart_GetHistory(panel, daqChart, numPoints, numScans,
                                         channelsPerScan, groupDataByChannels,
                                         samples, sampleTimes, remove);
}

static int Internal_DAQ_Chart_GetHistory(int panel, int daqChart, int *numPoints, int *numScans, int *channelsPerScan,
                    int groupDataByChannels, double samples[], double sampleTimes[], int remove)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = 0;
    int             nextScanIndex = 0;
    int             chartHistoryFull;
    int             numScansBeforeWrap;
    int             numScansAfterWrap;

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

    if (numPoints)
        *numPoints = chartInfo->historyLength * chartInfo->numChannels;

    if (channelsPerScan)
        *channelsPerScan = chartInfo->numChannels;

    if (numScans)
        *numScans = chartInfo->historyLength;

    if (chartInfo->historyBufferSize > 0 && chartInfo->historyLength > 0 && chartInfo->numChannels > 0)
        {
        nextScanIndex = chartInfo->nextScanIndex;
        chartHistoryFull = chartInfo->historyLength == chartInfo->historyBufferSize;

        if (chartHistoryFull)
            {
            if (nextScanIndex > 0)
                {
                numScansBeforeWrap = chartInfo->historyBufferSize - nextScanIndex;
                numScansAfterWrap = nextScanIndex;
                }
            else
                {
                numScansBeforeWrap = 0;
                numScansAfterWrap = chartInfo->historyBufferSize;
                }
            }
        else
            {
            numScansBeforeWrap = 0;
            numScansAfterWrap = nextScanIndex;
            }

        if (samples)
            {
            if (numScansBeforeWrap > 0)  /* move points stored before we wrapped around the array */
                memmove(samples, &chartInfo->dataHistory[nextScanIndex * chartInfo->numChannels],
                    numScansBeforeWrap * chartInfo->numChannels * sizeof(double));

            memmove(&samples[numScansBeforeWrap * chartInfo->numChannels], chartInfo->dataHistory, numScansAfterWrap * chartInfo->numChannels * sizeof(double));

            if (groupDataByChannels)
                TransposeData(samples, VAL_DOUBLE, chartInfo->historyLength * chartInfo->numChannels, chartInfo->numChannels);
            }

        if (sampleTimes)
            {
            if (numScansBeforeWrap > 0)  /* move points stored before we wrapped around the array */
                memmove(sampleTimes, &chartInfo->timeHistory[nextScanIndex],
                    numScansBeforeWrap * sizeof(double));

            memmove(&sampleTimes[numScansBeforeWrap], chartInfo->timeHistory, numScansAfterWrap * sizeof(double));
            }
        }

    if (remove) /* empty the history buffer */
        {
        errChk( PlotPendingScans(chartInfo));   /* don't delete the data before plotting it */
        chartInfo->historyLength = 0;
        chartInfo->nextPlotIndex = 0;
        chartInfo->nextScanIndex = 0;
        }

Error:
    return error;
}

/*****************************************************************************************/

static int UpdateNumTraces(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     index;
    int     numAlarms;
    Alarm   *alarm;
    int     numTraces;
    int     traceNumber;
    int     channelIndex;

    numAlarms = ListNumItems(chartInfo->alarmList);
    chartInfo->numAlarmsToPlot = 0;
    chartInfo->numAlarmLimitsToPlot = 0;

    for (index = 1; index <= numAlarms; index++)
        {
        alarm = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);
        if (alarm->plot)
            {
            chartInfo->numAlarmsToPlot++;
            if (IsRangeMode(alarm->mode))
                chartInfo->numAlarmLimitsToPlot += 2;
            else
                chartInfo->numAlarmLimitsToPlot++;
            }
        }

    numTraces = chartInfo->numChannels + chartInfo->numAlarmLimitsToPlot;

    if (numTraces > 0)
        errChk( SetCtrlAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_NUM_TRACES, numTraces));

    traceNumber = chartInfo->numChannels + 1;
    for (index = 1; index <= numAlarms; index++)
        {
        alarm = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);

        channelIndex = GetIndexFromChannel(chartInfo, alarm->channel);

        alarm->valid = channelIndex >= 0;
        alarm->channelIndex = channelIndex;

        if (alarm->plot)
            {
            alarm->traceIndex = traceNumber;
            errChk( SetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, traceNumber, ATTR_TRACE_COLOR, alarm->color));
            traceNumber++;
            if (IsRangeMode(alarm->mode))
                {
                errChk( SetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, traceNumber, ATTR_TRACE_COLOR, alarm->color));
                traceNumber++;
                }
            }
        }

    errChk( UpdateAlarmDlgCtrls(chartInfo));

Error:
    return error;
}

/*****************************************************************************************/

static void WarnIfAlarmsHaveInvalidChannels(DAQ_ChartInfo chartInfo)
{
    int     numAlarms;
    int     index;
    Alarm   *alarmPtr;
    int     warn = FALSE;

        /*  must suppress when recalling or reverting settings because we can't simultaneously recall the alarms and the channel string */
    if (!chartInfo->recallingSettings && !chartInfo->revertingSettings)
        {
        numAlarms = ListNumItems(chartInfo->alarmList);

        for (index = 1; index <= numAlarms && !warn; index++)
            {
            alarmPtr = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);

            if (!alarmPtr->valid)
                warn = TRUE;
            }

        if (warn)
            DaqCtrl_DisplayError((DaqCtrl)chartInfo, DAQ_CHART_WARNING_INVALID_CHANNEL, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);
        }
}

/*****************************************************************************************/

static void CancelLoggingDueToAttributeChange(DAQ_ChartInfo chartInfo)
{
    int isLogging = FALSE;

    DAQ_Chart_GetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOGGING_ON, &isLogging);

    if (isLogging)
        {
        Internal_DAQ_Chart_SetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOGGING_ON, FALSE);
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, DAQ_CHART_WARNING_DATALOGGING_TERMINATED, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);
        }
}

/*****************************************************************************************/

static int SetDataLogHeaderChannelString(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    char    *buf = 0;
    int     channelStringLen;
    char    *channelString;

    if (chartInfo->channelString)
        channelString = chartInfo->channelString;
    else
        channelString = "";

    channelStringLen = strlen(channelString);
    nullChk( buf = malloc(channelStringLen + 50));

    sprintf(buf, "CHANNEL STRING: \"%s\"", channelString);
    errChk( DataLog_SetAttribute(chartInfo->dataLog, ATTR_DATALOG_EXTRA_HEADER_INFO, buf));

Error:
    free(buf);
    return error;
}

/*****************************************************************************************/

static int SetDaqChartChannelString(DAQ_ChartInfo chartInfo, char *newValue)
{
    int             error = 0, channelError = 0;
    char            *newChannelString;
    int             valueLength;
    unsigned long   groupSize = 0;
    void            *newCache = 0;

    if (!newValue)
        errChk( UIENullPointerPassed);

    if (*newValue)
        channelError = (error = GetNumChannelsNoBOLE (chartInfo->device, newValue, chartInfo->ioDirection, &groupSize));

    /* Added check for noDriverErr, johnm - 5/30/96 */
    if ((error == unknownDeviceErr) || (error == noDriverErr)) /* don't let a bad device selection cause a bad channel string error */
        errChk( error);
    /*******************************************/
    
        /*  don't complain about string syntax in dialog as the user types in the 
            channel string */

    if (groupSize + ListNumItems(chartInfo->alarmList) > 64)       /* check for too many strip chart traces */
        {
        if (chartInfo->dialog && chartInfo->dialogDisplayed && chartInfo->updatingCtrlFromDialog && chartInfo->dialogEvent == EVENT_COMMIT)
            SetCtrlVal(chartInfo->dialog, chartInfo->channelStringCtrl, chartInfo->channelString);  /* reset channel string control, the entered value has been rejected */
        errChk( DAQ_CHART_ERROR_MAX_64_CHANNELS_AND_ALARMS);
        }

    if (groupSize <= 0 && chartInfo->dialog && chartInfo->dialogDisplayed && chartInfo->updatingCtrlFromDialog && chartInfo->dialogEvent != EVENT_COMMIT)
        {
        error = 0;
        goto Error; /* channel string is bad, but its not finished being typed, so give the user another chance before complaining */
        }

    if (!chartInfo->channelString || strcmp(newValue, chartInfo->channelString) != 0)
        {
        valueLength = Min(strlen(newValue), MAX_CHANNEL_STRING_LEN);

        nullChk( newChannelString = malloc(valueLength + 1));
        StringCopyMax(newChannelString, newValue, MAX_CHANNEL_STRING_LEN + 1);

        free(chartInfo->channelString);
        chartInfo->channelString = newChannelString;
        chartInfo->numChannels = groupSize;

        chartInfo->scanSize = groupSize * sizeof(double);

        chartInfo->readCacheIndex = 0;
        chartInfo->readCacheNumScans = 0;
        if (groupSize > 0)
            nullChk( newCache = malloc(ASYNC_READ_CACHE_SIZE * sizeof(double) * groupSize));
        else
            newCache = 0;
        free(chartInfo->scanCache);
        chartInfo->scanCache = newCache;

        CancelLoggingDueToAttributeChange(chartInfo);
        errChk( DataLog_SetAttribute(chartInfo->dataLog, ATTR_DATALOG_NUM_CHANNELS, groupSize));

            /* recreate the chart history with the appropriate number of channels */
        errChk( CreateHistory(chartInfo, chartInfo->historyBufferSize, chartInfo->numChannels));

        errChk( UpdateNumTraces(chartInfo));

        errChk( channelError);  /* added by johnm 6/11/96, check for channel error once the user has commited */
        errChk( UpdateChartHWTimingMode(chartInfo));

        if (!*newValue)
            DaqCtrl_DisplayError((DaqCtrl)chartInfo, 0, 0, 0);      /* clear a displayed error if the channel string is empty */

        WarnIfAlarmsHaveInvalidChannels(chartInfo);
        errChk( SetDataLogHeaderChannelString(chartInfo));
        }

Error:
    return error;
}

/*****************************************************************************************/

static int  Safe_AIClearAcquisition(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int savedBOLE = SetBOLE (FALSE);    /* disable break on library errors */

    if (chartInfo->asyncTaskId)
        errChk( AIClearAcquisition (chartInfo->asyncTaskId));

Error:
    chartInfo->asyncTaskId = 0;
    if (error == -10608)
        error = 0;

    SetBOLE(savedBOLE);

    return error;
}

/*****************************************************************************************/

    /*  returns a easyIO error, or 0 */
static int ConfigureChartHWTimingMode(DAQ_ChartInfo chartInfo)
{
    int             error = 0;
    unsigned long   taskId = 0;
    double          actualFrequency;
    int             hwTimedAcqBufferSize;
    int             savedBOLE = SetBOLE (FALSE);    /* disable break on library errors */

    errChk( Safe_AIClearAcquisition (chartInfo));
    FlushReadCache(chartInfo);

    if (chartInfo->enabled)
        {
        hwTimedAcqBufferSize = (int)Max(MIN_HW_TIMED_BUFFER_SIZE, HW_TIMED_SCAN_BUF_LENGTH_IN_SECONDS * chartInfo->scanFrequency);
        hwTimedAcqBufferSize = (hwTimedAcqBufferSize + 1) & ~0x1;       /* must be even !!??! */

        error =  AIStartAcquisition (chartInfo->device, chartInfo->channelString, hwTimedAcqBufferSize,
               chartInfo->scanFrequency, chartInfo->highLimit, chartInfo->lowLimit,
               &actualFrequency, &taskId);

            /*  This is to allow you to set the mode to hardware timing even if it is currently not an option
                because a transfer is in progress.  If the blocking transfer terminates, then the chart
                will start it acquisition.
            */
        if (error == -10609)
            {
            error = -error;     /* transfer in progress, make it a warning so we can still set the mode to hardware timing */
            DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_DAQ_ERROR_DISPLAY_TIME_TILL_CLEARED);
            actualFrequency = chartInfo->scanFrequency;
            taskId = 0;
            }

        errChk( error);

        chartInfo->hwTimedAcqStartTime = Timer();
        chartInfo->hwTimedScanNum = 0;

        chartInfo->scanFrequency = actualFrequency;
        chartInfo->asyncTaskId = taskId;
        chartInfo->hwTimedAcqBufferSize = hwTimedAcqBufferSize;

        if (chartInfo->dialog)
            errChk( SetCtrlVal(chartInfo->dialog, chartInfo->actualFrequencyNumeric, actualFrequency));
        }

Error:
    SetBOLE(savedBOLE);
    return error;
}

/*****************************************************************************************/

static int UpdateChartHWTimingMode(DAQ_ChartInfo chartInfo)
{
    int error = 0;

    if (chartInfo->useHardwareTiming)
        errChk( ConfigureChartHWTimingMode(chartInfo));

Error:
    return error;
}

/*****************************************************************************************/

static int SetChartHWTimingMode(DAQ_ChartInfo chartInfo, int useHardwareTiming)
{
    int error = 0;

    if (!chartInfo->useHardwareTiming!= !useHardwareTiming)
        {
        if (useHardwareTiming)
            {
            errChk( ConfigureChartHWTimingMode(chartInfo));
            chartInfo->useHardwareTiming = TRUE;
            }
        else
            {
            chartInfo->useHardwareTiming = FALSE;
            errChk( Safe_AIClearAcquisition (chartInfo));
            }
        }

Error:
    return error;
}

/*****************************************************************************************/

static void IncrementScanPosition(DAQ_ChartInfo chartInfo)
{
    chartInfo->nextScanIndex = (chartInfo->nextScanIndex + 1) % chartInfo->historyBufferSize;
    if (chartInfo->historyLength < chartInfo->historyBufferSize)
        chartInfo->historyLength++;
}

/*****************************************************************************************/

static int AcquireScans(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     daqError = 0;
    double  *time;
    double  *scanPtr;
    int     scanReady;
    double  startTime;
    double  currTime;
    int     moreScansLeft;

    startTime = Timer();
    do  {
        scanPtr = &chartInfo->dataHistory[chartInfo->nextScanIndex * chartInfo->numChannels];
        time = &chartInfo->timeHistory[chartInfo->nextScanIndex];

        daqError = GetSingleScan(chartInfo, scanPtr, time, &scanReady, &moreScansLeft);

        if (daqError != 0)
            {
            DaqCtrl_DisplayError((DaqCtrl)chartInfo, daqError, MIN_ERROR_DISPLAY_TIME, MIN_DAQ_ERROR_DISPLAY_TIME_TILL_CLEARED);
            CallCtrlCallback(chartInfo->panel, chartInfo->baseCtrl, EVENT_DAQ_CHART_DATA_READY, (int)scanPtr, daqError, 0);\
            }

        if (scanReady)
            {
            DaqCtrl_DisplayError((DaqCtrl)chartInfo, 0, 0, 0);   /* success, clear any displayed error */

            IncrementScanPosition(chartInfo);

            CallCtrlCallback(chartInfo->panel, chartInfo->baseCtrl, EVENT_DAQ_CHART_DATA_READY, (int)scanPtr, daqError, 0);

            CheckAlarms(chartInfo, scanPtr, time);

            errChk( DataLog_LogData(chartInfo->dataLog, time, scanPtr, 1));

               /* make sure data doesn't scroll out of history buffer before it is plotted */
            if (NumPendingScansToPlot(chartInfo) >=( chartInfo->historyBufferSize - 1))
                {error = (PlotPendingScans(chartInfo)); if (error < 0) {goto Error;}}
            }

        currTime = Timer();
        } while (moreScansLeft && daqError == 0 && currTime - startTime < SCAN_CATCHUP_TIME);

        UpdateScanBackLogDisplay(chartInfo, currTime);
        UpdateActualFrequencyDisplay(chartInfo, currTime);
        UpdateDataLogStateDisplay(chartInfo, currTime);

Error:
    return error;
}

/*****************************************************************************************/

static void UpdateScanBackLogDisplay(DAQ_ChartInfo chartInfo, double time)
{
    if (chartInfo->dialog)
        {
        chartInfo->numScansLeft = 0;
        if (chartInfo->useHardwareTiming)
            {
            int savedBole = SetBOLE(FALSE);
            AICheckAcquisition (chartInfo->asyncTaskId, &chartInfo->numScansLeft);
            SetBOLE(savedBole);
            }

        if (chartInfo->hwTimedAcqIsFallingBehind && chartInfo->numScansLeft <= (unsigned)(chartInfo->hwTimedAcqBufferSize / 4))
            {
            chartInfo->hwTimedAcqIsFallingBehind = FALSE;   /* if hardware timed buffer is 1/4 full or less */
            SetCtrlAttribute(chartInfo->dialog, chartInfo->scanBackLogNumeric, ATTR_TEXT_COLOR, VAL_BLACK);
            }
        else
        if (!chartInfo->hwTimedAcqIsFallingBehind && chartInfo->numScansLeft > (unsigned)(chartInfo->hwTimedAcqBufferSize / 2)) /* if hardware timed buffer is more than 1/2 full */
            {
            chartInfo->hwTimedAcqIsFallingBehind = TRUE;
            SetCtrlAttribute(chartInfo->dialog, chartInfo->scanBackLogNumeric, ATTR_TEXT_COLOR, VAL_RED);
            }

        if (chartInfo->dialog && chartInfo->dialogDisplayed && time > chartInfo->nextScanBackLogUpdateTime)
            {
            SetCtrlVal(chartInfo->dialog, chartInfo->scanBackLogNumeric, chartInfo->numScansLeft);
            chartInfo->nextScanBackLogUpdateTime = time + SCAN_BACKLOG_UPDATE_INTERVAL;
            }
        }
}

/*****************************************************************************************/

static void UpdateActualFrequencyDisplay(DAQ_ChartInfo chartInfo, double time)
{
    double  elapsedTime;
    double  interval = DAQ_CTRL_MIN_ACTUAL_FREQ_UPDATE_INTERVAL;

    if (!chartInfo->useHardwareTiming)
        {
        if (chartInfo->scanFrequency > 0.0)
            interval += 1/chartInfo->scanFrequency;     /* must get at least one cycle in the measurement */

        if (time > interval + chartInfo->lastActualFreqDisplayTime)
            {
            if (chartInfo->lastActualFreqDisplayTime > 0.0)     /* if this isn't the first time through */
                {
                elapsedTime = time - chartInfo->lastActualFreqDisplayTime;

                if (elapsedTime > 0.0 && chartInfo->dialog && chartInfo->dialogDisplayed)
                    SetCtrlVal(chartInfo->dialog, chartInfo->actualFrequencyNumeric, chartInfo->numScansSinceLastActualFreqDisplayTime / elapsedTime);
                }

            chartInfo->lastActualFreqDisplayTime = time;
            chartInfo->numScansSinceLastActualFreqDisplayTime = 0;
            }
        }
}

/*****************************************************************************************/

static void UpdateDataLogStateDisplay(DAQ_ChartInfo chartInfo, double time)
{
    if (time > MIN_DATALOG_PROGRESS_UPDATE_INTERVAL + chartInfo->lastDatalogStateDisplayTime)
        {
        int isActive = FALSE;

        chartInfo->lastDatalogStateDisplayTime = time;

        DAQ_Chart_GetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOGGING_ON, &isActive);
        if (isActive)
            UpdateLoggingStateIndicator(chartInfo);
        }
}

/*****************************************************************************************/

static int NumPendingScansToPlot(DAQ_ChartInfo chartInfo)
{
    if (chartInfo->nextScanIndex >= chartInfo->nextPlotIndex)
        return chartInfo->nextScanIndex - chartInfo->nextPlotIndex;
    else
        return chartInfo->historyBufferSize - chartInfo->nextPlotIndex + chartInfo->nextScanIndex;
}

/*****************************************************************************************/

static int PlotPendingScans(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     numScansToPlot;
    int     numValuesPerScanPlot;
    double  *plotBuffer = 0;
    int     index;
    int     plotIndex;
    int     numChannels;
    int     numAlarmsToPlot;
    int     numAlarmLimitsToPlot;

    numChannels = chartInfo->numChannels;
    numAlarmsToPlot = chartInfo->numAlarmsToPlot;
    numAlarmLimitsToPlot = chartInfo->numAlarmLimitsToPlot;

    numScansToPlot = NumPendingScansToPlot(chartInfo);

    numValuesPerScanPlot = numAlarmLimitsToPlot + numChannels;

    if (numValuesPerScanPlot > 0)
        {
        nullChk( plotBuffer = (double *)malloc(numScansToPlot * numValuesPerScanPlot * sizeof(double)));

        plotIndex = chartInfo->nextPlotIndex;
        for (index = 0; index < numScansToPlot; index++)
            {
            memmove(plotBuffer + index * numValuesPerScanPlot, &chartInfo->dataHistory[plotIndex * numChannels], numChannels * sizeof(double));

            if (chartInfo->numAlarmsToPlot > 0)
                {
                if (index == 0)
                    GetPlottedAlarmLimits(chartInfo, plotBuffer + numChannels);
                else     /* copy from first iteration since values will be the same */
                    memmove(plotBuffer + index * numValuesPerScanPlot + numChannels, plotBuffer + numChannels, numAlarmLimitsToPlot * sizeof(double));
                }

            plotIndex = (plotIndex + 1) % chartInfo->historyBufferSize;
            }

        chartInfo->nextPlotIndex = plotIndex;

        errChk( PlotStripChart (chartInfo->panel, chartInfo->baseCtrl, plotBuffer,
                                numScansToPlot * numValuesPerScanPlot, 0, 0, VAL_DOUBLE));
        }

Error:
    free(plotBuffer);
    return 0;
}

/*****************************************************************************************/

static void GetPlottedAlarmLimits(DAQ_ChartInfo chartInfo, double *dest)
{
    int     numAlarms;
    int     index;
    Alarm   *alarm;

    numAlarms = ListNumItems(chartInfo->alarmList);
    for (index = 1; index <= numAlarms; index++)
        {
        alarm = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);
        if (alarm->plot)
            {
            *dest++ = alarm->limit;
            if (IsRangeMode(alarm->mode))
                *dest++ = alarm->lowerLimit;
            }
        }
}

/*****************************************************************************************/

static void  CVICALLBACK FinishHWTimedTransferRestartWarning(void *callbackData)
{
    DAQ_ChartInfo   chartInfo = (DAQ_ChartInfo)callbackData;

    if (chartInfo->dialog)
        {
        SetCtrlAttribute(chartInfo->dialog, chartInfo->frequencyNumeric, ATTR_LABEL_COLOR, VAL_BLACK);
        SetCtrlAttribute(chartInfo->dialog, chartInfo->useHardwareTimingCheckBox, ATTR_LABEL_COLOR, VAL_BLACK);
        SetCtrlAttribute(chartInfo->dialog, chartInfo->scanBackLogNumeric, ATTR_TEXT_COLOR, VAL_BLACK);
        }
}

/*****************************************************************************************/

static void  DoHWTimedTransferRestartWarning(DAQ_ChartInfo chartInfo, int errorCode)
{
    if (chartInfo->warnOnHWTimedAcqRestarts)
        {
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, errorCode, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

        if (chartInfo->dialog)
            {
            int color;

            GetCtrlAttribute(chartInfo->dialog, chartInfo->useHardwareTimingCheckBox, ATTR_LABEL_COLOR, &color);

            if (color != VAL_RED)
                {
                if (PostDelayedCall (FinishHWTimedTransferRestartWarning, (void *)chartInfo, 3.0))
                    {
                    SetCtrlAttribute(chartInfo->dialog, chartInfo->scanBackLogNumeric, ATTR_TEXT_COLOR, VAL_RED);
                    SetCtrlAttribute(chartInfo->dialog, chartInfo->useHardwareTimingCheckBox, ATTR_LABEL_COLOR, VAL_RED);
                    if (errorCode == DAQ_CHART_WARNING_HW_TIMED_BUF_OVERRUN)
                        SetCtrlAttribute(chartInfo->dialog, chartInfo->frequencyNumeric, ATTR_LABEL_COLOR, VAL_RED);
                    }
                }
            }
        }

    return;
}

/*****************************************************************************************/

static int HandleHWTimedAcqOverrun(DAQ_ChartInfo chartInfo)
{
    int error = 0;

    CallCtrlCallback(chartInfo->panel, chartInfo->baseCtrl, EVENT_DAQ_CHART_HW_TIMED_SCAN_RESTARTED, VAL_DAQ_CHART_HW_TIMED_SCAN_BUF_OVERRUN, 0, 0);

    DoHWTimedTransferRestartWarning(chartInfo, DAQ_CHART_WARNING_HW_TIMED_BUF_OVERRUN);
    errChk( ConfigureChartHWTimingMode(chartInfo));

Error:
    return error;
}

/*****************************************************************************************/

static int GetCachedScan(DAQ_ChartInfo chartInfo, double *scan, double *time, int *scanReady)
{
    int error = 0;

    if (chartInfo->readCacheNumScans <= 0)
        errChk( FillReadCache(chartInfo));

    if (chartInfo->readCacheNumScans > 0)
        {
        memmove(scan, &chartInfo->scanCache[chartInfo->readCacheIndex * chartInfo->numChannels], chartInfo->scanSize);;
        chartInfo->readCacheIndex++;
        chartInfo->readCacheNumScans--;
        *time = chartInfo->hwTimedAcqStartTime + chartInfo->hwTimedScanNum / chartInfo->scanFrequency;
        chartInfo->hwTimedScanNum++;
        *scanReady = TRUE;
        }

    if (chartInfo->readCacheNumScans <= 0)  /* alway keep something in the cache if possible, so we can look at the cache to see if data is available */
        errChk( FillReadCache(chartInfo));

Error:
    return error;
}

/*****************************************************************************************/

static int FillReadCache(DAQ_ChartInfo chartInfo)
{
    int             error = 0;
    int             numScansToRead;
    unsigned long   numScansInHWBuf;
    unsigned long   backLog;

    if (chartInfo->readCacheNumScans < 1)
        {
        if (chartInfo->asyncTaskId == 0)    /* must have failed to restart previously, try again */
            goto RestartAsync;

        error = AICheckAcquisition (chartInfo->asyncTaskId, &numScansInHWBuf);

        if (error == -10846 || numScansInHWBuf > (unsigned)chartInfo->hwTimedAcqBufferSize)
            goto OverRun;

        if (error == -10608 || error == -10005)    /* transfer must have been aborted by non-async call to same device */
            {
            CallCtrlCallback(chartInfo->panel, chartInfo->baseCtrl, EVENT_DAQ_CHART_HW_TIMED_SCAN_RESTARTED, VAL_DAQ_CHART_HW_TIMED_TRANSFER_ABORTED, 0, 0);

            DoHWTimedTransferRestartWarning(chartInfo, DAQ_CHART_WARNING_HW_TIMED_TRANSFER_ABORTED);
            goto RestartAsync;
            }

        errChk( error);

        if (numScansInHWBuf > 0)
            {
            numScansToRead = Min(numScansInHWBuf, ASYNC_READ_CACHE_SIZE);
            numScansToRead = (int)Min(numScansToRead, MAX_SECONDS_OF_CACHED_SCANS * chartInfo->scanFrequency + 1);  /* +1 means read at least one scan */
            error = AIReadAcquisition (chartInfo->asyncTaskId, numScansToRead,
                                       CONSECUTIVE_MODE, &backLog, GROUP_BY_SCAN,
                                       chartInfo->scanCache);
            if (error == -10846)
                goto OverRun;

            errChk( error);
            chartInfo->readCacheNumScans = numScansToRead;
            chartInfo->readCacheIndex = 0;
            }
        }
Error:
    return error;

OverRun:
    return HandleHWTimedAcqOverrun(chartInfo);

RestartAsync:
    return UpdateChartHWTimingMode(chartInfo);
}

/*****************************************************************************************/

static void FlushReadCache(DAQ_ChartInfo chartInfo)
{
    chartInfo->readCacheNumScans = 0;
    chartInfo->readCacheIndex = 0;
}

/*****************************************************************************************/

static int GetSingleHWTimedScan(DAQ_ChartInfo chartInfo, double *scan, double *time, int *scanReady, int *moreScansLeft)
{
    int error = 0;

    *moreScansLeft = FALSE;
    if (chartInfo->numChannels > 0)
        {
        errChk( GetCachedScan(chartInfo, scan, time, scanReady));
        *moreScansLeft = chartInfo->readCacheNumScans > 0;
        }
    else
        *scanReady = FALSE;

Error:
    return error;
}

/*****************************************************************************************/

    /*  returns a Daq Error code */
static int GetSingleScan(DAQ_ChartInfo chartInfo, double *scan, double *time, int *scanReady, int *moreScansLeft)
{
    int error = 0;
    int savedBOLE = SetBOLE(FALSE);    /* disable break on library errors */

    *scanReady = FALSE;

    if (chartInfo->numChannels > 0)
        {
        if (chartInfo->useHardwareTiming)
            {
            errChk( GetSingleHWTimedScan(chartInfo, scan, time, scanReady, moreScansLeft));
            }
        else
            {
            errChk( AISampleChannels (chartInfo->device, chartInfo->channelString, chartInfo->highLimit, chartInfo->lowLimit, scan));
            *time = Timer();
            *scanReady = TRUE;
            *moreScansLeft = FALSE;
            chartInfo->numScansSinceLastActualFreqDisplayTime++;
            }
        }

Error:
    SetBOLE(savedBOLE);

    return error;
}

/*****************************************************************************************/

static void CVIFUNC HandleErrorNotification(DAQ_ChartInfo chartInfo, int errorCode, char *errorMsg)
{
    char    errorBuf[512];

    sprintf(errorBuf, "Error/Warning %d: %s", errorCode, errorMsg);

    DataLog_LogMessage(chartInfo->dataLog, errorBuf, 0);
}

/*****************************************************************************************/

static int CVIFUNC ConfirmRevert(DAQ_ChartInfo chartInfo, RevertReply *reply)
{
    int     error = 0;
    int     simpleRevertPossible;
    int     channelStringChanged;
    int     fileFormatChanged;
    int     logFileNameChanged;
    int     fileFormat;
    char    *logFile = 0;
    int     choice;
    int     loggingIsOn;

    *reply = DONT_REVERT;
    channelStringChanged = strcmp(chartInfo->channelString, chartInfo->savedChannelString) != 0;

    errChk( DataLog_GetAttribute(chartInfo->dataLog, ATTR_DATALOG_LOG_FILE_FORMAT, &fileFormat));
    fileFormatChanged = chartInfo->savedFileFormat != fileFormat;

    errChk( DataLog_GetAttribute(chartInfo->dataLog, ATTR_DATALOG_DATA_LOG_FILE_COPY, &logFile));
    logFileNameChanged = strcmp(chartInfo->savedDataLogFilePath, logFile) != 0;

    errChk( DataLog_GetAttribute(chartInfo->dataLog, ATTR_DATALOG_LOGGING_ON, &loggingIsOn));

    simpleRevertPossible =  !loggingIsOn || (!channelStringChanged && !fileFormatChanged && !logFileNameChanged);

    if (simpleRevertPossible)
        {
        errChk( choice = GenericMessagePopup ("Revert Configuration Changes", 
            "Are you sure you want to revert the configuration to the state\n"
            "when the dialog was invoked?\n\n"
            "(Note:  The Data Logging Start/Terminate and Pause/Resume\n"
            "buttons will not be reverted.)\n",
            "__Revert", "__Cancel", 
            0, 0, 0, 0, VAL_GENERIC_POPUP_BTN1, VAL_GENERIC_POPUP_BTN1, VAL_GENERIC_POPUP_BTN2));

        if (!ObjectExists(chartInfo))
            choice = VAL_GENERIC_POPUP_BTN2;

        if (choice == VAL_GENERIC_POPUP_BTN1)
            *reply = ALL_CHANGES_WILL_REVERT;
        }
    else
        {
        char    msg[400];

        sprintf(msg,    "Are you sure you want to revert the configuration to the state\n"
                        "when the dialog was invoked?\n\n"
                        "The following settings cannot be reverted unless data logging is\n"
                        "terminated:\n\n");

        if (channelStringChanged)
            strcat(msg, "Channel String\n");
        if (fileFormatChanged)
            strcat(msg, "Data Log file Format\n");
        if (logFileNameChanged)
            strcat(msg, "Data Log File Path\n");

        errChk( choice = GenericMessagePopup ("Revert Configuration Changes", msg,
                             "__Terminate And Revert",
                             "__Revert Other Settings", "__Cancel", 0, 0, 0,
                             VAL_GENERIC_POPUP_BTN1, VAL_GENERIC_POPUP_BTN1, VAL_GENERIC_POPUP_BTN3));

        if (!ObjectExists(chartInfo))
             choice = VAL_GENERIC_POPUP_BTN3;

        if (choice == VAL_GENERIC_POPUP_BTN1)
            errChk( DataLog_SetAttribute(chartInfo->dataLog, ATTR_DATALOG_LOGGING_ON, FALSE));

        if (choice == VAL_GENERIC_POPUP_BTN1)
            *reply = ALL_CHANGES_WILL_REVERT;
        else
        if (choice == VAL_GENERIC_POPUP_BTN2)
            *reply = SOME_CHANGES_WILL_REVERT;
        }

Error:
        return error;
}

/*****************************************************************************************/

static int CVIFUNC_C Internal_DAQ_Chart_SetAttribute(int panel, int daqChart, int attribute, ...)
{
    int             error = 0;
    va_list         parmInfo;

    va_start(parmInfo, attribute);
    error = DAQ_Chart_SetAttributeFromParmInfo(panel, daqChart, attribute, parmInfo);
    va_end(parmInfo);

    return error;
}

int CVIFUNC_C DAQ_Chart_SetAttribute(int panel, int daqChart, int attribute, ...)
{
    int             error = 0;
    va_list         parmInfo;

    va_start(parmInfo, attribute);
    error = DAQ_Chart_SetAttributeFromParmInfo(panel, daqChart, attribute, parmInfo);
    va_end(parmInfo);

    return error;
}


int CVIFUNC DAQ_Chart_SetAttributeFromParmInfo(int panel, int daqChart, int attribute, va_list parmInfo)
{
    int             error = 0;
    int             newIntValue;
    double          newDoubleValue;
    DAQ_ChartInfo   chartInfo = 0;

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

    switch (attribute)
        {
        case ATTR_DAQ_CHART_DEVICE:
            newIntValue = va_arg(parmInfo, int);
            if (newIntValue != Pin(newIntValue, MIN_DEVICE, MAX_DEVICE))
                errChk( UIEValueIsInvalidOrOutOfRange);
            if (newIntValue != chartInfo->device)
                {
                chartInfo->device = newIntValue;
                errChk( UpdateChartHWTimingMode(chartInfo));
                errChk( UpdateNumTraces(chartInfo));    /* alarms previously marked as having invalid channels may have valid channels now */
                }
            break;
        case ATTR_DAQ_CHART_CHANNEL_STRING:
            errChk( SetDaqChartChannelString(chartInfo, va_arg(parmInfo, char *)));
           break;
        case ATTR_DAQ_CHART_REFRESH_FREQUENCY:
            newDoubleValue = (double)va_arg(parmInfo, double);
            if (newDoubleValue != Pin(newDoubleValue, DAQ_CHART_MIN_REFRESH_FREQUENCY, DAQ_CHART_MAX_REFRESH_FREQUENCY))
                errChk( UIEValueIsInvalidOrOutOfRange);

            errChk( SetCtrlAttribute(panel, chartInfo->refreshTimerCtrl, ATTR_INTERVAL, (double)1.0 / newDoubleValue));
            break;
        case ATTR_DAQ_CHART_SCAN_FREQUENCY:
            newDoubleValue = (double)va_arg(parmInfo, double);
            if (newDoubleValue != Pin(newDoubleValue, DAQ_CHART_MIN_SCAN_FREQUENCY, DAQ_CHART_MAX_SCAN_FREQUENCY))
                errChk(UIEValueIsInvalidOrOutOfRange);
            if (newDoubleValue != chartInfo->scanFrequency)
                {
                errChk( SetCtrlAttribute(panel, chartInfo->scanTimerCtrl, ATTR_INTERVAL, (double)1.0/newDoubleValue));
                chartInfo->scanFrequency = newDoubleValue;
                errChk( UpdateChartHWTimingMode(chartInfo));
                }
            break;
        case ATTR_DAQ_CHART_ENABLED:
            newIntValue = !!va_arg(parmInfo, int);
            if (newIntValue != chartInfo->enabled)
                {
                errChk( SetCtrlAttribute(panel, chartInfo->refreshTimerCtrl, ATTR_ENABLED, newIntValue));
                errChk( SetCtrlAttribute(panel, chartInfo->scanTimerCtrl, ATTR_ENABLED, newIntValue));
                chartInfo->enabled = newIntValue;
                errChk( UpdateChartHWTimingMode(chartInfo));
                if (!chartInfo->enabled && chartInfo->dialog)
                    {
                    errChk( SetCtrlVal(chartInfo->dialog, chartInfo->actualFrequencyNumeric, 0.0));
                    DaqCtrl_ClearDisplayedError((DaqCtrl)chartInfo);
                    }
                chartInfo->autoErrorClearingNeeded = !chartInfo->enabled;
                }
            break;
        case ATTR_DAQ_CHART_VISIBLE:
        case ATTR_DAQ_CHART_ALLOW_RCLICK_EDIT:
        case ATTR_DAQ_CHART_EDIT_DIALOG_TITLE:
        case ATTR_DAQ_CHART_EDIT_DIALOG_DISPLAYED:
        case ATTR_DAQ_CHART_EDIT_DIALOG_MODAL:
        case ATTR_DAQ_CHART_IO_DIRECTION:
        case ATTR_DAQ_CHART_VERSION:
        case ATTR_DAQ_CHART_ERROR_NOTIFICATION:
            errChk( DaqCtrl_SetAttributeFromParmInfo((DaqCtrl)chartInfo, attribute,  parmInfo));
            break;
        case ATTR_DAQ_CHART_HIGH_LIMIT:
            newDoubleValue = (double)va_arg(parmInfo, double);
            if (newDoubleValue != chartInfo->highLimit)
                {
                chartInfo->highLimit = newDoubleValue;
                errChk( UpdateChartHWTimingMode(chartInfo));
                if (chartInfo->tieCtrlRangeToLimits)
                    {
                    int savedBole = SetBOLE(FALSE);

                    error = SetAxisRange (panel, daqChart, VAL_NO_CHANGE, 0.0, 0.0, VAL_MANUAL, chartInfo->lowLimit, chartInfo->highLimit);
                    SetBOLE(savedBole);
                    errChk( error);
                    }
                }
            break;
        case ATTR_DAQ_CHART_LOW_LIMIT:
            newDoubleValue = (double)va_arg(parmInfo, double);
            if (newDoubleValue != chartInfo->lowLimit)
                {
                chartInfo->lowLimit = newDoubleValue;
                errChk( UpdateChartHWTimingMode(chartInfo));
                if (chartInfo->tieCtrlRangeToLimits)
                    {
                    int savedBole = SetBOLE(FALSE);

                    error = SetAxisRange (panel, daqChart, VAL_NO_CHANGE, 0.0, 0.0, VAL_MANUAL, chartInfo->lowLimit, chartInfo->highLimit);
                    SetBOLE(savedBole);
                    errChk( error);
                    }
                }
            break;
        case ATTR_DAQ_CHART_NUM_CHANNELS:
            errChk( UIEAttributeNotSettable);
            break;
        case ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE:
            {
            int newSize;

            newSize = va_arg(parmInfo, int);
            if (newSize < MIN_DAQ_CHART_HISTORY_BUFFER_SIZE)
                errChk( UIEValueIsInvalidOrOutOfRange);

            if  (newSize != chartInfo->historyBufferSize)
                errChk( CreateHistory(chartInfo, newSize, chartInfo->numChannels));
            }
            break;
        case ATTR_DAQ_CHART_HISTORY_LENGTH:
            errChk( UIEAttributeNotSettable);
            break;
        case ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG:
            newIntValue = !!va_arg(parmInfo, int);
            if (newIntValue != chartInfo->useHardwareTiming)
                errChk( SetChartHWTimingMode(chartInfo, newIntValue));
            break;
        case ATTR_DAQ_CHART_WARN_ON_HW_TIMING_RESTARTS:
            chartInfo->warnOnHWTimedAcqRestarts = va_arg(parmInfo, int) != 0;
            break;
        case ATTR_DAQ_CHART_TIE_CTRL_RANGE_TO_LIMITS:
            newIntValue = va_arg(parmInfo, int);
            if (chartInfo->tieCtrlRangeToLimits != newIntValue)
                {
                chartInfo->tieCtrlRangeToLimits = newIntValue;

                if (chartInfo->tieCtrlRangeToLimits)
                    {
                    double  max, min;
                    int     savedBole;

                    errChk( DAQ_Chart_GetAttribute(panel, daqChart, ATTR_DAQ_CHART_LOW_LIMIT, &min));
                    errChk( DAQ_Chart_GetAttribute(panel, daqChart, ATTR_DAQ_CHART_HIGH_LIMIT, &max));
                    savedBole = SetBOLE(FALSE);
                    error = SetAxisRange (panel, daqChart, VAL_NO_CHANGE, 0, 0, VAL_MANUAL, min, max);
                    SetBOLE(savedBole);
                    errChk( error);
                    }
                }
            break;
        default:
            if (DataLog_IsAttribute(attribute))
                errChk( DataLog_SetAttributeFromParmInfo(chartInfo->dataLog, attribute, parmInfo));
            else
                errChk( UIEInvalidAttribute);
            break;
        }

        /* make sure the editing dialog displays the latest changes */
    if (!chartInfo->updatingCtrlFromDialog && chartInfo->dialog && chartInfo->dialogDisplayed)
        {
        errChk( CopyAttributesToDialog(chartInfo));
        errChk( UpdateDialogDimming(chartInfo));
        }

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_GetAttribute(int panel, int daqChart, int attribute, void *value)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = 0;

    if (!value)
        errChk( UIENullPointerPassed);

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

    switch (attribute)
        {
        case ATTR_DAQ_CHART_DEVICE:
        case ATTR_DAQ_CHART_CHANNEL_STRING:
        case ATTR_DAQ_CHART_CHANNEL_STRING_COPY:
        case ATTR_DAQ_CHART_CHANNEL_STRING_LENGTH:
        case ATTR_DAQ_CHART_VISIBLE:
        case ATTR_DAQ_CHART_ALLOW_RCLICK_EDIT:
        case ATTR_DAQ_CHART_EDIT_DIALOG_MODAL:
        case ATTR_DAQ_CHART_EDIT_DIALOG_DISPLAYED:
        case ATTR_DAQ_CHART_EDIT_DIALOG_TITLE:
        case ATTR_DAQ_CHART_EDIT_DIALOG_TITLE_COPY:
        case ATTR_DAQ_CHART_EDIT_DIALOG_TITLE_LENGTH:
        case ATTR_DAQ_CHART_IO_DIRECTION:
        case ATTR_DAQ_CHART_HIGH_LIMIT:
        case ATTR_DAQ_CHART_LOW_LIMIT:
        case ATTR_DAQ_CHART_TIE_CTRL_RANGE_TO_LIMITS:
        case ATTR_DAQ_CHART_VERSION:
        case ATTR_DAQ_CHART_ERROR_NOTIFICATION:
            errChk( DaqCtrl_GetAttribute((DaqCtrl)chartInfo, attribute, value));
            break;
        case ATTR_DAQ_CHART_REFRESH_FREQUENCY:
            {
            double  interval;

            errChk( GetCtrlAttribute(panel, chartInfo->refreshTimerCtrl, ATTR_INTERVAL, &interval));
            *(double *)value = 1.0 / interval;
            }
            break;
        case ATTR_DAQ_CHART_SCAN_FREQUENCY:
            *(double *)value = chartInfo->scanFrequency;
            break;
        case ATTR_DAQ_CHART_ENABLED:
            errChk( GetCtrlAttribute(panel, chartInfo->refreshTimerCtrl, ATTR_ENABLED, value));
            break;
        case ATTR_DAQ_CHART_NUM_CHANNELS:
            *(int *)value = chartInfo->numChannels;
            break;
        case ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE:
            *(int *)value = chartInfo->historyBufferSize;
            break;
        case ATTR_DAQ_CHART_HISTORY_LENGTH:
            *(int *)value = chartInfo->historyLength;
            break;
        case ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG:
            *(int *)value = !!chartInfo->useHardwareTiming;
            break;
        case ATTR_DAQ_CHART_WARN_ON_HW_TIMING_RESTARTS:
            *(int *)value = !!chartInfo->warnOnHWTimedAcqRestarts;
            break;
        default:
            if (DataLog_IsAttribute(attribute))
                errChk( DataLog_GetAttribute(chartInfo->dataLog, attribute, value));
            else
                errChk( UIEInvalidAttribute);
            break;
        }

Error:
    return error;
}

/*****************************************************************************************/

    /* the callback data is the DAQ_ChartInfo */
static int CVICALLBACK DAQ_ChartRefreshTimerCallback(int panel, int timerControl, int event, void *callbackData, int eventData1, int eventData2)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo;

    UNUSED(panel);
    UNUSED(timerControl);
    UNUSED(eventData1);
    UNUSED(eventData2);

    chartInfo = (DAQ_ChartInfo)callbackData;

    switch (event)
        {
        case EVENT_TIMER_TICK:
            errChk( PlotPendingScans(chartInfo));
            break;
        case EVENT_DISCARD:
            chartInfo->refreshTimerCtrl = 0;   /* if the panel is discarded, the timer might be automatically discarded before the chart. */
            break;
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/*****************************************************************************************/

    /* the callback data is the DAQ_ChartInfo */
static int CVICALLBACK DAQ_ChartScanTimerCallback(int panel, int timerControl, int event, void *callbackData, int eventData1, int eventData2)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo;

    UNUSED(panel);
    UNUSED(timerControl);
    UNUSED(eventData1);
    UNUSED(eventData2);

    chartInfo = (DAQ_ChartInfo)callbackData;

    switch (event)
        {
        case EVENT_TIMER_TICK:
            errChk( UpdateInput(chartInfo));
            break;
        case EVENT_DISCARD:
            chartInfo->scanTimerCtrl = 0;   /* if the panel is discarded, the timer might be automatically discarded before the chart. */
            break;
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/*****************************************************************************************/

static int CVIFUNC GetDAQChartErrorString(DAQ_ChartInfo chartInfo, int errorCode, char **errorTitle, char **errorMessage)
{
    UNUSED(chartInfo);

    switch (errorCode)
        {
        case DAQ_CHART_WARNING_HW_TIMED_BUF_OVERRUN:
            *errorTitle = "WARNING: HW Timed Transfer Buffer Overrun.";
            *errorMessage =
                "A hardware timed transfer was restarted because the "
                "transfer buffer overflowed before all of the data points could be "
                "processed.  This will result in a larger time gap between some data points. "
                "To avoid this use a faster machine, or try reducing the scan rate "
                "and/or the amount of processing and plotting.";
            return 1;
        case DAQ_CHART_WARNING_HW_TIMED_TRANSFER_ABORTED:
            *errorTitle =   "WARNING: HW Timed Transfer Interrupted.";
            *errorMessage =
                "A hardware timed transfer was restarted after being unexpectedly "
                "terminated.  This will result in a larger time gap between some data points. "
                "This may be the result of an attempt to read from the same device using a "
                "NI-DAQ analog input function during the hardware timed transfer.";
            return 1;
        case DAQ_CHART_WARNING_INVALID_CHANNEL:
            *errorTitle = "WARNING: Some alarm(s) have invalid channels";
            *errorMessage =
                "One or more alarms have channels which are not included in the channel string.";
            return 1;
        case DAQ_CHART_WARNING_DATALOGGING_TERMINATED:
            *errorTitle = "WARNING: Data Logging Terminated";
            *errorMessage =
                "Data logging was terminated because the channel string was changed.";
            return 1;
        case DAQ_CHART_ERROR_MAX_64_CHANNELS_AND_ALARMS:
            *errorTitle = "Error: Too many channels and/or alarms";
            *errorMessage =
                "The number of channels plus the number of alarms cannot exceed 64.";
            return 1;
        default:
            return 0;
        }
}

/*****************************************************************************************/

static int UpdateInput(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     channelStringPresent;

    if (chartInfo->ioDirection != ANALOG_OUTPUT)
        {
        channelStringPresent =  chartInfo->channelString && *chartInfo->channelString;

        if (channelStringPresent)
            errChk( AcquireScans(chartInfo));
        }

Error:
    return error;
}

/*****************************************************************************************/

static void CheckAlarms(DAQ_ChartInfo chartInfo, double *scan, double *time)
{
    int     index;
    int     numAlarms;
    int     numChannels;
    Alarm   *alarm;
    int     wasOn;
    enum    {GOOD_SIDE_OF_DEADBAND, IN_DEAD_BAND, BAD_SIDE_OF_DEADBAND} valueState;
    int     isOn;
    double  value;

    numAlarms = ListNumItems(chartInfo->alarmList);
    numChannels = chartInfo->numChannels;

    for (index = 1; index <= numAlarms; index++)
        {
        alarm = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);

        if (alarm->valid && alarm->channelIndex < numChannels)
            {
            value = scan[alarm->channelIndex];
            wasOn = alarm->on;

            switch (alarm->mode)
                {
                default:
                case VAL_ALARM_ABOVE_LIMIT:
                    if (value > alarm->limit)
                        valueState = BAD_SIDE_OF_DEADBAND;
                    else
                    if (value > alarm->limit - alarm->deadband)
                        valueState = IN_DEAD_BAND;
                    else
                        valueState = GOOD_SIDE_OF_DEADBAND;
                    break;
                case VAL_ALARM_BELOW_LIMIT:
                    if (value < alarm->limit)
                        valueState = BAD_SIDE_OF_DEADBAND;
                    else
                    if (value < alarm->limit + alarm->deadband)
                        valueState = IN_DEAD_BAND;
                    else
                        valueState = GOOD_SIDE_OF_DEADBAND;
                    break;
                case VAL_ALARM_IN_RANGE:
                    if (value < alarm->limit && value > alarm->lowerLimit)
                        valueState = BAD_SIDE_OF_DEADBAND;
                    else
                    if (value < alarm->limit + alarm->deadband  && value > alarm->lowerLimit - alarm->deadband)
                        valueState = IN_DEAD_BAND;
                    else
                        valueState = GOOD_SIDE_OF_DEADBAND;
                    break;
                case VAL_ALARM_OUT_OF_RANGE:
                    if (value > alarm->limit || value < alarm->lowerLimit)
                        valueState = BAD_SIDE_OF_DEADBAND;
                    else
                    if (value > alarm->limit - alarm->deadband  || value < alarm->lowerLimit + alarm->deadband)
                        valueState = IN_DEAD_BAND;
                    else
                        valueState = GOOD_SIDE_OF_DEADBAND;
                    break;
                }

            if (wasOn)
                isOn = (valueState != GOOD_SIDE_OF_DEADBAND);
            else
                isOn = (valueState == BAD_SIDE_OF_DEADBAND);

            alarm->on = isOn;
            if (wasOn != isOn)
                {
                if (isOn)
                    SendAlarmEvent(chartInfo, alarm, EVENT_DAQ_CHART_ALARM_ON, value, time);
                else
                    SendAlarmEvent(chartInfo, alarm, EVENT_DAQ_CHART_ALARM_OFF, value, time);
                }
            }
        }

    HandleAlarmControlledConditionalLogging(chartInfo);
}

/*****************************************************************************************/

static void HandleAlarmControlledConditionalLogging(DAQ_ChartInfo chartInfo)
{
    int numAlarms;
    int waitForCondition;
    int logConditionally;

    numAlarms = ListNumItems(chartInfo->alarmList);

    DataLog_GetAttribute(chartInfo->dataLog, ATTR_DATALOG_LOG_CONDITIONALLY, &logConditionally);    /* no error can occur */

    if (logConditionally && numAlarms > 0)
        {
        Alarm   *alarmPtr;
        int     index;
        int     numCheckedAlarms = 0;

        waitForCondition = TRUE;
        for (index = 1; index <= numAlarms && waitForCondition; index++)
            {
            alarmPtr = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);

            if (alarmPtr->logAction)
                {
                numCheckedAlarms++;

                if (alarmPtr->on)
                    {
                    if (alarmPtr->logAction & ALARM_LOG_WHILE_ON)
                        waitForCondition = FALSE;
                    }
                else
                    {
                    if (alarmPtr->logAction & ALARM_LOG_WHILE_OFF)
                        waitForCondition = FALSE;
                    }
                }
            }
            if (numCheckedAlarms == 0)      /* if no alarms have conditions, disable the feature, otherwise logging would never be on */
                waitForCondition = FALSE;
        }
    else
        waitForCondition = FALSE;

    DataLog_SetAttribute(chartInfo->dataLog, ATTR_DATALOG_PAUSE_FOR_CONDITION, waitForCondition);   /* no error possible */
}

/*****************************************************************************************/

    /*  time is optional parameter (pass null to have it ignored) */
static void SendAlarmEvent(DAQ_ChartInfo chartInfo, Alarm *alarm, int event, double value, double *time)
{
    char    alarmMsg[200];

    CallCtrlCallback(chartInfo->panel, chartInfo->baseCtrl, event, (int)alarm->name, (int)&value, 0);

    if (IsRangeMode(alarm->mode))
        sprintf(alarmMsg,"ALARM %s EVENT for '%s'. Value = %.6f, Lower Limit = %.6f, Upper Limit = %.6f",
            event == EVENT_DAQ_CHART_ALARM_ON ? "ON" : "OFF", alarm->name,
            value, alarm->lowerLimit, alarm->limit);
    else
        sprintf(alarmMsg,"ALARM %s EVENT for '%s'. Value = %.6g, Limit = %.6f",
            event == EVENT_DAQ_CHART_ALARM_ON ? "ON" : "OFF", alarm->name,
            value, alarm->limit);

    DataLog_LogMessage(chartInfo->dataLog, alarmMsg, time);
}

/*****************************************************************************************/

static int CVICALLBACK DAQ_ChartCallback(int panel, int daqChart, int event, void *callbackData, int eventData1, int eventData2)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = (DAQ_ChartInfo)callbackData;

    UNUSED(eventData1);
    UNUSED(eventData2);

    switch (event)
        {
        case EVENT_DISCARD:
            DiscardDAQ_ChartInfo(chartInfo);
            break;
        case EVENT_RIGHT_CLICK:
            {
            int allowRightClickEdits;
            int result;

            errChk( DAQ_Chart_GetAttribute(panel, daqChart, ATTR_DAQ_CHART_ALLOW_RCLICK_EDIT, &allowRightClickEdits));

            if (allowRightClickEdits)
                {
                errChk( result = DaqCtrl_RunRightClickPopupMenu(panel, MakePoint(eventData2, eventData1)));
                if (result)
                    errChk( Internal_DAQ_Chart_SetAttribute(panel, daqChart, ATTR_DAQ_CHART_EDIT_DIALOG_DISPLAYED, 1));
                }
            }
            break;
         }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/*****************************************************************************************/

static int CreateDummyEditingDAQChart(int *panel, int *dummyChart, int modal)
{
    int error = 0;
    int width, height;
    int margin = 3 * DLG_MARGIN;

    *panel = 0;
    errChk( *panel = NewPanel (0, "DAQ Chart For Quick Configuration Editing", 0, 0, 10, 10));
    errChk( *dummyChart = Internal_DAQ_Chart_Create (*panel, "Temporary DAQ Chart For Configuration Editing", margin, margin, 0, 0, 0));

    errChk( GetCtrlAttribute(*panel, *dummyChart, ATTR_WIDTH, &width));
    errChk( GetCtrlAttribute(*panel, *dummyChart, ATTR_HEIGHT, &height));

    errChk( SetPanelAttribute(*panel, ATTR_WIDTH, 2 * margin + width));
    errChk( SetPanelAttribute(*panel, ATTR_HEIGHT, 2 * margin + height));

    errChk( SetPanelAttribute (*panel, ATTR_CLOSE_ITEM_VISIBLE, FALSE));

    errChk( SetPanelPos(*panel, VAL_AUTO_CENTER, VAL_AUTO_CENTER));
    errChk( Internal_DAQ_Chart_SetAttribute (*panel, *dummyChart, ATTR_DAQ_CHART_HIGH_LIMIT, ATTR_DAQ_CHART_UPDATE_HIGH_LIMIT_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute (*panel, *dummyChart, ATTR_DAQ_CHART_LOW_LIMIT, ATTR_DAQ_CHART_UPDATE_LOW_LIMIT_DEFAULT));
    errChk( Internal_DAQ_Chart_SetAttribute (*panel, *dummyChart, ATTR_DAQ_CHART_TIE_CTRL_RANGE_TO_LIMITS, TRUE));

    if (modal)
        errChk( InstallPopup(*panel));
    else
        errChk( DisplayPanel(*panel));

Error:
    if (error < 0 && *panel > 0)
        {
        DiscardPanel(*panel);
        *panel = 0;
        }

    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_EditPopup(int panel, int daqChart, char *dialogTitle, int modal)
{
    int     error = 0;
    int     deletePanel = FALSE;
    DaqCtrl daqCtrl = 0;

    if (panel == 0 && daqChart == 0)
        {
        errChk( CreateDummyEditingDAQChart(&panel, &daqChart, modal));
        deletePanel = TRUE;
        }

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, &daqCtrl));
    errChk (DaqCtrl_EditPopup(daqCtrl, dialogTitle, modal));

Error:
    if (deletePanel && panel > 0)
        DiscardPanel(panel);

    return error;
}

/*****************************************************************************************/

    /*  copy the control's attributes to the dialog fields */
static int CVIFUNC CopyAttributesToDialog(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     panel, ctrl, dialog;
    double  frequency;
    int     useHardwareTiming;
    double  refreshFrequency;
    int     historySize;

    dialog = chartInfo->dialog;
    panel = chartInfo->panel;
    ctrl = chartInfo->baseCtrl;

    errChk( DaqCtrl_CopyAttributesToDialog((DaqCtrl)chartInfo));

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_SCAN_FREQUENCY, &frequency));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_REFRESH_FREQUENCY, &refreshFrequency));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE, &historySize));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG, &useHardwareTiming));

    errChk( SetCtrlVal(chartInfo->dialog, chartInfo->frequencyNumeric, frequency));
    errChk( SetCtrlVal(chartInfo->dialog, chartInfo->refreshNumeric, refreshFrequency));
    errChk( SetCtrlVal(chartInfo->dialog, chartInfo->historySizeNumeric, historySize));
    errChk( SetCtrlVal(chartInfo->dialog, chartInfo->useHardwareTimingCheckBox, useHardwareTiming));

    errChk( CopyDataLogAttrToDialog(chartInfo));

    UpdateDialogDimming(chartInfo);

Error:
    return error;
}
/*****************************************************************************************/

    /*  copy the control's datalog attributes to the dialog fields */
static int CopyDataLogAttrToDialog(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     panel, ctrl, dialog;
    int     loggingPaused;
    int     logEvents;
    int     fileFormat;
    int     startOnLoad;
    int     loggingOn;
    char    dataLogPath[MAX_PATHNAME_LEN];
    char    eventLogPath[MAX_PATHNAME_LEN];

    dialog = chartInfo->dialog;
    panel = chartInfo->panel;
    ctrl = chartInfo->baseCtrl;

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_PAUSED, &loggingPaused));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_EVENTS, &logEvents));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_FILE_FORMAT, &fileFormat));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_START_LOGGING_ON_LOAD, &startOnLoad));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_ON, &loggingOn));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_DATA_LOG_FILE, dataLogPath));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_EVENT_LOG_FILE, eventLogPath));

    errChk( SetCtrlVal(dialog, chartInfo->pauseLoggingToggle, loggingPaused));
    errChk( SetCtrlVal(dialog, chartInfo->logEventsCheckBox, logEvents));
    errChk( SetCtrlVal(dialog, chartInfo->fileFormatRing, fileFormat));
    errChk( SetCtrlVal(dialog, chartInfo->startOnLoadCheckBox, startOnLoad));
    errChk( SetCtrlVal(dialog, chartInfo->startLoggingToggle, loggingOn));
    errChk( SetCtrlVal(dialog, chartInfo->dataLogFileString, dataLogPath));
    errChk( SetCtrlVal(dialog, chartInfo->eventLogFileString, eventLogPath));
    errChk( UpdateLoggingStateIndicator(chartInfo));

Error:
    return error;
}

/*****************************************************************************************/

static int UpdateLoggingStateIndicator(DAQ_ChartInfo chartInfo)
{
    int     error;
    int     loggingState;
    char    *stateString = "";
    char    stateStrBuf[50];
    int     numScans;
    int     panel;
    int     ctrl;
    int     dialog;

    panel = chartInfo->panel;
    ctrl = chartInfo->baseCtrl;
    dialog = chartInfo->dialog;

    if (dialog > 0)
        {
        errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_STATE, &loggingState));

        switch (loggingState)
            {
            case DATA_LOG_ACTIVE:
                errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_NUM_SCANS_LOGGED, &numScans));
                stateString = stateStrBuf;
                sprintf(stateStrBuf, "Logging Active (%d scans)", numScans);
                break;
            case DATA_LOG_PAUSED:
                stateString = "Logging Paused";
                break;
            case DATA_LOG_STOPPED:
                stateString = "Logging Off";
                break;
            case DATA_LOG_WAITING_FOR_CONDITION:
                stateString = "Waiting For Condition";
                break;
            }

        errChk( SetCtrlVal(dialog, chartInfo->loggingStateString, stateString));
        }

Error:
    return error;
}

/*****************************************************************************************/

    /* save the current attribute values of the control */
static int CVIFUNC RememberAttributes(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int panel = chartInfo->panel;
    int ctrl  = chartInfo->baseCtrl;

    errChk( DaqCtrl_RememberAttributes((DaqCtrl)chartInfo));    /* base class functionality */

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_REFRESH_FREQUENCY, &chartInfo->savedRefreshFrequency));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_SCAN_FREQUENCY, &chartInfo->savedScanFrequency));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG, &chartInfo->savedUseHardwareTiming));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE, &chartInfo->savedHistorySize));
    errChk( RememberAlarms(chartInfo))
    errChk( RememberTraceColors(chartInfo));
    errChk( RememberDataLogAttributes(chartInfo));

Error:
    return error;
}

/*****************************************************************************************/

    /* save the current attribute values of the control */
static int RememberDataLogAttributes(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int panel = chartInfo->panel;
    int ctrl  = chartInfo->baseCtrl;

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_PAUSED, &chartInfo->savedLoggingPaused));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_EVENTS, &chartInfo->savedLogEvents));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_FILE_FORMAT, &chartInfo->savedFileFormat));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_START_LOGGING_ON_LOAD, &chartInfo->savedStartOnLoad));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_ON, &chartInfo->savedLoggingOn));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_DATA_LOG_FILE, &chartInfo->savedDataLogFilePath));

    free(chartInfo->savedColumnHeadings);
    free(chartInfo->savedComment);
    chartInfo->savedColumnHeadings = 0;
    chartInfo->savedComment = 0;

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE, &chartInfo->savedPromptOnOverwrite));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_FIELD_WIDTH, &chartInfo->savedFieldWidth));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_WRITE_LOG_HEADER, &chartInfo->savedWriteHeader));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_COLUMN_HEADINGS_COPY, &chartInfo->savedColumnHeadings));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_COMMENT_COPY, &chartInfo->savedComment));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_MODE, &chartInfo->savedLogMode));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_NUM_SCANS_TO_LOG, &chartInfo->savedNumScans));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_CONDITIONALLY, &chartInfo->savedLogConditionally));


Error:
    return error;
}

/*****************************************************************************************/

static int RememberAlarms(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     numAlarms, index;
    Alarm   alarm, *ptrToAlarmToCopy;

    if (!chartInfo->savedAlarmList)
        nullChk( chartInfo->savedAlarmList = ListCreate(sizeof(alarm)));
    else
        ListClear(chartInfo->savedAlarmList);

    numAlarms = ListNumItems(chartInfo->alarmList);

    for (index = 1; index <= numAlarms; index++)
        {
        ptrToAlarmToCopy = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);
        errChk( CopyAlarm(ptrToAlarmToCopy, &alarm));

        nullChk( ListInsertItem(chartInfo->savedAlarmList, &alarm, END_OF_LIST));
        }

Error:
    if (error < 0)
        DisposeSavedAlarmList(chartInfo);

    return error;
}

/*****************************************************************************************/

static int RememberTraceColors(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int numTraces, index;
    int color;

    if (!chartInfo->savedTraceColorList)
        nullChk( chartInfo->savedTraceColorList = ListCreate(sizeof(int)));
    else
        ListClear(chartInfo->savedTraceColorList);

    errChk( GetCtrlAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_NUM_TRACES, &numTraces));

    for (index = 1; index <= numTraces; index++)
        {
        errChk( GetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, index, ATTR_TRACE_COLOR, &color));
        nullChk( ListInsertItem(chartInfo->savedTraceColorList, &color, END_OF_LIST));
        }

Error:
    if (error < 0)
        {
        ListDispose(chartInfo->savedTraceColorList);
        chartInfo->savedTraceColorList = 0;
        }

    return error;
}

/*****************************************************************************************/

    /* copy the saved values of the attributes back to the control */
static int CVIFUNC RevertAttributes(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int panel = chartInfo->panel;
    int ctrl = chartInfo->baseCtrl;
    int loggingOn;

    chartInfo->revertingSettings = TRUE;
    chartInfo->updatingCtrlFromDialog = TRUE;

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_ON, &loggingOn));

        /*  clear the alarms before the channel string is set because otherwise if the number of channels from the new configuration
            plus the number of alarms from the current configuration exceed 64, an error occurs due to the strip chart limit of 64 traces
        */
    DisposeAllChartAlarms(chartInfo);   /* clears the charts alarm list, freeing resources owned by the alarms */

    if (loggingOn)  /* can't change channel string while logging, it would terminate logging */
        DaqCtrl_ChannelStringIsRevertable((DaqCtrl)chartInfo, FALSE);

    errChk( DaqCtrl_RevertAttributes((DaqCtrl)chartInfo));

    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_SCAN_FREQUENCY, chartInfo->savedScanFrequency));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_REFRESH_FREQUENCY, chartInfo->savedRefreshFrequency));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG, chartInfo->savedUseHardwareTiming));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE, chartInfo->savedHistorySize));

    errChk( RevertAlarms(chartInfo));
    errChk( RevertTraceColors(chartInfo));
    errChk( RevertDataLogSettings(chartInfo));

Error:
    chartInfo->updatingCtrlFromDialog = FALSE;
    chartInfo->revertingSettings = FALSE;
    DaqCtrl_ChannelStringIsRevertable((DaqCtrl)chartInfo, TRUE);

    if (error >= 0)
        WarnIfAlarmsHaveInvalidChannels(chartInfo);     /*  do this now, since it was suppresed when reverting settings because we can't simultaneously revert the alarms and the channel string */

    return error;
}

/*****************************************************************************************/

    /* copy the saved values of the datalog attributes back to the control */
static int RevertDataLogSettings(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     panel = chartInfo->panel;
    int     ctrl = chartInfo->baseCtrl;
    int     loggingOn;

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_ON, &loggingOn));

    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOG_EVENTS, chartInfo->savedLogEvents));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_START_LOGGING_ON_LOAD, chartInfo->savedStartOnLoad));

    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE, chartInfo->savedPromptOnOverwrite));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOG_FIELD_WIDTH, chartInfo->savedFieldWidth));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_WRITE_LOG_HEADER, chartInfo->savedWriteHeader));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_COLUMN_HEADINGS, chartInfo->savedColumnHeadings));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_COMMENT, chartInfo->savedComment));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOG_MODE, chartInfo->savedLogMode));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_NUM_SCANS_TO_LOG, chartInfo->savedNumScans));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOG_CONDITIONALLY, chartInfo->savedLogConditionally));

    if (!loggingOn) /* can't revert some things while logging */
        {
        errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOG_FILE_FORMAT, chartInfo->savedFileFormat));
        errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_DATA_LOG_FILE, chartInfo->savedDataLogFilePath));
        }

Error:
    return error;
}

/*****************************************************************************************/

static int RevertTraceColors(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int numTraces, index;
    int color;

    if (chartInfo->savedTraceColorList)
        {
        errChk( GetCtrlAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_NUM_TRACES, &numTraces));

        numTraces = Min(numTraces, ListNumItems(chartInfo->savedTraceColorList));

        for (index = 1; index <= numTraces; index++)
            {
            ListGetItem(chartInfo->savedTraceColorList, &color, index);
            errChk( SetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, index, ATTR_TRACE_COLOR, color));
            }
        }

Error:
    return error;
}

/*****************************************************************************************/

static int RevertAlarms(DAQ_ChartInfo chartInfo)
{
    int error = 0;

    if (chartInfo->savedAlarmList)
        errChk( ReplaceAllChartAlarms(chartInfo, chartInfo->savedAlarmList));

Error:
    if (error < 0)
        DisposeSavedAlarmList(chartInfo);

    return error;
}

/*****************************************************************************************/

    /*  newAlarmList is emptied, but not disposed */
static int ReplaceAllChartAlarms(DAQ_ChartInfo chartInfo, ListType newAlarmList)
{
    int error = 0;

    DisposeAllChartAlarms(chartInfo);   /* clears the charts alarm list, freeing resources owned by the alarms */
    nullChk( ListAppend(chartInfo->alarmList, newAlarmList));   /* copies the saved list into the alarm list */
    ListClear(newAlarmList);
    errChk( UpdateNumTraces(chartInfo));

Error:
    return error;
}

/*****************************************************************************************/

    /* copy current dialog control settings to the control's attributes */
static int CVIFUNC ApplyDialogSettings(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     dialog;
    int     panel;
    int     ctrl;
    int     useHardwareTiming;
    int     historySize;
    double  refreshFrequency;
    double  scanFrequency;

    dialog = chartInfo->dialog;
    panel = chartInfo->panel;
    ctrl = chartInfo->baseCtrl;

    chartInfo->updatingCtrlFromDialog = TRUE;

    errChk( DaqCtrl_ApplyDialogSettings((DaqCtrl)chartInfo));

    errChk( GetCtrlVal(dialog, chartInfo->useHardwareTimingCheckBox, &useHardwareTiming));
    errChk( GetCtrlVal(dialog, chartInfo->frequencyNumeric, &scanFrequency));
    errChk( GetCtrlVal(dialog, chartInfo->refreshNumeric, &refreshFrequency));
    errChk( GetCtrlVal(dialog, chartInfo->historySizeNumeric, &historySize));

    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_REFRESH_FREQUENCY, refreshFrequency));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_SCAN_FREQUENCY, (double)scanFrequency));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE, historySize));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG, useHardwareTiming));

    errChk( ApplyDataLogDialogSettings(chartInfo));

Error:
    chartInfo->updatingCtrlFromDialog = FALSE;
    return error;
}

/*****************************************************************************************/

    /* copy current dialog datalog control settings to the control's attributes */
static int ApplyDataLogDialogSettings(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     dialog;
    int     panel;
    int     ctrl;
    int     loggingOn;
    int     loggingPaused;
    int     logEvents;
    int     fileFormat;
    int     startOnLoad;

    dialog = chartInfo->dialog;
    panel = chartInfo->panel;
    ctrl = chartInfo->baseCtrl;

    errChk( GetCtrlVal(dialog, chartInfo->pauseLoggingToggle, &loggingPaused));
    errChk( GetCtrlVal(dialog, chartInfo->logEventsCheckBox, &logEvents));
    errChk( GetCtrlVal(dialog, chartInfo->fileFormatRing, &fileFormat));
    errChk( GetCtrlVal(dialog, chartInfo->startOnLoadCheckBox, &startOnLoad));
    errChk( GetCtrlVal(dialog, chartInfo->startLoggingToggle, &loggingOn));

    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_PAUSED, loggingPaused));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOG_EVENTS, logEvents));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOG_FILE_FORMAT, fileFormat));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_START_LOGGING_ON_LOAD, startOnLoad));
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_ON, loggingOn));

Error:
    return error;
}

/*****************************************************************************************/

static int CVIFUNC UpdateDialogDimming(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int ioDirection;
    int dialog;

    dialog = chartInfo->dialog;

    if (dialog > 0)
        {
        errChk( DaqCtrl_UpdateDialogDimming((DaqCtrl)chartInfo));

#ifdef INCLUDE_IO_DIRECTION
        errChk( GetCtrlVal(dialog, chartInfo->ioDirectionRing, &ioDirection));
        errChk( SetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_DIMMED, ioDirection != ANALOG_INPUT));
        errChk( SetCtrlAttribute(dialog, chartInfo->highLimitNumeric, ATTR_DIMMED, ioDirection != ANALOG_INPUT));
        errChk( SetCtrlAttribute(dialog, chartInfo->lowLimitNumeric, ATTR_DIMMED, ioDirection != ANALOG_INPUT));
#else
        UNUSED(ioDirection);
#endif

        errChk( SetCtrlAttribute(dialog, chartInfo->scanBackLogNumeric, ATTR_DIMMED, !chartInfo->useHardwareTiming));

        errChk( UpdateDataLogDialogDimming(chartInfo));
        }

Error:
    return error;
}

/*****************************************************************************************/

static int UpdateDataLogDialogDimming(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int dialog;
    int panel;
    int ctrl;
    int loggingOn;
    int logEvents;
    int fileFormat;

    dialog = chartInfo->dialog;
    panel = chartInfo->panel;
    ctrl = chartInfo->baseCtrl;

    if (dialog > 0)
        {
        errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOGGING_ON, &loggingOn));
        errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_EVENTS, &logEvents));
        errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_LOG_FILE_FORMAT, &fileFormat));

        errChk( SetCtrlAttribute(dialog, chartInfo->nameLogFileButton, ATTR_DIMMED, loggingOn));
        errChk( SetCtrlAttribute(dialog, chartInfo->eventLogFileString, ATTR_DIMMED, !logEvents));
        errChk( SetCtrlAttribute(dialog, chartInfo->fileFormatRing, ATTR_DIMMED, loggingOn));
        errChk( SetCtrlAttribute(dialog, chartInfo->textFormatButton, ATTR_DIMMED, loggingOn));
        errChk( SetCtrlAttribute(dialog, chartInfo->advancedLoggingButton, ATTR_DIMMED, loggingOn));
        errChk( SetCtrlAttribute(dialog, chartInfo->convertLogFileButton, ATTR_DIMMED, loggingOn));
        }

Error:
    return error;
}

/*****************************************************************************************/

static int UpdateAlarmDlgCtrls(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     index;
    Alarm   alarm;
    int     numAlarms;
    char    alarmListBoxString[256];

    if (chartInfo->dialog > 0)
        {
        numAlarms = ListNumItems(chartInfo->alarmList);

        errChk( ClearListCtrl(chartInfo->dialog, chartInfo->alarmListBox));

        for (index = 1; index <= numAlarms; index++)
            {
            ListGetItem(chartInfo->alarmList, &alarm, index);

            sprintf(alarmListBoxString, "\033fg%s%s", alarm.valid ? "default" : "FF0000", alarm.name);
            if (!alarm.valid)
                strcat(alarmListBoxString, " (invalid channel)");
            errChk( InsertListItem (chartInfo->dialog, chartInfo->alarmListBox, -1, alarmListBoxString, index));
            }

        SetCtrlAttribute(chartInfo->dialog, chartInfo->deleteAlarmButton, ATTR_DIMMED, numAlarms == 0);
        SetCtrlAttribute(chartInfo->dialog, chartInfo->editAlarmButton, ATTR_DIMMED, numAlarms == 0);
        }

Error:
    return error;
}

/*****************************************************************************************/

static int DAQ_Chart_FormatPopup(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     ctrlLeft;
    int     ctrlWidth;
    int     ctrlTop;
    int     dialog;
    int     popupRingWidth;
    int     sheetWidth;
    int     hMargin;
    int     sheetMargin;
    Rect    r;

    dialog = chartInfo->dialog;
    sheetMargin = hMargin = chartInfo->hMargin;

    errChk( DaqCtrl_CalcMinDialogWidth((DaqCtrl)chartInfo));
    errChk( CalcDataLogTabMinDialogWidth(chartInfo));

    errChk( DaqCtrl_FormatDialog((DaqCtrl)chartInfo));
    errChk( FormatDataLogTab(chartInfo));

        /* make related buttons the same width, help button must be last since it might be zero */
    errChk( SetCtrlsToAttributeExtreme (dialog, ATTR_WIDTH, VAL_TB_MAX, 0, chartInfo->editColorsButton, chartInfo->editAlarmButton,
        chartInfo->newAlarmButton, chartInfo->deleteAlarmButton, chartInfo->helpButton, 0));

        /*** position the refresh ring ***/
    errChk( GetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_TOP, &ctrlTop));
    errChk( SetCtrlAttribute(dialog, chartInfo->refreshRing, ATTR_TOP, ctrlTop));

    errChk( GetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_LEFT, &ctrlLeft));
    errChk( GetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_WIDTH, &ctrlWidth));
    errChk( GetCtrlAttribute(dialog, chartInfo->refreshRing, ATTR_WIDTH, &popupRingWidth));
    errChk( SetCtrlAttribute(dialog, chartInfo->refreshRing, ATTR_LEFT, ctrlLeft + ctrlWidth - popupRingWidth));
    errChk( SetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_WIDTH, ctrlWidth - popupRingWidth));

    errChk( DaqCtrl_FinishDialogSizing((DaqCtrl)chartInfo));

    errChk( EasyTab_GetBounds(dialog, chartInfo->sheetGroupCtrl, VAL_EASY_TAB_INTERIOR_BOUNDS, &r));
    sheetWidth = r.width;
    errChk(SetCtrlAttribute(dialog, chartInfo->alarmListBox, ATTR_WIDTH, sheetWidth - 2 * hMargin));
    errChk(SetCtrlAttribute(dialog, chartInfo->alarmListBox, ATTR_LABEL_LEFT, hMargin + sheetMargin));
    ExpandCtrlToMeetLabel(dialog, chartInfo->channelStringCtrl, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);

Error:
    return error;
}

/*****************************************************************************************/

static int CVIFUNC BuildDaqChartPopup(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int savedWaitCursor;
    int dialog = 0;
    int height;

    savedWaitCursor = GetWaitCursorState();
    SetWaitCursor(1);

    errChk( CreateDaqDlg((DaqCtrl)chartInfo));

    dialog = chartInfo->dialog;     /* initialized by CreateDaqDlg() */

    errChk( CreateDlgLists((DaqCtrl)chartInfo));
    errChk( CreateErrorControls((DaqCtrl)chartInfo));
    errChk( CreateConfigCtrls((DaqCtrl)chartInfo));
    errChk( CreateGeneralCtrls((DaqCtrl)chartInfo));
    errChk( CreateDAQCtrls((DaqCtrl)chartInfo));
    errChk( CreateStdButtons((DaqCtrl)chartInfo));
    errChk( CreateDlgSheetGroup((DaqCtrl)chartInfo));

    errChk( chartInfo->frequencyNumeric =   NewCtrl (dialog, CTRL_NUMERIC, "Requested Scan Frequency (Hz):", 0, 0));
    errChk( chartInfo->actualFrequencyNumeric = NewCtrl (dialog, CTRL_NUMERIC, "Actual Scan Frequency (Hz):", 0, 0));
    errChk( chartInfo->scanBackLogNumeric = NewCtrl (dialog, CTRL_NUMERIC, "Scan Backlog:", 0, 0));
    errChk( chartInfo->useHardwareTimingCheckBox =      NewCtrl (dialog, CTRL_CHECK_BOX, "Use Hardware Timing", 0, 0));
    errChk( chartInfo->refreshNumeric =     NewCtrl (dialog, CTRL_NUMERIC, "Chart Refresh Frequency (Hz):", 0, 0));
    errChk( chartInfo->refreshRing =        NewCtrl (dialog, CTRL_POPUP_MENU_RING, "", 0, 0));
    errChk( chartInfo->historySizeNumeric = NewCtrl (dialog, CTRL_NUMERIC, "History Size (in scans):", 0, 0));
    errChk( chartInfo->editColorsButton =   NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "Edit __Trace Colors...", 0, 0));

    errChk( chartInfo->alarmListBox =       NewCtrl (dialog, CTRL_LIST, "Alarm List:", 0, 0));
    errChk( chartInfo->newAlarmButton =     NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__New Alarm...", 0, 0));
    errChk( chartInfo->editAlarmButton =    NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Edit Alarm...", 0, 0));
    errChk( chartInfo->deleteAlarmButton =  NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Delete Alarm", 0, 0));

    errChk( CreateLoggingControls(chartInfo));

    errChk( MakeCtrlGrayRecessedIndicator(dialog, chartInfo->scanBackLogNumeric));
    errChk( MakeCtrlGrayRecessedIndicator(dialog, chartInfo->actualFrequencyNumeric));

        /*  Create and configure refresh ring */
    errChk( chartInfo->refreshPopupMenuBar = NewMenuBar(0));
    errChk( chartInfo->refreshPopupMenu = NewMenu (chartInfo->refreshPopupMenuBar, "", -1));
    errChk( NewMenuItem (chartInfo->refreshPopupMenuBar, chartInfo->refreshPopupMenu,
            "Very Fast", -1, 0, 0, (void *)1)); 
    errChk( NewMenuItem (chartInfo->refreshPopupMenuBar, chartInfo->refreshPopupMenu,
            "Fast", -1, 0, 0, (void *)2)); 
    errChk( NewMenuItem (chartInfo->refreshPopupMenuBar, chartInfo->refreshPopupMenu,
            "Normal", -1, 0, 0, (void *)3)); 
    errChk( NewMenuItem (chartInfo->refreshPopupMenuBar, chartInfo->refreshPopupMenu,
            "Slow", -1, 0, 0, (void *)4)); 
    errChk( NewMenuItem (chartInfo->refreshPopupMenuBar, chartInfo->refreshPopupMenu,
            "Very Slow", -1, 0, 0, (void *)5)); 

    errChk( SetCtrlAttribute (dialog, chartInfo->refreshRing, ATTR_CTRL_MODE, VAL_INDICATOR));
    errChk( GetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_HEIGHT, &height));
    errChk( SetCtrlAttribute(dialog, chartInfo->refreshRing, ATTR_HEIGHT, height));


        /*** Build lists of the controls on the dialog to make formatting easier later */

    nullChk( AppendToIntegerList(chartInfo->ctrlsWithLabelsOnLeft, 0, chartInfo->frequencyNumeric,
        chartInfo->actualFrequencyNumeric, chartInfo->refreshNumeric, chartInfo->historySizeNumeric,
        chartInfo->scanBackLogNumeric, 0));

    nullChk( AppendToIntegerList(chartInfo->ctrlsWithLabelsOnLeft, 0, chartInfo->dataLogFileString,
        chartInfo->eventLogFileString, chartInfo->fileFormatRing, 0));

    nullChk( ListInsertAfterItem(chartInfo->daqTabCtrls, &chartInfo->frequencyNumeric, &chartInfo->lowLimitNumeric, 0));
    nullChk( ListInsertAfterItem(chartInfo->daqTabCtrls, &chartInfo->actualFrequencyNumeric, &chartInfo->frequencyNumeric, 0));
    nullChk( ListInsertAfterItem(chartInfo->daqTabCtrls, &chartInfo->scanBackLogNumeric, &chartInfo->actualFrequencyNumeric, 0));
    nullChk( ListInsertAfterItem(chartInfo->daqTabCtrls, &chartInfo->useHardwareTimingCheckBox, &chartInfo->scanBackLogNumeric, 0));

    nullChk( chartInfo->alarmTabCtrls = BuildIntegerList(0, chartInfo->alarmListBox, chartInfo->newAlarmButton,
        chartInfo->editAlarmButton, chartInfo->deleteAlarmButton, 0));

    nullChk( ListInsertAfterItem(chartInfo->generalTabCtrls, &chartInfo->refreshNumeric, &chartInfo->titleString, 0));
    nullChk( ListInsertAfterItem(chartInfo->generalTabCtrls, &chartInfo->refreshRing, &chartInfo->refreshNumeric, 0));
    nullChk( ListInsertAfterItem(chartInfo->generalTabCtrls, &chartInfo->historySizeNumeric, &chartInfo->refreshRing, 0));
    nullChk( ListInsertAfterItem(chartInfo->generalTabCtrls, &chartInfo->editColorsButton, &chartInfo->tieLimitsCheckBox, 0));

        /* same as generalTabCtrls except minus refreshRing */
    nullChk( ListInsertAfterItem(chartInfo->generalTabCtrlsToDistribute, &chartInfo->refreshNumeric, &chartInfo->titleString, 0));
    nullChk( ListInsertAfterItem(chartInfo->generalTabCtrlsToDistribute, &chartInfo->historySizeNumeric, &chartInfo->refreshNumeric, 0));
    nullChk( ListInsertAfterItem(chartInfo->generalTabCtrlsToDistribute, &chartInfo->editColorsButton, &chartInfo->tieLimitsCheckBox, 0));

    nullChk( AppendToIntegerList(chartInfo->ctrlsOnLeft, 0,
        chartInfo->useHardwareTimingCheckBox, chartInfo->editColorsButton, chartInfo->alarmListBox,
        chartInfo->editAlarmButton, chartInfo->editAlarmButton, chartInfo->newAlarmButton, chartInfo->deleteAlarmButton,
        0));

    nullChk( AppendToIntegerList(chartInfo->ctrlsOnLeft, 0,
        chartInfo->nameLogFileButton, chartInfo->logEventsCheckBox, chartInfo->textFormatButton,
        chartInfo->advancedLoggingButton,
        0));

    nullChk( chartInfo->dataLogTabCtrls = BuildIntegerList(0, chartInfo->nameLogFileButton, chartInfo->dataLogFileString,
        chartInfo->eventLogFileString, chartInfo->logEventsCheckBox, chartInfo->fileFormatRing, chartInfo->textFormatButton,
        chartInfo->logControlTextMsg, chartInfo->logControlDeco, chartInfo->startLoggingToggle, chartInfo->pauseLoggingToggle,
        chartInfo->loggingStateString, chartInfo->startOnLoadCheckBox,
        chartInfo->advancedLoggingButton, chartInfo->convertLogFileButton, 0));

    nullChk( chartInfo->ctrlListsToDistribute = BuildIntegerList(0, chartInfo->daqTabCtrls, chartInfo->alarmTabCtrls,
        chartInfo->generalTabCtrlsToDistribute, chartInfo->configTabCtrls, 0));

#ifndef INCLUDE_IO_DIRECTION
    ListRemoveItem(chartInfo->daqTabCtrls, 0, ListFindItem(chartInfo->daqTabCtrls, &chartInfo->ioDirectionRing, FRONT_OF_LIST, 0));
    SetCtrlAttribute(chartInfo->dialog, chartInfo->ioDirectionRing, ATTR_VISIBLE, FALSE);
#endif

    errChk( EasyTab_AddSheet(dialog, chartInfo->sheetGroupCtrl, "DAQ", NULL, 1, VAL_TAB_SHEET_FROM_LIST, chartInfo->daqTabCtrls)); 
    errChk( EasyTab_AddSheet(dialog, chartInfo->sheetGroupCtrl, "Alarms", NULL, 1, VAL_TAB_SHEET_FROM_LIST, chartInfo->alarmTabCtrls)); 
    errChk( EasyTab_AddSheet(dialog, chartInfo->sheetGroupCtrl, "General", NULL, 1, VAL_TAB_SHEET_FROM_LIST, chartInfo->generalTabCtrls)); 
    errChk( EasyTab_AddSheet(dialog, chartInfo->sheetGroupCtrl, "Data Logging", NULL, 1, VAL_TAB_SHEET_FROM_LIST, chartInfo->dataLogTabCtrls)); 
    errChk( EasyTab_AddSheet(dialog, chartInfo->sheetGroupCtrl, "Save/Recall", NULL, 1, VAL_TAB_SHEET_FROM_LIST, chartInfo->configTabCtrls)); 

        /* set precisions */
    errChk( SetCtrlAttribute (dialog, chartInfo->frequencyNumeric, ATTR_PRECISION, 5));
    errChk( SetCtrlAttribute(dialog, chartInfo->actualFrequencyNumeric, ATTR_PRECISION, 5));

        /* Set data types */
    errChk( SetCtrlAttribute(dialog, chartInfo->historySizeNumeric, ATTR_DATA_TYPE, VAL_INTEGER));
    errChk( SetCtrlAttribute(dialog, chartInfo->frequencyNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dialog, chartInfo->actualFrequencyNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dialog, chartInfo->scanBackLogNumeric, ATTR_DATA_TYPE, VAL_INTEGER));


    errChk( SetCtrlAttribute (dialog, chartInfo->frequencyNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( SetCtrlAttribute (dialog, chartInfo->refreshNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( SetCtrlAttribute (dialog, chartInfo->historySizeNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( SetCtrlAttribute (dialog, chartInfo->refreshNumeric, ATTR_INCR_VALUE, 0.1));
    errChk( SetCtrlAttribute(dialog, chartInfo->historySizeNumeric, ATTR_MIN_VALUE, MIN_DAQ_CHART_HISTORY_BUFFER_SIZE));
    errChk( SetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_MIN_VALUE, DAQ_CHART_MIN_REFRESH_FREQUENCY));
    errChk( SetCtrlAttribute(dialog, chartInfo->refreshNumeric, ATTR_MAX_VALUE, DAQ_CHART_MAX_REFRESH_FREQUENCY));
    errChk( SetCtrlAttribute(dialog, chartInfo->frequencyNumeric, ATTR_MIN_VALUE, DAQ_CHART_MIN_SCAN_FREQUENCY));
    errChk( SetCtrlAttribute(dialog, chartInfo->frequencyNumeric, ATTR_MAX_VALUE, DAQ_CHART_MAX_SCAN_FREQUENCY));

    errChk( SetCtrlVal(dialog, chartInfo->actualFrequencyNumeric, chartInfo->scanFrequency));

        /* Install Control Callbacks */
    errChk( InstallCtrlCallback (dialog, chartInfo->frequencyNumeric,     DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->refreshNumeric,       DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->historySizeNumeric,   DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->useHardwareTimingCheckBox,        DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->refreshRing,          RefreshRingCallback,  (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->editColorsButton,     EditColorsCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->deleteAlarmButton,    EditAlarmsCtrlCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->newAlarmButton,       EditAlarmsCtrlCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->editAlarmButton,      EditAlarmsCtrlCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->alarmListBox,         EditAlarmsCtrlCallback, (void *)chartInfo));

    errChk( InstallCtrlCallback (dialog, chartInfo->logEventsCheckBox,      DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->fileFormatRing,         DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->startLoggingToggle,     DAQ_CtrlDialogCallbackNoChange, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->pauseLoggingToggle,     DAQ_CtrlDialogCallbackNoChange, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->startOnLoadCheckBox,    DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->frequencyNumeric,       DAQ_CtrlDialogCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->frequencyNumeric,       DAQ_CtrlDialogCallback, (void *)chartInfo));

    errChk( InstallCtrlCallback (dialog, chartInfo->nameLogFileButton,      NameLogFileBtnCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->advancedLoggingButton,  AdvancedLoggingBtnCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->convertLogFileButton,   ConvertLogFileBtnCallback, (void *)chartInfo));
    errChk( InstallCtrlCallback (dialog, chartInfo->textFormatButton,       TextFormatBtnCallback, (void *)chartInfo));

        /* Arrange the controls and size the popup */
    errChk( DAQ_Chart_FormatPopup(chartInfo));

    errChk( UpdateAlarmDlgCtrls(chartInfo));

    errChk( SetPanelPos(dialog, VAL_AUTO_CENTER, VAL_AUTO_CENTER));

Error:
    if (error < 0)
        {
        if (dialog > 0)
            DiscardPanel(dialog);
        chartInfo->dialog = 0;
        }

    SetWaitCursor(savedWaitCursor);
    return error;
}

/*****************************************************************************************/

    /* it is expected that the alarm passed will not yet be in the chart alarm list */
static int MakeAlarmNameUnique(DAQ_ChartInfo chartInfo, Alarm *alarm)
{
    int     error = 0;
    Alarm   *alarmPtr;
    char    *newName = 0;
    char    *baseName = 0;
    int     newNameLength;
    int     baseNameLength;
    int     index;
    int     newSuffixNumber = 2;
    char    newSuffixNumberBuf[32];

    while(1)
        {
        alarmPtr = AlarmPtr(chartInfo, alarm->name);
        if (!alarmPtr || alarmPtr == alarm)
            break;

        if (!baseName)  /* strip trailing digits off current name */
            {
            nullChk(baseName = StrDup(alarm->name));
            baseNameLength = strlen(baseName);
            for (index = baseNameLength; index > 0; index--)
                {
                if (isdigit(baseName[index - 1]))
                    baseName[index - 1] = 0;
                else
                    break;
                }
            }

        sprintf(newSuffixNumberBuf, "%d", newSuffixNumber++);
        newNameLength = strlen(newSuffixNumberBuf) + baseNameLength;
        nullChk( newName = (char *)malloc(newNameLength + 1));
        sprintf(newName, "%s%s", baseName, newSuffixNumberBuf);     

        free(alarm->name);
        alarm->name = newName;
        newName = 0;
        }

Error:
    free(newName);
    free(baseName);
    return 0;
}

/*****************************************************************************************/

static int CVICALLBACK EditAlarmsCtrlCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    DAQ_ChartInfo   chartInfo;
    int             error = 0;
    int             swallowedEvent = 0;
    int             ctrlIndex;
    Alarm           alarmCopy;
    int             disposeAlarm = FALSE;
    int             result;

    UNUSED(dialog);
    UNUSED(eventData1);
    UNUSED(eventData2);

    chartInfo = (DAQ_ChartInfo)callbackData;

    if (event == EVENT_COMMIT)
        {
        swallowedEvent = 1;   /* this dialog isn't known to the user's code, so we don't want him to get its events if he uses GetUserEvent() */

        if (control == chartInfo->deleteAlarmButton)
            {
            errChk( GetCtrlIndex(chartInfo->dialog, chartInfo->alarmListBox, &ctrlIndex));
            if (ctrlIndex >= 0)
                {
                DisposeAlarmByIndex(chartInfo, ctrlIndex+1);
                errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));
                }
            }
        else
        if (control == chartInfo->newAlarmButton)
            {
            if (chartInfo->numChannels + ListNumItems(chartInfo->alarmList) == 64)     /* check for too many strip chart traces */
                errChk( DAQ_CHART_ERROR_MAX_64_CHANNELS_AND_ALARMS);

            errChk( CopyAlarm(&gDefaultAlarm, &alarmCopy));
            disposeAlarm = TRUE;
            errChk( MakeAlarmNameUnique(chartInfo, &alarmCopy));
            errChk( result = EditAlarmDialog(chartInfo, "Create New Alarm", &alarmCopy));
            if (result != CANCELED)
                {
                errChk( Internal_DAQ_Chart_CreateAlarm(chartInfo->panel, chartInfo->baseCtrl, alarmCopy.name, alarmCopy.channel,
                    alarmCopy.limit, alarmCopy.lowerLimit, alarmCopy.deadband, alarmCopy.mode, alarmCopy.plot, alarmCopy.color));
                errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));
                }
            }
        else
        if (control == chartInfo->editAlarmButton || control == chartInfo->alarmListBox)
            {
            errChk( GetCtrlIndex(chartInfo->dialog, chartInfo->alarmListBox, &ctrlIndex));
            if (ctrlIndex >= 0)
                {
                Alarm   *alarmPtr;

                alarmPtr = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, ctrlIndex+1);
                errChk( CopyAlarm(alarmPtr, &alarmCopy));
                disposeAlarm = TRUE;
                errChk( result = EditAlarmDialog(chartInfo, "Edit Alarm", &alarmCopy));
                if (result == CHANGE_MADE)
                    {
                    errChk( ReplaceAlarmByIndex(chartInfo, ctrlIndex+1, &alarmCopy));
                    disposeAlarm = FALSE;
                    errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));
                    errChk( SetCtrlIndex(chartInfo->dialog, chartInfo->alarmListBox, ctrlIndex));
                    }
                }
            }
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    if (disposeAlarm)
        FreeAlarmResources(&alarmCopy);

    return swallowedEvent;
}

/*****************************************************************************************/

static int CVICALLBACK EditColorsCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    DAQ_ChartInfo   chartInfo;
    int             error = 0;

    UNUSED(dialog);
    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    chartInfo = (DAQ_ChartInfo)callbackData;

    if (event == EVENT_COMMIT)
        errChk( SelectChartColors(chartInfo));

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/*****************************************************************************************/

static int CVICALLBACK RefreshRingCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    DAQ_ChartInfo   chartInfo;
    int             item;
    int             top, left, height, width;

    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    chartInfo = (DAQ_ChartInfo)callbackData;

    switch (event)
        {
        case EVENT_LEFT_CLICK:
        case EVENT_LEFT_DOUBLE_CLICK:
            GetCtrlAttribute(chartInfo->dialog, chartInfo->refreshRing, ATTR_TOP, &top);
            GetCtrlAttribute(chartInfo->dialog, chartInfo->refreshRing, ATTR_LEFT, &left);
            GetCtrlAttribute(chartInfo->dialog, chartInfo->refreshRing, ATTR_HEIGHT, &height);
            GetCtrlAttribute(chartInfo->dialog, chartInfo->refreshRing, ATTR_WIDTH, &width);

            item = RunPopupMenu(chartInfo->refreshPopupMenuBar, chartInfo->refreshPopupMenu, dialog,
                top + height, left, top, left, height, width);


            if (item > 0 && ObjectExists(chartInfo))
                {
                void    *callbackData;
                double  frequency;

                GetMenuBarAttribute(chartInfo->refreshPopupMenuBar, item, ATTR_CALLBACK_DATA, &callbackData);
                switch ((int)callbackData)
                    {
                    case 1:
                        frequency = VAL_DAQ_CHART_REFRESH_VERY_FAST;
                        break;
                    case 2:
                        frequency = VAL_DAQ_CHART_REFRESH_FAST;
                        break;
                    default:
                    case 3:
                        frequency = VAL_DAQ_CHART_REFRESH_NORMAL;
                        break;
                    case 4:
                        frequency = VAL_DAQ_CHART_REFRESH_SLOW;
                        break;
                    case 5:
                        frequency = VAL_DAQ_CHART_REFRESH_VERY_SLOW;
                        break;
                    }

                SetCtrlVal(chartInfo->dialog, chartInfo->refreshNumeric, frequency);
                return DAQ_CtrlDialogCallback(chartInfo->dialog, chartInfo->refreshNumeric,
                            EVENT_VAL_CHANGED, (void *)chartInfo, 0,0);

                }
        break;
        }

    return 1;
}

/*****************************************************************************************/

        /* use same plot styles as the DAQ Chart does: */

static void GetTraceAttributes(int panel, int stripChart, int trace, int *color,
        int *pointStyle, int *plotStyle, int *lineStyle)
{
    GetTraceAttribute (panel, stripChart, trace, ATTR_TRACE_COLOR, color);
    GetTraceAttribute (panel, stripChart, trace, ATTR_TRACE_POINT_STYLE, pointStyle);
    GetTraceAttribute (panel, stripChart, trace, ATTR_PLOT_STYLE, plotStyle);
    GetTraceAttribute (panel, stripChart, trace, ATTR_LINE_STYLE, lineStyle);
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_PlotHistory(int panel, int daqChartCtrl, int graphPanel, int graph, int plotVsTime,
    int numScansToPlot, int plotAlarmLimits, int deleteCurrentPlots, int remove, int matchYAxisRange)
{
    int     error = 0;
    int     numScans;
    int     index;
    int     refreshGraphSavedValue;
    int     channelsPerScan;
    double  *samples = 0;
    double  *times = 0;
    int     color;
    int     pointStyle;
    int     plotStyle;
    int     lineStyle;
    int     numScansNotToPlot;

    errChk( Internal_DAQ_Chart_GetHistoryCopy(panel, daqChartCtrl, 0, &numScans, &channelsPerScan, 1, &samples, &times, remove));

    if (numScansToPlot < numScans && numScansToPlot > 0)
        {
        numScansNotToPlot = numScans - numScansToPlot;
        }
    else
        {
        numScansToPlot = numScans;
        numScansNotToPlot = 0;
        }

    if (deleteCurrentPlots)
        DeleteGraphPlot(graphPanel, graph, -1, VAL_DELAYED_DRAW);

    GetCtrlAttribute (graphPanel, graph, ATTR_REFRESH_GRAPH, &refreshGraphSavedValue);
    SetCtrlAttribute (graphPanel, graph, ATTR_REFRESH_GRAPH, 0);

    if (matchYAxisRange)
        {
        double  yMin, yMax;

        errChk( GetAxisRange (panel, daqChartCtrl, 0, 0, 0, 0, &yMin, &yMax));
        errChk( SetAxisRange (graphPanel, graph, VAL_NO_CHANGE, 0, 0, VAL_MANUAL, yMin, yMax));
        }

    if (numScansToPlot > 0)
        {
        for (index = 0; index < channelsPerScan; index++)
            {
                /* use same plot styles as the DAQ Chart does: */
            GetTraceAttributes(panel, daqChartCtrl, index+1, &color, &pointStyle, &plotStyle, &lineStyle);

            if (plotVsTime)
                PlotXY (graphPanel, graph, times + numScansNotToPlot, &samples[index*numScans + numScansNotToPlot],
                    numScansToPlot, VAL_DOUBLE, VAL_DOUBLE, plotStyle, pointStyle,
                    lineStyle, 1, color);
            else
                PlotY (graphPanel, graph, &samples[index*numScans + numScansNotToPlot], numScansToPlot, VAL_DOUBLE, plotStyle,
                       pointStyle, lineStyle, 1, color);
            }


        if (plotAlarmLimits)
            {
            int             numAlarms;
            DAQ_ChartInfo   chartInfo = 0;
            int             alarmTraceIndex = 1;
            Alarm           *alarm;
            double          x[2];
            double          y[2];

            errChk( GetDAQCtrlInfo(panel, daqChartCtrl, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

            numAlarms = ListNumItems(chartInfo->alarmList);

           if (plotVsTime)
                {
                x[0] = times[numScansNotToPlot];
                x[1] = times[numScans-1];
                }
            else
                {
                x[0] = 0;
                x[1] = numScansToPlot-1;
                }

            for (index = 1; index <= numAlarms; index++)
                {
                alarm = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);
                if (alarm->plot)
                    {
                        /* use same plot styles as the DAQ Chart does: */
                    GetTraceAttributes(panel, daqChartCtrl, chartInfo->numChannels + alarmTraceIndex, &color, &pointStyle, &plotStyle, &lineStyle);

                    y[0] = y[1] = alarm->limit;
                    PlotXY (graphPanel, graph, x, y, 2, VAL_DOUBLE, VAL_DOUBLE, plotStyle, pointStyle, lineStyle, 1, color);

                    if (IsRangeMode(alarm->mode))
                        {
                        y[0] = y[1] = alarm->lowerLimit;
                        PlotXY (graphPanel, graph, x, y, 2, VAL_DOUBLE, VAL_DOUBLE, plotStyle, pointStyle, lineStyle, 1, color);
                        }

                    alarmTraceIndex++;
                    }
                }
            }
    }

    SetCtrlAttribute (graphPanel, graph, ATTR_REFRESH_GRAPH, refreshGraphSavedValue);

Error:
    free(samples);
    free(times);

    return error;
}

/*****************************************************************************************/

static int GetIndexFromChannel(DAQ_ChartInfo chartInfo, char *channel)
{
    int             error = 0;
    unsigned long   numChannels;
    long            indexArray[1];
    int             savedBOLE = SetBOLE(FALSE);     /* disable break on library errors */

    errChk( GetNumChannelsNoBOLE (chartInfo->device, channel, chartInfo->ioDirection, &numChannels));
    if (numChannels != 1)
        errChk( -10007);

    errChk( GetChannelIndices (chartInfo->device, chartInfo->channelString, channel, chartInfo->ioDirection, indexArray));

Error:
    SetBOLE(savedBOLE);
    if (error < 0)
        return error;
    else
        return (int)indexArray[0];
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_CreateAlarm(int panel, int daqChart, char *alarmName, char *channel,
            double limit, double lowerLimit, double deadband, int mode, int plotTheLimit, int traceColor)
{
    return Internal_DAQ_Chart_CreateAlarm(panel, daqChart, alarmName, channel,
                                          limit, lowerLimit, deadband, mode,
                                          plotTheLimit, traceColor);
}

static int Internal_DAQ_Chart_CreateAlarm(int panel, int daqChart, char *alarmName, char *channel,
            double limit, double lowerLimit, double deadband, int mode, int plotTheLimit, int traceColor)
{
    int             error = 0;
    Alarm           alarm;
    DAQ_ChartInfo   chartInfo = 0;
    int             alarmAdded = 0;
    int             savedBOLE = SetBOLE (FALSE);    /* disable break on library errors */

    memset(&alarm, 0, sizeof(alarm));

    if (!alarmName || !channel)
        errChk( UIENullPointerPassed);

    if (IsRangeMode(mode) < 0)
        errChk( UIEValueIsInvalidOrOutOfRange);

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

    if (AlarmPtr(chartInfo, alarmName) != 0)
        errChk( ToolErr_DuplicateItemOrValue);

    if (chartInfo->numChannels + ListNumItems(chartInfo->alarmList) > 64)      /* check for too many strip chart traces */
        errChk( DAQ_CHART_ERROR_MAX_64_CHANNELS_AND_ALARMS);

    errChk( AllocateAlarmResources(&alarm, alarmName, channel, limit, lowerLimit, deadband, mode, plotTheLimit, traceColor, gDefaultAlarm.logAction));

    nullChk( ListInsertItem (chartInfo->alarmList, &alarm, END_OF_LIST));
    alarmAdded = TRUE;

    errChk( UpdateNumTraces(chartInfo));

Error:
    if (error < 0)
        {
        if (alarmAdded)
            ListRemoveItem(chartInfo->alarmList, 0, END_OF_LIST);

        FreeAlarmResources(&alarm);
        }

    SetBOLE(savedBOLE);
    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_DiscardAlarm(int panel, int daqChart, char *alarmName)
{
    int             error = 0;
    Alarm           alarm;
    DAQ_ChartInfo   chartInfo = 0;
    int             index, numAlarms;
    int             alarmDeleted = FALSE;

    alarm.name = 0;     /* bypass UP problem */
    alarm.channel = 0;  /* bypass UP problem */

    if (!alarmName)
        errChk( UIENullPointerPassed);

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

    numAlarms = ListNumItems(chartInfo->alarmList);

    for (index = numAlarms; index > 0; index--)
        {
        ListGetItem(chartInfo->alarmList, &alarm, index);

        if (strcmp(alarmName, alarm.name) == 0)
            {
            DisposeAlarmByIndex(chartInfo, index);
            alarmDeleted = TRUE;
            }
        }

Error:
    if(error < 0)
        return error;
    else
        return alarmDeleted;
}

/*****************************************************************************************/

static int AllocateAlarmResources(Alarm *alarm, char *alarmName, char *channel, double limit, double lowerLimit, double deadband,
        int mode, int plotTheLimit, int traceColor, int logAction)
{
    int error = 0;

    alarm->name = 0;
    alarm->channel = 0;

    nullChk( alarm->name = StrDup(alarmName));
    nullChk( alarm->channel = StrDup(channel));

    alarm->limit = limit;
    alarm->lowerLimit = lowerLimit;
    alarm->deadband = deadband;
    alarm->mode = mode;
    alarm->plot = plotTheLimit;
    alarm->on = FALSE;
    alarm->color = traceColor;
    alarm->valid = TRUE;
    alarm->logAction = logAction;

Error:
    if (error < 0)
        FreeAlarmResources(alarm);

    return error;
}

/*****************************************************************************************/

static int ReplaceAlarmByIndex(DAQ_ChartInfo chartInfo, int indexOfAlarmToReplace, Alarm *replacementAlarm)
{
    int error = 0;
    int savedBOLE = SetBOLE (FALSE);    /* disable break on library errors */

    if (indexOfAlarmToReplace <= ListNumItems(chartInfo->alarmList))
        {
        nullChk( ListInsertItem (chartInfo->alarmList, replacementAlarm, indexOfAlarmToReplace));
        DisposeAlarmByIndex(chartInfo, indexOfAlarmToReplace + 1);
        }

    errChk( UpdateNumTraces(chartInfo));

Error:
    SetBOLE(savedBOLE);
    return error;
}

/*****************************************************************************************/

    /*  copies the alarm pointed to by src to the memory pointed to by dest, duplicating any attached resources */
static int CopyAlarm(Alarm *src, Alarm  *dest)
{
    int error = 0;

    memset(dest, 0, sizeof(*dest));
    nullChk( dest->name = StrDup(src->name));
    nullChk( dest->channel = StrDup(src->channel));
    dest->limit = src->limit;
    dest->lowerLimit = src->lowerLimit;
    dest->deadband = src->deadband;
    dest->mode = src->mode;
    dest->plot = src->plot;
    dest->channelIndex = src->channelIndex;
    dest->color = src->color;
    dest->on = src->on;
    dest->valid = src->valid;
    dest->logAction = src->logAction;

Error:
    if (error < 0)
        FreeAlarmResources(dest);
    return error;
}

/*****************************************************************************************/

    /*  returns 0 if the alarm doesn't exist */
static Alarm *AlarmPtr(DAQ_ChartInfo chartInfo, char *alarmName)
{
    Alarm   *alarmPtr;
    int     index, numAlarms;

    if (!alarmName)
        return 0;

    numAlarms = ListNumItems(chartInfo->alarmList);

    for (index = numAlarms; index > 0; index--)
        {
        alarmPtr = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);

        if (strcmp(alarmName, alarmPtr->name) == 0)
            return alarmPtr;
        }

    return 0;
}

/*****************************************************************************************/

int CVIFUNC_C DAQ_Chart_SetAlarmAttribute(int panel, int chart, char *alarmName, int attribute, ...)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = 0;
    va_list         parmInfo;

    va_start(parmInfo, attribute);
    errChk( GetDAQCtrlInfo(panel, chart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));
    errChk( Internal_SetAlarmAttributeFromParmInfo(chartInfo, alarmName, attribute, parmInfo));

Error:
    va_end(parmInfo);
    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_SetAlarmAttributeFromParmInfo(int panel, int daqChart, char alarmName[], int attribute, va_list parmInfo)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = 0;

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));
    errChk( Internal_SetAlarmAttributeFromParmInfo(chartInfo, alarmName, attribute, parmInfo));

Error:
    return error;
}

/*****************************************************************************************/

static int Internal_SetAlarmAttributeFromParmInfo(DAQ_ChartInfo chartInfo, char *alarmName, int attribute, va_list parmInfo)
{
    int     error = 0;
    Alarm   *alarmPtr;
    int     intValue;
    char    *tempString = 0;
    char    *stringPtr;

    if (!alarmName)
        errChk( UIENullPointerPassed);

    alarmPtr = AlarmPtr(chartInfo, alarmName);
    if (!alarmPtr)
        errChk( UIEValueIsInvalidOrOutOfRange);

    switch (attribute)
        {
        case ATTR_ALARM_LIMIT:
            alarmPtr->limit = va_arg(parmInfo, double);
            break;
        case ATTR_ALARM_LOWER_LIMIT:
            alarmPtr->lowerLimit = va_arg(parmInfo, double);
            break;
        case ATTR_ALARM_DEADBAND:
            alarmPtr->deadband = va_arg(parmInfo, double);
            break;
        case ATTR_ALARM_MODE:
            intValue = va_arg(parmInfo, int);
            switch (intValue)
                {
                case VAL_ALARM_ABOVE_LIMIT:
                case VAL_ALARM_BELOW_LIMIT:
                case VAL_ALARM_IN_RANGE:
                case VAL_ALARM_OUT_OF_RANGE:
                    alarmPtr->mode = intValue;
                    break;
                default:
                    errChk( UIEValueIsInvalidOrOutOfRange);
                }
            break;
        case ATTR_ALARM_PLOT:
            intValue = !!va_arg(parmInfo, int);
            if (!!alarmPtr->plot != intValue)
                {
                alarmPtr->plot = intValue;
                errChk( UpdateNumTraces(chartInfo));
                }
            break;
        case ATTR_ALARM_TRACE_COLOR:
            alarmPtr->color = va_arg(parmInfo, int);
            errChk( SetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, alarmPtr->traceIndex, ATTR_TRACE_COLOR, alarmPtr->color));
            break;
        case ATTR_ALARM_NAME:
            stringPtr = va_arg(parmInfo, char *);
            if (!stringPtr)
                errChk( UIENullPointerPassed);

            if (strcmp(stringPtr, alarmPtr->name) != 0)
                {
                if (AlarmPtr(chartInfo, stringPtr) != 0)
                    errChk( ToolErr_DuplicateItemOrValue);

                nullChk(tempString = StrDup(stringPtr));
                free(alarmPtr->name);
                alarmPtr->name = tempString;
                tempString = 0;
                }
            break;
        case ATTR_ALARM_CHANNEL:
            stringPtr = va_arg(parmInfo, char *);
            if (!stringPtr)
                errChk( UIENullPointerPassed);

            if (strcmp(stringPtr, alarmPtr->channel) != 0)
                {
                nullChk(tempString = StrDup(stringPtr));
                free(alarmPtr->channel);
                alarmPtr->channel = tempString;
                tempString = 0;
                errChk( UpdateNumTraces(chartInfo));
                WarnIfAlarmsHaveInvalidChannels(chartInfo);
                }
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC DAQ_Chart_GetAlarmAttribute(int panel, int chart, char *alarmName, int attribute, void *value)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = 0;

    errChk( GetDAQCtrlInfo(panel, chart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));
    errChk( Internal_GetAlarmAttribute(chartInfo, alarmName, attribute, value));

Error:
    return error;
}

/*****************************************************************************************/

static int Internal_GetAlarmAttribute(DAQ_ChartInfo chartInfo, char *alarmName, int attribute, void *value)
{
    int     error = 0;
    Alarm   *alarmPtr;

    if (!alarmName || !value)
        errChk( UIENullPointerPassed);

    alarmPtr = AlarmPtr(chartInfo, alarmName);
    if (!alarmPtr)
        errChk( UIEValueIsInvalidOrOutOfRange);

    switch (attribute)
        {
        case ATTR_ALARM_LIMIT:
            *(double *)value = alarmPtr->limit;
            break;
        case ATTR_ALARM_LOWER_LIMIT:
            *(double *)value = alarmPtr->lowerLimit;
            break;
        case ATTR_ALARM_DEADBAND:
            *(double *)value = alarmPtr->deadband;
            break;
        case ATTR_ALARM_MODE:
            *(int *)value = alarmPtr->mode;
            break;
        case ATTR_ALARM_PLOT:
            *(int *)value = alarmPtr->plot;
            break;
        case ATTR_ALARM_TRACE_COLOR:
            *(int *)value = alarmPtr->color;
            break;
        case ATTR_ALARM_NAME:
            strcpy((char *)value, alarmPtr->name);
            break;
        case ATTR_ALARM_CHANNEL:
            strcpy((char *)value, alarmPtr->channel);
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }

Error:
    return error;
}

/*****************************************************************************************/

static void FreeAlarmResources(Alarm *alarm)
{
    free(alarm->name);
    free(alarm->channel);
    alarm->name = 0;
    alarm->channel = 0;
}

/*****************************************************************************************/

static void DisposeAllChartAlarms(DAQ_ChartInfo chartInfo)
{
    int index;
    int numItems;

    if (chartInfo->alarmList)
        {
        numItems = ListNumItems(chartInfo->alarmList);
        for (index = numItems; index > 0; index--)
             DisposeAlarmByIndex(chartInfo, index);
        }
}

/*****************************************************************************************/

static void FreeAlarmList(ListType alarmList)
{
    int     index;
    int     numItems;
    Alarm   *alarmPtr;

    if (alarmList)
        {
        numItems = ListNumItems(alarmList);
        for (index = 1; index <= numItems; index++)
            {
            alarmPtr = (Alarm *)ListGetPtrToItem(alarmList, index);
            FreeAlarmResources(alarmPtr);
            }

        ListDispose(alarmList);
        }
}

/*****************************************************************************************/

static void DisposeSavedAlarmList(DAQ_ChartInfo chartInfo)
{
    FreeAlarmList(chartInfo->savedAlarmList);
    chartInfo->savedAlarmList = 0;
}

/*****************************************************************************************/

    /*  this is the only function that should actually delete an alarm owned by the chart */
static void DisposeAlarmByIndex(DAQ_ChartInfo chartInfo, int index)
{
    Alarm   alarm;

    alarm.name = 0;     /* bypass UP problem */
    alarm.channel = 0;  /* bypass UP problem */

    ListRemoveItem(chartInfo->alarmList, &alarm, index);
    FreeAlarmResources(&alarm);

    UpdateNumTraces(chartInfo);
}

/*******************************************************/

static void CVICALLBACK DAQ_Chart_ApplyColorChanges(ListType colorList, void *callbackData, int colorListIndex)
{
    DAQ_ChartInfo       chartInfo = (DAQ_ChartInfo)callbackData;
    int                 index;
    int                 numItems;
    ColorChangeEntry    *entry;
    int                 numAlarms;
    Alarm               *alarm;
    int                 alarmIndex = 1;
    int                 traceIndex = 1;

    UNUSED(colorListIndex);

    if (ObjectExists(chartInfo))
        {
        numItems = ListNumItems(colorList);
        numAlarms = ListNumItems(chartInfo->alarmList);

        for (index = 1; index <= numItems; index++)
            {
            entry = (ColorChangeEntry*)ListGetPtrToItem(colorList, index);
            SetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, traceIndex, ATTR_TRACE_COLOR, entry->color);
            traceIndex++;

SkipNonPlottedAlarm:
            if (index > chartInfo->numChannels && alarmIndex <= numAlarms)
                {
                alarm = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, alarmIndex);
                if (!alarm->plot)
                    {
                    alarmIndex++;
                    goto SkipNonPlottedAlarm;
                    }
                alarm->color = entry->color;
                alarmIndex++;
                if (IsRangeMode(alarm->mode))
                    {
                    SetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, traceIndex, ATTR_TRACE_COLOR, entry->color);
                    traceIndex++;
                    }
                }
            }
        }
}

/*******************************************************/

static ListType CreateChartColorList(DAQ_ChartInfo chartInfo)
{
    int                 error = 0;
    ListType            colorList = 0;
    ColorChangeEntry    entry;
    int                 index;
    char                name[200];
    char                channelNameBuf[MAX_CHANNEL_NAME_LENGTH + 1];
    Alarm               *alarm;
    int                 numAlarms;
    int                 traceNumber = 1;

    nullChk( colorList = ListCreate(sizeof(entry)));

    for (index = 1; index <= chartInfo->numChannels; index++)
        {
        int savedBole = SetBOLE(FALSE);     /* in case GetChannelNameFromIndex() fails */

        if (GetChannelNameFromIndex (chartInfo->device,chartInfo->channelString, index - 1, chartInfo->ioDirection, channelNameBuf) < 0)
            sprintf(name, "Trace Number %d", traceNumber);
        else
            sprintf(name, "Channel %s", channelNameBuf);
        SetBOLE(savedBole);

        nullChk(entry.name = StrDup(name));
        GetTraceAttribute (chartInfo->panel, chartInfo->baseCtrl, traceNumber, ATTR_TRACE_COLOR, &entry.color);
        entry.defaultColor = entry.color;
        nullChk( ListInsertItem(colorList, &entry, END_OF_LIST));
        traceNumber++;
        }

    numAlarms = ListNumItems(chartInfo->alarmList);

    for (index = 1; index <= numAlarms; index++)
        {
        alarm = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);
        if (alarm->plot)
            {
            if (IsRangeMode(alarm->mode))
                sprintf(name, "Alarm '%s', channel=%s%s, range=%.2f to %.2f", alarm->name, alarm->channel, alarm->valid ? "" : " (not a valid channel)", alarm->lowerLimit, alarm->limit);
            else
                sprintf(name, "Alarm '%s', channel=%s%s, limit=%.2f", alarm->name, alarm->channel, alarm->valid ? "" : " (not a valid channel)", alarm->limit);
            nullChk(entry.name = StrDup(name));
            entry.color = alarm->color;
            entry.defaultColor = entry.color;
            nullChk( ListInsertItem(colorList, &entry, END_OF_LIST));
            }
        }
Error:
    if (error < 0)
        {
        DisposeChartColorList(colorList);
        return 0;
        }
    else
        return colorList;
}

/*******************************************************/

static void DisposeChartColorList(ListType colorList)
{
    if (colorList)
        {
        int                 index;
        int                 numItems;
        ColorChangeEntry    *entryPtr;

        numItems = ListNumItems(colorList);

        for (index = 1; index <= numItems; index++)
            {
            entryPtr = (ColorChangeEntry *)ListGetPtrToItem(colorList, index);
            free(entryPtr->name);
            }

        ListDispose(colorList);
        }
}

/**********************************************************************/

static int SelectChartColors(DAQ_ChartInfo chartInfo)
{
    int         error = 0;
    int         result;
    ListType    colorList = 0;
    int         savedWaitCursor = GetWaitCursorState ();

    SetWaitCursor (1);

    nullChk( colorList = CreateChartColorList(chartInfo));
    SetWaitCursor(0);
    errChk( result = ColorChangePopup("Select Trace Colors", colorList, FALSE, DAQ_Chart_ApplyColorChanges, (void*)chartInfo));
    if (result > 0 && ObjectExists(chartInfo))
        errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));

Error:
    SetWaitCursor(savedWaitCursor);
    DisposeChartColorList(colorList);
    return error;
}

/**********************************************************************/

    /*  returns positive if in range mode, 0 if in level mode, and negative if bad mode parameter */
static int IsRangeMode(int limitMode)
{
    switch (limitMode)
        {
        case VAL_ALARM_ABOVE_LIMIT:
        case VAL_ALARM_BELOW_LIMIT:
            return 0;
        case VAL_ALARM_IN_RANGE:
        case VAL_ALARM_OUT_OF_RANGE:
            return 1;
        default:
            return -1;
        }
}

/**********************************************************************/

static int  AdjustAlarmDlgForLimitMode(int dlg, int limitRing, int limitNumeric, int lowerLimitNumeric)
{
    int error;
    int limitMode;

    errChk( GetCtrlVal(dlg, limitRing, &limitMode));

    switch (limitMode)
        {
        case VAL_ALARM_ABOVE_LIMIT:
        case VAL_ALARM_BELOW_LIMIT:
            errChk( SetCtrlAttribute(dlg, limitNumeric, ATTR_LABEL_TEXT, "Limit (Volts):"));
            errChk( SetCtrlAttribute(dlg, lowerLimitNumeric, ATTR_DIMMED, 1));
            break;
        case VAL_ALARM_IN_RANGE:
        case VAL_ALARM_OUT_OF_RANGE:
            errChk( SetCtrlAttribute(dlg, limitNumeric, ATTR_LABEL_TEXT, "Upper Limit (Volts):"));
            errChk( SetCtrlAttribute(dlg, lowerLimitNumeric, ATTR_DIMMED, 0));
            break;
        }
Error:
    return error;
}

/************************************************************************/

    /*  returns 1 if limits are valid, 0 if they are invalid, or a negative error code */
static int  CheckAlarmDlgLimitValidity(int dlg, int limitRing, int limitNumeric, int lowerLimitNumeric, int deadBandNumeric)
{
    int     error;
    int     limitMode;
    double  upperLimit, lowerLimit, deadBand;

    errChk( GetCtrlVal(dlg, limitRing, &limitMode));

    if (IsRangeMode(limitMode))
        {
        errChk( GetCtrlVal(dlg, limitNumeric, &upperLimit));
        errChk( GetCtrlVal(dlg, lowerLimitNumeric, &lowerLimit));
        errChk( GetCtrlVal(dlg, deadBandNumeric, &deadBand));

        if (upperLimit <= lowerLimit)
            {
            MessagePopup("Alarm Specification Error", "The lower limit must be less than the upper limit.");
            return 0;
            }

        if (upperLimit - lowerLimit <= deadBand * 2)
            {
            MessagePopup("Alarm Specification Error", 
                "The range is too small for the specified deadband width.\n\n"
                "To accomodate the upper and lower deadbands, the difference between the\n"
                "upper and lower limits must be greater than twice the width of the deadband.");
            return 0;
            }
        }

Error:
    if (error < 0)
        return error;
    else
        return 1;
}

/************************************************************************/

    /*  returns a ResponseCode or a negative error code */
static EditAlarmDialog(DAQ_ChartInfo chartInfo, char *dlgTitle, Alarm *alarm)
{
    int error = 0;
    int dlg = 0;
    int ok, cancel, nameString, limitRing, plotCheckBox, channelString, traceColorNumeric;
    int limitNumeric, lowerLimitNumeric, deadbandNumeric;
    int changed = FALSE;
    int hMargin = DLG_MARGIN;
    int vMargin = DLG_MARGIN;
    int top;
    int maxLabelWidth;
    int ctrlLeft;
    int ctrlHeight;
    int ctrlWidth;
    int contentWidth;
    int done = FALSE;
    int evtPanel, evtCtrl;
    int btnWidth;
    int labelToCtrlGap = DLG_LABEL_TO_CTRL_MIN_H_GAP + 50;  /* wider for this small dialog */
    int channelStringWidth;
    int canceled = FALSE;
    char *newName = 0, *newChannel = 0;
    int limitRingWidth;
    int limitRingLabelWidth;
    int limitRingLabelEncroachment; /* number of pixels the limit ring uses out of the label area */

    errChk(dlg = NewPanel(0, dlgTitle, 0,0, 10,10));
    SetPanelAttribute(dlg, ATTR_CONFORM_TO_SYSTEM, TRUE);

    errChk( nameString =        NewCtrl (dlg, CTRL_STRING, "Alarm Name:", 0, 0));
    errChk( channelString =     NewCtrl (dlg, CTRL_STRING, "Channel String:", 0, 0));
    errChk( limitNumeric =      NewCtrl (dlg, CTRL_NUMERIC, "Limit (Volts):", 0, 0));
    errChk( lowerLimitNumeric = NewCtrl (dlg, CTRL_NUMERIC, "Lower Limit (Volts):", 0, 0));
    errChk( limitRing =         NewCtrl (dlg, CTRL_RING, "Alarm Mode:", 0, hMargin));
    errChk( deadbandNumeric =   NewCtrl (dlg, CTRL_NUMERIC, "Deadband Width (Volts):", 0, 0));
    errChk( traceColorNumeric = NewCtrl (dlg, CTRL_COLOR_NUMERIC, "Trace Color:", 0, 0));
    errChk( plotCheckBox =      NewCtrl (dlg, CTRL_CHECK_BOX, "Plot Alarm", 0, hMargin));
    errChk( ok =                NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "__OK", 0, hMargin));
    errChk( cancel =            NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "  __Cancel  ", 0, hMargin));

    errChk( SetCtrlAttribute(dlg, limitNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dlg, lowerLimitNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dlg, deadbandNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dlg, deadbandNumeric, ATTR_MIN_VALUE, 0.0));

    errChk( SetCtrlAttribute(dlg, nameString, ATTR_MAX_ENTRY_LENGTH, DAQ_CHART_MAX_ALARM_NAME_LENGTH));

    errChk( InsertListItem (dlg, limitRing, -1, "Trigger Above Limit", VAL_ALARM_ABOVE_LIMIT));
    errChk( InsertListItem (dlg, limitRing, -1, "Trigger Below Limit", VAL_ALARM_BELOW_LIMIT));
    errChk( InsertListItem (dlg, limitRing, -1, "Trigger In Range", VAL_ALARM_IN_RANGE));
    errChk( InsertListItem (dlg, limitRing, -1, "Trigger Out Of Range", VAL_ALARM_OUT_OF_RANGE));

    errChk( SizeRingCtrlToText(dlg, limitRing));
    errChk( GetCtrlAttribute(dlg, limitRing, ATTR_WIDTH, &limitRingWidth));
    errChk( GetCtrlAttribute(dlg, limitRing, ATTR_LABEL_WIDTH, &limitRingLabelWidth));

        /* find the largest label */
    errChk( AttributeMaxAndMin(dlg, ATTR_LABEL_WIDTH, 0, &maxLabelWidth, 0, 0, 0,
        nameString, channelString, limitNumeric, deadbandNumeric, traceColorNumeric, 0));

    errChk( GetCtrlAttribute(dlg, nameString, ATTR_WIDTH, &ctrlWidth));
    limitRingLabelEncroachment = limitRingWidth - ctrlWidth;

    maxLabelWidth = Max(maxLabelWidth, limitRingLabelWidth + limitRingLabelEncroachment);
    ctrlLeft = hMargin + maxLabelWidth + labelToCtrlGap;
    contentWidth =  maxLabelWidth + labelToCtrlGap + ctrlWidth;
    top = vMargin;

    errChk( PutLabelsOnLeft (dlg, labelToCtrlGap, 0, nameString, channelString, limitNumeric, lowerLimitNumeric, limitRing, deadbandNumeric, traceColorNumeric, 0));
    errChk( SetAttributeForCtrls (dlg, ATTR_LEFT, ctrlLeft, 0, nameString, channelString, limitNumeric, lowerLimitNumeric, deadbandNumeric, traceColorNumeric, 0));
    errChk( SetCtrlAttribute(dlg, limitRing, ATTR_LEFT, ctrlLeft - limitRingLabelEncroachment));

    errChk( SetAttributeForCtrls (dlg, ATTR_LABEL_LEFT, hMargin, 0, nameString, channelString, limitNumeric, lowerLimitNumeric, limitRing, deadbandNumeric, traceColorNumeric, 0));

    errChk( DistributeCtrls (dlg, VAL_TB_VERTICAL_DISTRIBUTION, VAL_TB_FIXED_GAP_SPACING, &top, DLG_CTRL_V_GAP, 0,
        nameString, channelString, limitNumeric, lowerLimitNumeric, limitRing, deadbandNumeric, traceColorNumeric, plotCheckBox, 0));

    ExpandCtrlToMeetLabel(dlg, channelString, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);
    GetCtrlAttribute(dlg, channelString, ATTR_WIDTH, &channelStringWidth);
    ExpandCtrlToMeetLabel(dlg, nameString, DLG_LABEL_TO_CTRL_MIN_H_GAP, channelStringWidth);

        /* make ok and cancel the same width */
    errChk( SetCtrlsToAttributeExtreme(dlg, ATTR_WIDTH, VAL_TB_MAX, 0, ok, cancel, 0));

        /* position ok and cancel vertically */
    top += DLG_CTRL_V_GAP * 2;
    errChk( SetAttributeForCtrls (dlg, ATTR_TOP, top, 0, ok, cancel, 0));

        /* position the ok and cancel horizontally */
    GetCtrlAttribute(dlg, ok, ATTR_WIDTH, &btnWidth);
    SetCtrlAttribute(dlg, cancel, ATTR_LEFT, hMargin + contentWidth - btnWidth);
    SetCtrlAttribute(dlg, ok, ATTR_LEFT, hMargin + contentWidth - 2 * btnWidth - DLG_BTN_H_GAP);

    errChk( GetCtrlAttribute(dlg, ok, ATTR_HEIGHT, &ctrlHeight));
    top += ctrlHeight + DLG_CTRL_V_GAP;

    errChk( SetPanelAttribute(dlg, ATTR_WIDTH, contentWidth + 2 * hMargin));
    errChk( SetPanelAttribute(dlg, ATTR_HEIGHT, top));

    errChk( SetCommonDialogShortcutKeys(dlg, ok, cancel, cancel));
    errChk( SetPanelPos(dlg, VAL_AUTO_CENTER, VAL_AUTO_CENTER));

    errChk( SetCtrlAttribute (dlg, nameString, ATTR_MAX_ENTRY_LENGTH, 50));
    errChk( SetCtrlAttribute (dlg, channelString, ATTR_MAX_ENTRY_LENGTH, 255));

    errChk( SetCtrlVal(dlg, nameString, alarm->name));
    errChk( SetCtrlVal(dlg, channelString, alarm->channel));
    errChk( SetCtrlVal(dlg, limitNumeric, alarm->limit));
    errChk( SetCtrlVal(dlg, lowerLimitNumeric, alarm->lowerLimit));
    errChk( SetCtrlVal(dlg, deadbandNumeric, alarm->deadband));
    errChk( SetCtrlVal(dlg, limitRing, alarm->mode));
    errChk( SetCtrlVal(dlg, traceColorNumeric, alarm->color));
    errChk( SetCtrlVal(dlg, plotCheckBox, alarm->plot));

    errChk( AdjustAlarmDlgForLimitMode(dlg, limitRing, limitNumeric, lowerLimitNumeric));

    InstallPopup(dlg);

    while (!done)
        {
        GetUserEvent(0, &evtPanel, &evtCtrl);

        if (!ObjectExists(chartInfo))
            {
            done = TRUE;
            canceled = TRUE;
            }
        else
        if (evtPanel == dlg)
            {
            if  (evtCtrl == cancel)
                {
                done = TRUE;
                canceled = TRUE;
                }
            else
            if (evtCtrl == ok)
                {
                errChk( done = CheckAlarmDlgLimitValidity(dlg, limitRing, limitNumeric, lowerLimitNumeric, deadbandNumeric));
                }
            else
            if (evtCtrl == limitRing)
                {
                errChk( AdjustAlarmDlgForLimitMode(dlg, limitRing, limitNumeric, lowerLimitNumeric));
                changed = TRUE;
                }
            else
            if (evtCtrl == nameString || evtCtrl == channelString || evtCtrl == limitNumeric || evtCtrl == lowerLimitNumeric
                    || evtCtrl == deadbandNumeric || evtCtrl == traceColorNumeric || evtCtrl == plotCheckBox)
                {
                changed = TRUE;
                }
            }
        }

    if (!canceled)
        {
        char    nameBuf[DAQ_CHART_MAX_ALARM_NAME_LENGTH + 1];
        char    channelBuf[MAX_CHANNEL_STRING_LEN + 1];

        errChk( GetCtrlVal(dlg, nameString, nameBuf));
        errChk( GetCtrlVal(dlg, channelString, channelBuf));
        errChk( GetCtrlVal(dlg, limitNumeric, &alarm->limit));
        errChk( GetCtrlVal(dlg, lowerLimitNumeric, &alarm->lowerLimit));
        errChk( GetCtrlVal(dlg, deadbandNumeric, &alarm->deadband));
        errChk( GetCtrlVal(dlg, limitRing, &alarm->mode));
        errChk( GetCtrlVal(dlg, traceColorNumeric, &alarm->color));
        errChk( GetCtrlVal(dlg, plotCheckBox, &alarm->plot));

        nullChk( newName = StrDup(nameBuf));
        nullChk( newChannel = StrDup(channelBuf));

        free(alarm->name);
        free(alarm->channel);
        alarm->name = newName;
        alarm->channel = newChannel;
        newChannel = 0;
        newName = 0;
        }

Error:
    free(newName);
    free(newChannel);

    if (dlg > 0)
        DiscardPanel(dlg);
    if (error < 0)
        return error;
    else
    if (canceled)
        return CANCELED;
    else
    if (changed)
        return CHANGE_MADE;
    else
        return NO_CHANGE_MADE;
}

/****************************************************************************************/

static int CVIFUNC SaveSettings(DAQ_ChartInfo chartInfo, char *settingsName)
{
    int     error = 0;
    IniText iniText = 0;
    char    section[MAX_SETTING_NAME_LENGTH + 128];     /* control prefix should never be anywhere near 128 characters */
    int     panel = chartInfo->panel;
    int     ctrl = chartInfo->baseCtrl;
    int     hardwareTiming, warn;
    int     historySize;
    double  scanFrequency, refreshFrequency;

    errChk( ReadConfigIni(&iniText));

    sprintf(section, "%s%s", DAQ_CHART_SETTINGS_PREFIX, settingsName);
    Ini_RemoveSection (iniText, section);

    errChk( DaqCtrl_WriteSettings((DaqCtrl)chartInfo, iniText, section));

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE, &historySize));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_SCAN_FREQUENCY, &scanFrequency));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_REFRESH_FREQUENCY, &refreshFrequency));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG, &hardwareTiming));
    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DAQ_CHART_WARN_ON_HW_TIMING_RESTARTS, &warn));

    errChk( Ini_PutInt      (iniText, section, HISTORY_SIZE_TAG, historySize));
    errChk( Ini_PutDouble   (iniText, section, SCAN_FREQUENCY_TAG, scanFrequency));
    errChk( Ini_PutDouble   (iniText, section, REFRESH_FREQUENCY_TAG, refreshFrequency));
    errChk( Ini_PutBoolean  (iniText, section, HARDWARE_TIMING_TAG, hardwareTiming));
    errChk( Ini_PutBoolean  (iniText, section, WARN_ON_RESTARTS_TAG, warn));

    errChk( WriteOutAlarms(chartInfo, iniText, section));
    errChk( WriteOutTraceColors(chartInfo, iniText, section));
    errChk( WriteOutDataLogAttributes(chartInfo->dataLog, iniText, section));

    errChk( WriteConfigIni(iniText));
    strcpy(chartInfo->settingsName, settingsName);      /* if success, this is the new settings name */

Error:
    Ini_Dispose (iniText);
    return error;
}

/****************************************************************************************/

static int WriteOutTraceColors(DAQ_ChartInfo chartInfo, IniText iniText, char *section)
{
    int     error = 0;
    int     numTraces;
    int     color;
    int     index;
    char    itemName[128];

    errChk( GetCtrlAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_NUM_TRACES, &numTraces));

    errChk( Ini_PutInt (iniText, section, NUM_TRACES_TAG, numTraces));

    for (index = 1; index <= numTraces; index++)
        {
        errChk( GetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, index, ATTR_TRACE_COLOR, &color));
        sprintf(itemName, TRACE_COLOR_TAG, index);
        errChk( Ini_PutInt (iniText, section, itemName, color));
        }

Error:
    return error;
}

/****************************************************************************************/

static int WriteOutAlarms(DAQ_ChartInfo chartInfo, IniText iniText, char *section)
{
    int     error = 0;
    int     numAlarms;
    Alarm   *alarmPtr;
    int     index;
    char    itemName[128];
    char    *modeTag;

    numAlarms = ListNumItems(chartInfo->alarmList);

    errChk( Ini_PutInt (iniText, section, NUM_ALARMS_TAG, numAlarms));

    for (index = 1; index <= numAlarms; index++)
        {
        alarmPtr = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);

        sprintf(itemName, ALARM_NAME_TAG, index);
        errChk( Ini_PutString (iniText, section, itemName, alarmPtr->name));

        sprintf(itemName, ALARM_CHANNEL_TAG, index);
        errChk( Ini_PutString (iniText, section, itemName, alarmPtr->channel));

        sprintf(itemName, ALARM_LIMIT_TAG, index);
        errChk( Ini_PutDouble (iniText, section, itemName, alarmPtr->limit));

        sprintf(itemName, ALARM_LOWER_LIMIT_TAG, index);
        errChk( Ini_PutDouble (iniText, section, itemName, alarmPtr->lowerLimit));

        sprintf(itemName, ALARM_DEADBAND_TAG, index);
        errChk( Ini_PutDouble (iniText, section, itemName, alarmPtr->deadband));

        sprintf(itemName, ALARM_MODE_TAG, index);
        switch (alarmPtr->mode)
            {
            case VAL_ALARM_ABOVE_LIMIT:
                modeTag = ALARM_TRIGGER_ABOVE_TAG_VAL;
                break;
            case VAL_ALARM_BELOW_LIMIT:
                modeTag = ALARM_TRIGGER_BELOW_TAG_VAL;
                break;
            case VAL_ALARM_IN_RANGE:
                modeTag = ALARM_TRIGGER_IN_RANGE_TAG_VAL;
                break;
            case VAL_ALARM_OUT_OF_RANGE:
                modeTag = ALARM_TRIGGER_OUT_OF_RANGE_TAG_VAL;
                break;
            }
        errChk( Ini_PutString (iniText, section, itemName, modeTag));

        sprintf(itemName, ALARM_PLOT_TAG, index);
        errChk( Ini_PutBoolean (iniText, section, itemName, alarmPtr->plot));

        sprintf(itemName, ALARM_COLOR_TAG, index);
        errChk( Ini_PutUInt (iniText, section, itemName, alarmPtr->color));

        sprintf(itemName, ALARM_LOG_WHILE_ON_TAG, index);
        errChk( Ini_PutBoolean (iniText, section, itemName, alarmPtr->logAction & ALARM_LOG_WHILE_ON));

        sprintf(itemName, ALARM_LOG_WHILE_OFF_TAG, index);
        errChk( Ini_PutBoolean (iniText, section, itemName, alarmPtr->logAction & ALARM_LOG_WHILE_OFF));
        }

Error:
    return error;
}

/****************************************************************************************/

#define BETA_ALARM_NAME_TAG                 "Alarm%d_Name"
#define BETA_ALARM_CHANNEL_TAG              "Alarm%d_Channel"
#define BETA_ALARM_LIMIT_TAG                "Alarm%d_Limit"
#define BETA_ALARM_LOWER_LIMIT_TAG          "Alarm%d_Lower Limit"
#define BETA_ALARM_DEADBAND_TAG             "Alarm%d_DeadBand"
#define BETA_ALARM_MODE_TAG                 "Alarm%d_TriggerAboveLimit"
#define BETA_ALARM_COLOR_TAG                "Alarm%d_Color"
#define BETA_ALARM_PLOT_TAG                 "Alarm%d_PlotLimit"

    /*  returns 0 (success), negative error code or positive warning code */
static int BetaReadInAlarms(DAQ_ChartInfo chartInfo, IniText iniText, char *section, char *errorDescription, int *doingGets)
{
    int         error = 0;
    int         numAlarms;
    int         index;
    char        itemName[128];
    char        name[DAQ_CHART_MAX_ALARM_NAME_LENGTH + 1];
    char        channel[MAX_CHANNEL_STRING_LEN + 1];
    double      limit, lowerLimit;
    int         aboveLimit;
    double      deadband;
    int         mode;
    int         plotLimit;
    int         logAction;
    unsigned int color;
    int         found = TRUE;
    Alarm       alarm;
    ListType    newAlarmList = 0;

    *doingGets = FALSE;

    /* set defaults */
    strcpy(name, gDefaultAlarm.name);
    strcpy(channel, gDefaultAlarm.channel);
    limit =  gDefaultAlarm.limit;
    lowerLimit = gDefaultAlarm.lowerLimit;
    deadband = gDefaultAlarm.deadband;
    mode = gDefaultAlarm.mode;
    plotLimit = gDefaultAlarm.plot;
    color = gDefaultAlarm.color;
    logAction = gDefaultAlarm.logAction;

    nullChk( newAlarmList = ListCreate(sizeof(alarm)));

    errChk( found = Ini_GetInt (iniText, section, NUM_ALARMS_TAG, &numAlarms));
    if (!found)
        {
        if (errorDescription)
            sprintf(errorDescription, "Could not find section '%s'", section);
        errChk( ToolErr_MissingSection);
        }

    nullChk( ListPreAllocate(newAlarmList, numAlarms));

    *doingGets = TRUE;
    for (index = 1; index <= numAlarms; index++)
        {
        memset (&alarm, 0, sizeof(alarm));

        sprintf(itemName, BETA_ALARM_NAME_TAG, index);
        errChk( found = Ini_GetStringIntoBuffer (iniText, section, itemName, name, DAQ_CHART_MAX_ALARM_NAME_LENGTH + 1));

        sprintf(itemName, BETA_ALARM_CHANNEL_TAG, index);
        errChk( found = Ini_GetStringIntoBuffer (iniText, section, itemName, channel, MAX_CHANNEL_STRING_LEN + 1));

        sprintf(itemName, BETA_ALARM_LIMIT_TAG, index);
        errChk( found = Ini_GetDouble (iniText, section, itemName, &limit));

        sprintf(itemName, BETA_ALARM_LOWER_LIMIT_TAG, index);
        errChk( found = Ini_GetDouble (iniText, section, itemName, &lowerLimit));

        sprintf(itemName, BETA_ALARM_DEADBAND_TAG, index);
        errChk( found = Ini_GetDouble (iniText, section, itemName, &deadband));

        sprintf(itemName, BETA_ALARM_MODE_TAG, index);
        errChk( found = Ini_GetBoolean (iniText, section, itemName, &aboveLimit));

        if (found)
            {
            if (aboveLimit)
                mode = VAL_ALARM_ABOVE_LIMIT;
            else
                mode = VAL_ALARM_BELOW_LIMIT;
            }

        sprintf(itemName, ALARM_PLOT_TAG, index);
        errChk( found = Ini_GetBoolean (iniText, section, itemName, &plotLimit));

        sprintf(itemName, ALARM_COLOR_TAG, index);
        errChk( found = Ini_GetUInt (iniText, section, itemName, &color));

        errChk( AllocateAlarmResources(&alarm, name, channel, limit, lowerLimit, deadband, mode, plotLimit, color, logAction));
        ListInsertItem(newAlarmList, &alarm, END_OF_LIST);  /* can't fail, memory pre-allocated */              
        }
    *doingGets = FALSE;

    errChk(ReplaceAllChartAlarms(chartInfo, newAlarmList));

Error:
    FreeAlarmList(newAlarmList);

    if (error < 0)
        {
        if (doingGets && errorDescription && !*errorDescription)
                sprintf(errorDescription, "Error reading from line %d", Ini_LineOfLastAccess (iniText));    
        }

    return error;
}

/****************************************************************************************/

    /*  returns 0 (success), negative error code or positive warning code */
static int ReadInAlarms(DAQ_ChartInfo chartInfo, IniText iniText, char *section, char *errorDescription, int *doingGets)
{
    int         error = 0;
    int         numAlarms;
    int         index;
    char        itemName[128];
    char        name[DAQ_CHART_MAX_ALARM_NAME_LENGTH + 1];
    char        channel[MAX_CHANNEL_STRING_LEN + 1];
    double      limit, lowerLimit;
    double      deadband;
    int         mode;
    int         plotLimit;
    unsigned int color;
    int         logAction;
    int         logWhileOn;
    int         logWhileOff;
    int         found = TRUE;
    Alarm       alarm;
    ListType    newAlarmList = 0;
    char        *modeTag = 0;

    *doingGets = FALSE;

    /* set defaults */
    strcpy(name, gDefaultAlarm.name);
    strcpy(channel, gDefaultAlarm.channel);
    limit =  gDefaultAlarm.limit;
    lowerLimit = gDefaultAlarm.lowerLimit;
    deadband = gDefaultAlarm.deadband;
    mode = gDefaultAlarm.mode;
    plotLimit = gDefaultAlarm.plot;
    logAction = gDefaultAlarm.logAction;

    nullChk( newAlarmList = ListCreate(sizeof(alarm)));

    errChk( found = Ini_GetInt (iniText, section, NUM_ALARMS_TAG, &numAlarms));
    if (!found)
        {
        if (errorDescription)
            sprintf(errorDescription, "Could not find section '%s'", section);
        errChk( ToolErr_MissingSection);
        }

    nullChk( ListPreAllocate(newAlarmList, numAlarms));

    *doingGets = TRUE;
    for (index = 1; index <= numAlarms; index++)
        {
        memset (&alarm, 0, sizeof(alarm));

        sprintf(itemName, ALARM_NAME_TAG, index);
        errChk( found = Ini_GetStringIntoBuffer (iniText, section, itemName, name, DAQ_CHART_MAX_ALARM_NAME_LENGTH + 1));

        sprintf(itemName, ALARM_CHANNEL_TAG, index);
        errChk( found = Ini_GetStringIntoBuffer (iniText, section, itemName, channel, MAX_CHANNEL_STRING_LEN + 1));

        sprintf(itemName, ALARM_LIMIT_TAG, index);
        errChk( found = Ini_GetDouble (iniText, section, itemName, &limit));

        sprintf(itemName, ALARM_LOWER_LIMIT_TAG, index);
        errChk( found = Ini_GetDouble (iniText, section, itemName, &lowerLimit));

        sprintf(itemName, ALARM_DEADBAND_TAG, index);
        errChk( found = Ini_GetDouble (iniText, section, itemName, &deadband));

        sprintf(itemName, ALARM_MODE_TAG, index);
        errChk( found = Ini_GetStringCopy (iniText, section, itemName, &modeTag));

        if (found)
            {
            if (StrICmp(modeTag, ALARM_TRIGGER_ABOVE_TAG_VAL) == 0)
                mode = VAL_ALARM_ABOVE_LIMIT;
            else
            if (StrICmp(modeTag, ALARM_TRIGGER_BELOW_TAG_VAL) == 0)
                mode = VAL_ALARM_BELOW_LIMIT;
            else
            if (StrICmp(modeTag, ALARM_TRIGGER_IN_RANGE_TAG_VAL) == 0)
                mode = VAL_ALARM_IN_RANGE;
            else
            if (StrICmp(modeTag, ALARM_TRIGGER_OUT_OF_RANGE_TAG_VAL) == 0)
                mode = VAL_ALARM_OUT_OF_RANGE;
            }

        sprintf(itemName, ALARM_PLOT_TAG, index);
        errChk( found = Ini_GetBoolean (iniText, section, itemName, &plotLimit));

        sprintf(itemName, ALARM_LOG_WHILE_ON_TAG, index);
        errChk( found = Ini_GetBoolean (iniText, section, itemName, &logWhileOn));
        if (found)
            {
            if (logWhileOn)
                logAction |= ALARM_LOG_WHILE_ON;
            else
                logAction &= ~ALARM_LOG_WHILE_ON;
            }

        sprintf(itemName, ALARM_LOG_WHILE_OFF_TAG, index);
        errChk( found = Ini_GetBoolean (iniText, section, itemName, &logWhileOff));
        if (found)
            {
            if (logWhileOff)
                logAction |= ALARM_LOG_WHILE_OFF;
            else
                logAction &= ~ALARM_LOG_WHILE_OFF;
            }

        sprintf(itemName, ALARM_COLOR_TAG, index);
        errChk( found = Ini_GetUInt (iniText, section, itemName, &color));

        errChk( AllocateAlarmResources(&alarm, name, channel, limit, lowerLimit, deadband, mode, plotLimit, color, logAction));
        ListInsertItem(newAlarmList, &alarm, END_OF_LIST);  /* can't fail, memory pre-allocated */              
        }
    *doingGets = FALSE;

    errChk(ReplaceAllChartAlarms(chartInfo, newAlarmList));

Error:
    FreeAlarmList(newAlarmList);
    free(modeTag);

    if (error < 0)
        {
        if (doingGets && errorDescription && !*errorDescription)
            {
#if 0   /* restore this code if we decide to add 'if (!found) errChk(ToolErr_MissingItem);' after each get */
            if (error == ToolErr_MissingItem && !found)
                sprintf(errorDescription, "Could not find item '%s'", itemName);
            else
#endif
                sprintf(errorDescription, "Error reading from line %d", Ini_LineOfLastAccess (iniText));    
            }
        }

    return error;
}

/****************************************************************************************/

    /* missing trace colors are ignored */
static int ReadInTraceColors(DAQ_ChartInfo chartInfo, IniText iniText, char *section, char *errorDescription, int *doingGets)
{
    int         error = 0;
    int         numTracesInFile;
    int         numTracesOnChart;
    int         numTracesToRecall;
    int         index;
    int         color;
    int         found;
    char        itemName[128];

    *doingGets = FALSE;

    errChk( GetCtrlAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_NUM_TRACES, &numTracesOnChart));
    errChk( found = Ini_GetInt (iniText, section, NUM_TRACES_TAG, &numTracesInFile));

    if (found)
        {
        numTracesToRecall = Min(numTracesOnChart, numTracesInFile);

        *doingGets = TRUE;
        for (index = 1; index <= numTracesToRecall; index++)
            {
            sprintf(itemName, TRACE_COLOR_TAG, index);
            *doingGets = TRUE;
            errChk( found = Ini_GetInt (iniText, section, itemName, &color));
            *doingGets = FALSE;

            if (found)
                errChk( SetTraceAttribute(chartInfo->panel, chartInfo->baseCtrl, index, ATTR_TRACE_COLOR, color));
            }
        }

Error:
    if (error < 0)
        {
        if (doingGets && errorDescription && !*errorDescription)
            sprintf(errorDescription, "Error reading from line %d", Ini_LineOfLastAccess (iniText));    

        return -error;  /* error reading a trace color is only a warning and must be a positive number */
        }
    else
        return 0;

}

/****************************************************************************************/

    /*  errorDescription should be a pointer to a buffer initialize to an empty string, but large
        enough for a short error message description.
        0 is returned for success, a negative error code is returned for fatal error, or a positive error code
        is returned for warnings.
    */
static int CVIFUNC RecallSettings(DAQ_ChartInfo chartInfo, char *settingsName, char *errorDescription)
{
    int     error = 0;
    IniText iniText = 0;
    char    section[MAX_SETTING_NAME_LENGTH + 256];
    int     panel = chartInfo->panel;
    int     ctrl = chartInfo->baseCtrl;
    int     historySize;
    double  scanFrequency, refreshFrequency;
    int     hardwareTiming, warn;
    int     doingGets = FALSE;
    int     foundHistorySize;
    int     foundScanFrequency, foundRefreshFrequency;
    int     foundHardwareTiming, foundWarn;
    int     startOnLoad;
    int     result, warning = 0;

    chartInfo->recallingSettings = TRUE;

    errChk(ReadConfigIni(&iniText));
    sprintf(section, "%s%s", chartInfo->settingsPrefix, settingsName);

    if (!Ini_SectionExists (iniText, section))
        {
        if (errorDescription)
            sprintf(errorDescription, "Couldn't find section '%s'", section);
        errChk( ToolErr_MissingSection);
        }

        /* setting the channel string (done by the base class function) will terminate logging anyway (and will display a warning).  Turn off logging now to avoid the warning */
    errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl , ATTR_DATALOG_LOGGING_ON, FALSE));

        /*  clear the alarms before the channel string is set because otherwise if the number of channels from the new configuration
            plus the number of alarms from the current configuration exceed 64, an error occurs due to the strip chart limit of 64 traces
        */
    DisposeAllChartAlarms(chartInfo);   /* clears the charts alarm list, freeing resources owned by the alarms */

    errChk( result = DaqCtrl_ReadSettings((DaqCtrl)chartInfo, iniText, section, errorDescription, &doingGets));
    if (!warning) warning = result;

    doingGets = TRUE;
    errChk( foundHistorySize        = Ini_GetInt (iniText, section, HISTORY_SIZE_TAG, &historySize));
    errChk( foundScanFrequency      = Ini_GetDouble (iniText, section, SCAN_FREQUENCY_TAG, &scanFrequency));
    errChk( foundRefreshFrequency   = Ini_GetDouble (iniText, section, REFRESH_FREQUENCY_TAG, &refreshFrequency));
    errChk( foundHardwareTiming     = Ini_GetBoolean    (iniText, section, HARDWARE_TIMING_TAG, &hardwareTiming));
    errChk( foundWarn               = Ini_GetBoolean    (iniText, section, WARN_ON_RESTARTS_TAG, &warn));
    doingGets = FALSE;

    if (foundHistorySize)
        result = Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_HISTORY_BUFFER_SIZE, historySize);
    if (!warning && result < 0) warning = -result;

    if (foundRefreshFrequency)
        result = Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_REFRESH_FREQUENCY, (double)refreshFrequency);
    if (!warning && result < 0) warning = -result;

    if (foundHardwareTiming)
        result = Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_USE_HARDWARE_TIMIMG, hardwareTiming);
    if (!warning && result < 0) warning = -result;

    if (foundWarn)
        result = Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_WARN_ON_HW_TIMING_RESTARTS, warn);
    if (!warning && result < 0) warning = -result;

    if (foundScanFrequency)
        result = Internal_DAQ_Chart_SetAttribute(panel, ctrl, ATTR_DAQ_CHART_SCAN_FREQUENCY, (double)scanFrequency);
    if (!warning && result < 0) warning = -result;

    if (chartInfo->lastRecalledVersion == DAQ_CHART_BETA_VERSION)
        errChk( result = BetaReadInAlarms(chartInfo, iniText, section, errorDescription, &doingGets));
    else
        errChk( result = ReadInAlarms(chartInfo, iniText, section, errorDescription, &doingGets));

    if (!warning && result) warning = -result;

    errChk( result = ReadInTraceColors(chartInfo, iniText, section, errorDescription, &doingGets));
    if (!warning && result) warning = result;

    errChk( result = ReadInDataLogAttributes(chartInfo->dataLog, iniText, section, errorDescription, &doingGets));
    if (!warning && result) warning = result;

    strcpy(chartInfo->settingsName, settingsName);      /* if success, this is the new settings name */

    errChk( DAQ_Chart_GetAttribute(panel, ctrl, ATTR_DATALOG_START_LOGGING_ON_LOAD, &startOnLoad));
    if (startOnLoad)
        errChk( Internal_DAQ_Chart_SetAttribute(panel, ctrl , ATTR_DATALOG_LOGGING_ON, TRUE));

Error:
    chartInfo->recallingSettings = FALSE;

    if (error < 0)
        {
        if (doingGets && errorDescription && !*errorDescription)
            sprintf(errorDescription, "Error reading from line %d", Ini_LineOfLastAccess (iniText));    
        }
    else
        {
        error = warning;
        if (warning == 0)
            WarnIfAlarmsHaveInvalidChannels(chartInfo);     /*  do this now, since it was suppresed when recalling settings because we can't simultaneously recall the alarms and the channel string */
        }

    Ini_Dispose (iniText);
    return error;
}

/**********************************************************/

int CVIFUNC DAQ_Chart_RecallConfig(int panel, int daqChart, char *configurationName)
{
    return Internal_DAQ_Chart_RecallConfig(panel, daqChart, configurationName);
}

static int Internal_DAQ_Chart_RecallConfig(int panel, int daqChart, char *configurationName)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = 0;

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

    if (configurationName && *configurationName)
        errChk( RecallSettings(chartInfo, configurationName, 0));

    errChk( DaqCtrl_UpdateConfigDlgCtrl((DaqCtrl)chartInfo));    /* update the config name in the dialog */

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/**********************************************************/

int CVIFUNC DAQ_Chart_SaveConfig(int panel, int daqChart, char *configurationName)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = 0;

    errChk( GetDAQCtrlInfo(panel, daqChart, DAQ_CHART_TYPE, (DaqCtrl *)&chartInfo));

    if (!configurationName)
        errChk( UIENullPointerPassed);

    if (*configurationName)
        errChk( SaveSettings(chartInfo, configurationName));

    errChk( DaqCtrl_UpdateConfigDlgCtrl((DaqCtrl)chartInfo));    /* update the config name in the dialog */

Error:
    if (error < 0)
        return error;
    else
        return 0;
}

/************************************/
/************************************/

static int CVICALLBACK NameLogFileBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    DAQ_ChartInfo   chartInfo = (DAQ_ChartInfo)callbackData;
    int             error = 0;
    int             result;

    UNUSED(dialog);
    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    if (event == EVENT_COMMIT)
        {
        errChk( result = DataLog_SelectLogFile(chartInfo->dataLog, FALSE));
        if (result == CHANGE_MADE)
            errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/*******************************************************************************/

static int CVICALLBACK AdvancedLoggingBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = (DAQ_ChartInfo)callbackData;
    int             result;

    UNUSED(dialog);
    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    if (event == EVENT_COMMIT)
        {
        errChk( result = AdvancedLoggingOptionsDlg(chartInfo));
        if (result == CHANGE_MADE)
            errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/*******************************************************************************/

static int CVICALLBACK ConvertLogFileBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int             error = 0;
    DAQ_ChartInfo   chartInfo = (DAQ_ChartInfo)callbackData;
    int             result;

    UNUSED(dialog);
    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    if (event == EVENT_COMMIT)
        {
        errChk( result = DataLog_ConvertBinaryFileDlg(chartInfo->dataLog, 0));
        if (result == CHANGE_MADE)
            errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/********************************************8***********************************/

static int CVICALLBACK TextFormatBtnCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    DAQ_ChartInfo   chartInfo = (DAQ_ChartInfo)callbackData;
    int             error = 0;
    int             result;

    UNUSED(dialog);
    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    if (event == EVENT_COMMIT)
        {
        errChk( result = DataLog_TextFormatDlg(chartInfo->dataLog));
        if (result == CHANGE_MADE)
            errChk( DaqCtrl_SetChangedState((DaqCtrl)chartInfo, TRUE));
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return 0;
}

/*******************************************************************************/

static int CreateLoggingControls(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int dialog;
    int fileFormat;

    dialog = chartInfo->dialog;

    errChk( chartInfo->nameLogFileButton =          NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Name Log File...", 0, 0));
    errChk( chartInfo->dataLogFileString =          NewCtrl (dialog, CTRL_STRING, "Data Log File:", 0, 0));
    errChk( chartInfo->eventLogFileString =         NewCtrl (dialog, CTRL_STRING, "Event Log File:", 0, 0));
    errChk( chartInfo->logEventsCheckBox =          NewCtrl (dialog, CTRL_CHECK_BOX, "Log Overruns, Alarm Events, and Errors", 0, 0));
    errChk( chartInfo->fileFormatRing =             NewCtrl (dialog, CTRL_RING, "File Format:", 0, 0));
    errChk( chartInfo->textFormatButton =           NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__File Layout...", 0, 0));        
    errChk( chartInfo->logControlTextMsg =          NewCtrl (dialog, CTRL_TEXT_MSG, "Logging Control", 0, 0));      
    errChk( chartInfo->logControlDeco   =           NewCtrl (dialog, CTRL_RAISED_FRAME, "", 0, 0));         
    errChk( chartInfo->startLoggingToggle =         NewCtrl (dialog, CTRL_SQUARE_TEXT_BUTTON, "", 0, 0));
    errChk( chartInfo->pauseLoggingToggle =         NewCtrl (dialog, CTRL_SQUARE_TEXT_BUTTON, "", 0, 0));
    errChk( chartInfo->loggingStateString =         NewCtrl (dialog, CTRL_STRING, "Logging State:", 0, 0));
    errChk( chartInfo->startOnLoadCheckBox =        NewCtrl (dialog, CTRL_CHECK_BOX, "Start Logging When Configuration Is Loaded", 0, 0));
    errChk( chartInfo->advancedLoggingButton =      NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Advanced...", 0, 0));
    errChk( chartInfo->convertLogFileButton =       NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Convert Binary Log File...", 0, 0));

    errChk( SetCtrlAttribute(dialog, chartInfo->startLoggingToggle, ATTR_ON_TEXT, "Terminate Logging"));
    errChk( SetCtrlAttribute(dialog, chartInfo->startLoggingToggle, ATTR_OFF_TEXT, "Start Logging Now"));
    errChk( SetCtrlAttribute(dialog, chartInfo->pauseLoggingToggle, ATTR_ON_TEXT, "Continue"));
    errChk( SetCtrlAttribute(dialog, chartInfo->pauseLoggingToggle, ATTR_OFF_TEXT, "Pause Now"));

    errChk( MakeCtrlGrayRecessedIndicator(dialog, chartInfo->dataLogFileString));
    errChk( MakeCtrlGrayRecessedIndicator(dialog, chartInfo->eventLogFileString));
    errChk( MakeCtrlGrayRecessedIndicator(dialog, chartInfo->loggingStateString));

    errChk( InsertListItem(dialog, chartInfo->fileFormatRing, -1, "Tab Delimited Text", DATA_LOG_TAB_DELIMITED_TEXT));
    errChk( InsertListItem(dialog, chartInfo->fileFormatRing, -1, "Comma Delimited Text", DATA_LOG_COMMA_DELIMITED_TEXT));
    errChk( InsertListItem(dialog, chartInfo->fileFormatRing, -1, "Constant Field Width Text", DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT));
    errChk( InsertListItem(dialog, chartInfo->fileFormatRing, -1, "Binary", DATA_LOG_BINARY));

    errChk( DAQ_Chart_GetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOG_FILE_FORMAT, &fileFormat));
    errChk( SetCtrlVal(dialog, chartInfo->fileFormatRing, fileFormat));

Error:
    return error;
}

/***************************************************************************************/

static int CalcDataLogTabMinDialogWidth(DAQ_ChartInfo chartInfo)
{
    int error = 0;
    int startStopPlusPauseWidth;
    int startStopWidth;
    int pauseWidth;
    int loggingStateWidth;
    int startOnLoadWidth;
    int decoContentWidth;
    int decoInsideHMargin;
    int minContentWidth;
    int minDialogWidth;
    int sheetMargin;
    int dialog;

    dialog = chartInfo->dialog;
    sheetMargin = decoInsideHMargin = chartInfo->hMargin;

        /* put the control labels on the left */
    errChk( PutLabelsOnLeft(dialog, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0, chartInfo->loggingStateString, 0));

    errChk( GetCtrlAttribute(dialog, chartInfo->startLoggingToggle, ATTR_WIDTH, &startStopWidth));
    errChk( GetCtrlAttribute(dialog, chartInfo->pauseLoggingToggle, ATTR_WIDTH, &pauseWidth));

    startStopPlusPauseWidth = startStopWidth + pauseWidth + DLG_BTN_H_GAP;
    errChk( GetCtrlBoundingRect(dialog, chartInfo->loggingStateString, 0,0,0, &loggingStateWidth));
    errChk( GetCtrlBoundingRect(dialog, chartInfo->startOnLoadCheckBox, 0,0,0, &startOnLoadWidth));

    decoContentWidth = Max(startStopPlusPauseWidth, loggingStateWidth);
    decoContentWidth = Max(decoContentWidth, startOnLoadWidth);

    minContentWidth = decoContentWidth + 2*decoInsideHMargin;
    minDialogWidth = minContentWidth + 2*chartInfo->hMargin + 2*sheetMargin;

    chartInfo->minDialogWidth = Max(chartInfo->minDialogWidth, minDialogWidth);

Error:
    return error;
}

/***************************************************************************************/

static int FormatDataLogTab(DAQ_ChartInfo chartInfo)
{
    int     error = 0;
    int     dialog = chartInfo->dialog;
    int     top, sheetTop;
    int     btnTop;
    int     contentWidth;
    int     sheetMargin;
    int     hMargin;
    int     decoInsideHMargin;
    int     decoContentWidth;
    int     decoContentLeft;
    int     decoContentRight;
    int     decoLeft;
    int     decoRight;
    int     decoWidth;
    Rect    r;

    errChk( EasyTab_GetBounds(dialog, chartInfo->sheetGroupCtrl, VAL_EASY_TAB_INTERIOR_BOUNDS, &r));
    sheetTop = r.top;
    top = sheetTop + chartInfo->vMargin - 3;    /* -3 is to try and squeeze down the datalog tab for 640X480 */
    hMargin = sheetMargin = chartInfo->hMargin;

    decoInsideHMargin = chartInfo->hMargin;
    contentWidth = chartInfo->minDialogWidth - 2*hMargin - 2*sheetMargin;
    decoLeft = hMargin + sheetMargin;
    decoWidth = contentWidth;
    decoRight = decoLeft + decoWidth;
    decoContentWidth = decoWidth - 2*decoInsideHMargin;
    decoContentLeft = decoLeft + decoInsideHMargin;
    decoContentRight = decoContentLeft + decoContentWidth;

    errChk( DistributeCtrls(dialog, VAL_TB_VERTICAL_DISTRIBUTION, VAL_TB_FIXED_GAP_SPACING, &top, DLG_CTRL_V_GAP-2, 0,
        chartInfo->nameLogFileButton, chartInfo->dataLogFileString, chartInfo->eventLogFileString,
        chartInfo->logEventsCheckBox, chartInfo->fileFormatRing, chartInfo->textFormatButton,
        chartInfo->logControlTextMsg, chartInfo->startLoggingToggle,
        chartInfo->loggingStateString, chartInfo->startOnLoadCheckBox, 0));

    top += DLG_CTRL_V_GAP;
    GetCtrlAttribute(dialog, chartInfo->startLoggingToggle, ATTR_TOP, &btnTop);
    SetCtrlAttribute(dialog, chartInfo->pauseLoggingToggle, ATTR_TOP, btnTop);

    SetCtrlAttribute(dialog, chartInfo->logControlDeco, ATTR_WIDTH, decoWidth);
    SetCtrlAttribute(dialog, chartInfo->logControlDeco, ATTR_LEFT, decoLeft);

    errChk( FormatGroupDecoration(dialog, chartInfo->logControlTextMsg, chartInfo->logControlDeco, decoLeft, top));

    SetCtrlAttribute(dialog, chartInfo->startLoggingToggle, ATTR_LEFT, decoContentLeft);
    SetCtrlRightPosition(dialog, chartInfo->pauseLoggingToggle, decoContentRight);

    SetCtrlRightPosition(dialog, chartInfo->loggingStateString, decoContentRight);
    SetCtrlAttribute(dialog, chartInfo->loggingStateString, ATTR_LABEL_LEFT, decoContentLeft);

    SetCtrlAttribute(dialog, chartInfo->startOnLoadCheckBox, ATTR_LEFT, decoContentLeft);

    top += DLG_CTRL_V_GAP;
    SetCtrlAttribute(dialog, chartInfo->advancedLoggingButton, ATTR_TOP, top);
    SetCtrlAttribute(dialog, chartInfo->convertLogFileButton, ATTR_TOP, top);
    SetCtrlRightPosition(dialog, chartInfo->convertLogFileButton, decoRight);

    ExpandCtrlToMeetLabel(dialog, chartInfo->dataLogFileString, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);
    ExpandCtrlToMeetLabel(dialog, chartInfo->eventLogFileString, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);
    ExpandCtrlToMeetLabel(dialog, chartInfo->loggingStateString, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);

    errChk( SizeRingCtrlToText (dialog, chartInfo->fileFormatRing));
    SetCtrlRightPosition(dialog, chartInfo->fileFormatRing, contentWidth + hMargin + sheetMargin);
    SetCtrlAttribute(dialog, chartInfo->fileFormatRing, ATTR_LABEL_LEFT, hMargin + sheetMargin);

Error:
    return error;
}

/*****************************************************************************************/

    /*  This function allocates a string containing default column headings intended
        for labeling the time and data from a scan.
    */

static int GetDefaultHeadings(DAQ_ChartInfo chartInfo, char **defaultHeadings)
{
    int     error = 0;
    int     numChannels;
    int     column;
    char    *defaults = 0;
    char    channelNameBuf[MAX_CHANNEL_NAME_LENGTH + 1];
    char    columnNameBuf[40 + MAX_CHANNEL_NAME_LENGTH + 1];
    int     savedBole = SetBOLE(FALSE);     /* in case GetChannelNameFromIndex() fails */

    errChk( DataLog_GetAttribute(chartInfo->dataLog, ATTR_DATALOG_NUM_CHANNELS, &numChannels));

    nullChk( AppendString(&defaults, "Time", -1));

    for (column = 1; column <= numChannels; column++)
        {
        nullChk( AppendString(&defaults, ", ", -1));

        if (GetChannelNameFromIndex (chartInfo->device, chartInfo->channelString, column - 1, chartInfo->ioDirection, channelNameBuf) < 0)
            sprintf(columnNameBuf, "Data Column %d", column);  /* print something reasonable if case of error */
        else
            sprintf(columnNameBuf, "Channel %s", channelNameBuf);

        nullChk( AppendString(&defaults, columnNameBuf, -1));
        }

Error:
    SetBOLE(savedBole);

    if (error < 0)
        {
        free(defaults);
        *defaultHeadings = 0;
        return error;
        }
    else
        {
        *defaultHeadings =defaults;
        return 0;
        }
}

/************************************************************************************/

    /*  This function replaces any unspecified column headings with a default column heading.
    */
static int ReviseColumnHeaders(DAQ_ChartInfo chartInfo, char **columnHeaders)
{
    int     error = 0;
    char    *defaultHeadings = 0;
    char    *revisedHeaders = 0;

    errChk( GetDefaultHeadings(chartInfo, &defaultHeadings));

    errChk( SpliceCommaSeparatedList(*columnHeaders, defaultHeadings, &revisedHeaders));
    free(*columnHeaders);
    *columnHeaders = revisedHeaders;
    revisedHeaders = 0;

Error:
    free(defaultHeadings);
    free(revisedHeaders);
    return error;
}

/*****************************************************************************************/

static void CVICALLBACK HandleDatalogNotification(void *callbackData, DataLogNotification event, void *eventData)
{
    DAQ_ChartInfo   chartInfo = (DAQ_ChartInfo)callbackData;
    int             error = 0;

    switch (event)
        {
        case DATALOG_CHANGED:
        case DATALOG_CHANGE_REJECTED:
            if (chartInfo->dialog && chartInfo->dialogDisplayed)
                {
                errChk( CopyDataLogAttrToDialog(chartInfo));
                errChk( UpdateDialogDimming(chartInfo));
                }
            break;
        case DATALOG_ERROR_ELABORATION:
            DaqCtrl_SetErrorElaboration((DaqCtrl)chartInfo, (char *)eventData);
            break;
        case DATALOG_REVISE_COLUMN_HEADERS:
            errChk( ReviseColumnHeaders(chartInfo, (char **)eventData));
        }

Error:
    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)chartInfo, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);
}

/*****************************************************************************************/

static int FillAlarmLogActionListBox(DAQ_ChartInfo chartInfo, int panel, int listBox)
{
    int     error = 0;
    Alarm   *alarmPtr;
    int     numAlarms;
    int     index;
    char    label[128];

    numAlarms = ListNumItems(chartInfo->alarmList);

    for (index = 1; index <= numAlarms; index++)
        {
        alarmPtr = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);
        sprintf(label, "%s - On", alarmPtr->name);
        errChk( InsertListItem (panel, listBox, -1, label, 0));
        errChk( CheckListItem(panel, listBox, (index - 1) * 2, alarmPtr->logAction & ALARM_LOG_WHILE_ON));
        sprintf(label, "%s - Off", alarmPtr->name);
        errChk( InsertListItem (panel, listBox, -1, label, 0));
        errChk( CheckListItem(panel, listBox, (index - 1) * 2 + 1, alarmPtr->logAction & ALARM_LOG_WHILE_OFF));
        }

Error:
    return error;
}

/*****************************************************************************************/

    /*  The list box is assumed to contain the log while on, log while off selections in its
        check marks.
    */
static int ApplyAlarmLogActionSelections(DAQ_ChartInfo chartInfo, int panel, int listBox)
{
    int     error = 0;
    Alarm   *alarmPtr;
    int     numAlarms;
    int     index;
    int     numItems;
    int     checked;

    numAlarms = ListNumItems(chartInfo->alarmList);
    errChk( GetNumListItems(panel, listBox, &numItems));

    numAlarms = Min(numItems * 2, numAlarms);   /* in case the number of alarms was changed in a timer callback (in this case they get what they get, but we make sure not to crash) */

    for (index = 1; index <= numAlarms; index++)
        {
        alarmPtr = (Alarm *)ListGetPtrToItem(chartInfo->alarmList, index);

        errChk( IsListItemChecked(panel, listBox, (index - 1) * 2, &checked));
        if (checked)
            alarmPtr->logAction |= ALARM_LOG_WHILE_ON;
        else
            alarmPtr->logAction &= ~ALARM_LOG_WHILE_ON;

        errChk( IsListItemChecked(panel, listBox, (index - 1) * 2 + 1, &checked));
        if (checked)
            alarmPtr->logAction |= ALARM_LOG_WHILE_OFF;
        else
            alarmPtr->logAction &= ~ALARM_LOG_WHILE_OFF;
        }

Error:
    return error;
}

/*****************************************************************************************/

static int AdvancedLoggingOptionsDlg(DAQ_ChartInfo  chartInfo)
{
    int     error = 0;
    int     dlg = 0;
    int     whenToStopMsg;
    int     whenToStopDeco;
    int     logModeRing;
    int     numScansNumeric;
    int     logTimeNumeric;
    int     conditionalLogMsg;
    int     conditionalLogDeco;
    int     controlLoggingCheckBox;
    int     promptOnOverwriteCheckBox;
    int     alarmListBox;
    int     ok, cancel;
    int     changed = FALSE;
    int     hMargin = DLG_MARGIN;
    int     vMargin = DLG_MARGIN;
    int     top;
    int     done = FALSE;
    int     evtPanel, evtCtrl;
    int     ctrlHeight;
    int     btnWidth;
    int     canceled = FALSE;
    int     frameContentWidth;
    int     frameContentRight;
    int     frameContentLeft;
    int     contentWidth;
    int     minContentWidth;
    int     minFrameContentWidth;
    int     dialogWidth;
    int     logMode;
    int     numScans;
    int     logConditionally;
    int     promptOnOverwrite;
    int     listIndex;

    errChk(dlg = NewPanel(0, "Advanced Logging Options", 0,0, 10,10));
    SetPanelAttribute(dlg, ATTR_CONFORM_TO_SYSTEM, TRUE);

    errChk( whenToStopMsg =         NewCtrl (dlg, CTRL_TEXT_MSG, "When To Stop Logging", 0, 0));
    errChk( whenToStopDeco =        NewCtrl (dlg, CTRL_RAISED_FRAME, "", 0, 0));
    errChk( logModeRing =           NewCtrl (dlg, CTRL_RING, "Log Mode:", 0, 0));
    errChk( numScansNumeric =       NewCtrl (dlg, CTRL_NUMERIC, "Number Of Scans To Log/Keep:", 0, 0));
    errChk( logTimeNumeric =        NewCtrl (dlg, CTRL_NUMERIC, "Log Time In Seconds:", 0, 0));

    errChk( conditionalLogMsg =     NewCtrl (dlg, CTRL_TEXT_MSG, "Conditional Logging", 0, 0));
    errChk( conditionalLogDeco =    NewCtrl (dlg, CTRL_RAISED_FRAME, "", 0, 0));
    errChk( controlLoggingCheckBox =NewCtrl (dlg, CTRL_CHECK_BOX, "Control Logging With Alarms", 0, 0));
    errChk( alarmListBox        =   NewCtrl (dlg, CTRL_LIST, "Log while any of the checked alarm\nconditions are true:", 0, 0));
    errChk( promptOnOverwriteCheckBox   =   NewCtrl (dlg, CTRL_CHECK_BOX, "Prompt Before Overwriting Log Files", 0, hMargin));

    errChk( ok =                    NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "__OK", 0, hMargin));
    errChk( cancel =                NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "  __Cancel  ", 0, hMargin));

    errChk( InsertListItem(dlg, logModeRing, -1, "Log Indefinitely", DATA_LOG_DONT_STOP));
    errChk( InsertListItem(dlg, logModeRing, -1, "Stop When All Scans Are Acquired", DATA_LOG_STOP_AT_SCAN_LIMIT));
    errChk( InsertListItem(dlg, logModeRing, -1, "Log Indefinitely, Overwriting Oldest", DATA_LOG_OVERWRITE_OLDEST));
    errChk( SizeRingCtrlToText(dlg, logModeRing));

    errChk( SetCtrlAttribute (dlg, numScansNumeric, ATTR_DATA_TYPE, VAL_INTEGER));
    errChk( SetCtrlAttribute(dlg, numScansNumeric, ATTR_MIN_VALUE, 1));
    errChk( SetCtrlAttribute (dlg, numScansNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( MakeCtrlGrayRecessedIndicator(dlg, logTimeNumeric));

    errChk( SetCtrlAttribute (dlg, alarmListBox, ATTR_CHECK_MODE, TRUE));

    top = vMargin;

    errChk( PutLabelsOnLeft (dlg, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0, logModeRing, numScansNumeric, logTimeNumeric, 0));
    errChk( AttributeMaxAndMin (dlg, ATTR_TB_BOUNDS_WIDTH, 0, &minFrameContentWidth, 0, 0, 0,
        logModeRing, numScansNumeric, logTimeNumeric, alarmListBox, 0));

    errChk( AttributeMaxAndMin (dlg, ATTR_TB_BOUNDS_WIDTH, 0, &minContentWidth, 0, 0, 0,
        promptOnOverwriteCheckBox, 0));

    frameContentWidth = Max(minFrameContentWidth, minContentWidth + 2 * hMargin);
    frameContentLeft = hMargin * 2;
    frameContentRight = frameContentLeft + frameContentWidth;
    contentWidth = frameContentWidth + 2 * hMargin;
    dialogWidth = contentWidth + 2 * hMargin;

    SetCtrlRightPosition(dlg, logModeRing, frameContentRight);
    SetCtrlRightPosition(dlg, numScansNumeric, frameContentRight);
    SetCtrlRightPosition(dlg, logTimeNumeric, frameContentRight);

    errChk( SetAttributeForCtrls (dlg, ATTR_LEFT, frameContentLeft, 0, controlLoggingCheckBox, alarmListBox, 0));
    errChk( SetAttributeForCtrls(dlg, ATTR_LABEL_LEFT, frameContentLeft, 0, logModeRing, numScansNumeric, logTimeNumeric, alarmListBox, 0));

    errChk( SetAttributeForCtrls (dlg, ATTR_WIDTH, frameContentWidth, 0, alarmListBox, 0));
    errChk( SetAttributeForCtrls (dlg, ATTR_WIDTH, contentWidth, 0, whenToStopDeco, conditionalLogDeco, 0));

    errChk( DistributeCtrls (dlg, VAL_TB_VERTICAL_DISTRIBUTION, VAL_TB_FIXED_GAP_SPACING, &top, DLG_CTRL_V_GAP, 0,
        whenToStopMsg, logModeRing, numScansNumeric, logTimeNumeric, 0));
    top += DLG_CTRL_V_GAP;

    errChk( FormatGroupDecoration(dlg, whenToStopMsg, whenToStopDeco, hMargin, top));
    top += DLG_CTRL_V_GAP;

    errChk( DistributeCtrls (dlg, VAL_TB_VERTICAL_DISTRIBUTION, VAL_TB_FIXED_GAP_SPACING, &top, DLG_CTRL_V_GAP, 0,
        conditionalLogMsg, controlLoggingCheckBox, alarmListBox, 0));
    top += DLG_CTRL_V_GAP;

    errChk( FormatGroupDecoration(dlg, conditionalLogMsg, conditionalLogDeco, hMargin, top));
    top += DLG_CTRL_V_GAP;

    SetCtrlAttribute(dlg, promptOnOverwriteCheckBox, ATTR_TOP, top);
    top += DLG_CTRL_V_GAP;

        /* make ok and cancel the same width */
    errChk( SetCtrlsToAttributeExtreme(dlg, ATTR_WIDTH, VAL_TB_MAX, 0, ok, cancel, 0));

        /* position ok and cancel vertically */
    top += DLG_CTRL_V_GAP * 2;
    errChk( SetAttributeForCtrls (dlg, ATTR_TOP, top, 0, ok, cancel, 0));

        /* position the ok and cancel horizontally */
    GetCtrlAttribute(dlg, ok, ATTR_WIDTH, &btnWidth);
    SetCtrlAttribute(dlg, cancel, ATTR_LEFT, hMargin + contentWidth - btnWidth);
    SetCtrlAttribute(dlg, ok, ATTR_LEFT, hMargin + contentWidth - 2 * btnWidth - DLG_BTN_H_GAP);

    GetCtrlAttribute(dlg, ok, ATTR_HEIGHT, &ctrlHeight);
    top += ctrlHeight + DLG_CTRL_V_GAP;

    SetPanelAttribute(dlg, ATTR_WIDTH, dialogWidth);
    SetPanelAttribute(dlg, ATTR_HEIGHT, top);

    errChk( SetCommonDialogShortcutKeys(dlg, ok, cancel, cancel));
    errChk( SetPanelPos(dlg, VAL_AUTO_CENTER, VAL_AUTO_CENTER));

    errChk( DAQ_Chart_GetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOG_MODE, &logMode));
    errChk( DAQ_Chart_GetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_NUM_SCANS_TO_LOG, &numScans));
    errChk( DAQ_Chart_GetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOG_CONDITIONALLY, &logConditionally));
    errChk( DAQ_Chart_GetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE, &promptOnOverwrite));
    errChk( SetCtrlVal(dlg, logModeRing, logMode));
    errChk( SetCtrlVal(dlg, numScansNumeric, numScans));
    errChk( SetCtrlVal(dlg, controlLoggingCheckBox, logConditionally));
    errChk( SetCtrlVal(dlg, promptOnOverwriteCheckBox, promptOnOverwrite));

    SetCtrlAttribute(dlg, controlLoggingCheckBox, ATTR_DIMMED, ListNumItems(chartInfo->alarmList) == 0);
    SetCtrlAttribute(dlg, logTimeNumeric, ATTR_DIMMED, logMode != DATA_LOG_STOP_AT_SCAN_LIMIT);
    SetCtrlAttribute(dlg, alarmListBox, ATTR_DIMMED, !logConditionally);
    errChk( FillAlarmLogActionListBox(chartInfo, dlg, alarmListBox));
    SetCtrlAttribute(dlg, numScansNumeric, ATTR_DIMMED, logMode == DATA_LOG_DONT_STOP);

    InstallPopup(dlg);

    while (!done)
        {
        GetUserEvent(0, &evtPanel, &evtCtrl);

        if (!ObjectExists(chartInfo))
            {
            done = TRUE;
            cancel = TRUE;
            }
        else
        if (evtPanel == dlg)
            {
            if  (evtCtrl == cancel)
                {
                done = TRUE;
                canceled = TRUE;
                }
            else
            if (evtCtrl == ok)
                {
                int numChecked;

                errChk( GetNumCheckedItems (dlg, alarmListBox, &numChecked));
                GetCtrlVal(dlg, controlLoggingCheckBox, &logConditionally);

                if (logConditionally && numChecked == 0)
                    {
                    int choice;

                    errChk( choice = GenericMessagePopup ("Conditional Logging Conflict", 
                        "The conditional logging check box is selected, but no alarms\n"
                        "are checked.  This will cause the conditional logging option\n"
                        "to be ignored.\n\n"
                        "Do you want to deactivate conditional logging?\n",
                         "__Deactivate", "__Ignore", "__Cancel", 0, 0, 0, VAL_GENERIC_POPUP_BTN1,
                                         VAL_GENERIC_POPUP_BTN1, VAL_GENERIC_POPUP_BTN3));
                    switch (choice)
                        {
                        case VAL_GENERIC_POPUP_BTN1:
                            SetCtrlVal(dlg,controlLoggingCheckBox, FALSE);
                            SetCtrlAttribute(dlg, alarmListBox, ATTR_DIMMED, TRUE);
                            changed = TRUE;
                            break;
                        case VAL_GENERIC_POPUP_BTN2:
                            done = TRUE;
                            break;
                        case VAL_GENERIC_POPUP_BTN3:
                            break;
                        }
                    }
                else
                    done = TRUE;
                }
            else
            if (evtCtrl == alarmListBox)
                {
                errChk( GetCtrlIndex(dlg, alarmListBox, &listIndex));
                if (listIndex >= 0)
                    {
                    int checked;

                        /*  make sure the On and the Off state aren't both checked at the same time */
                    errChk( IsListItemChecked (dlg, alarmListBox, listIndex, &checked));
                    if (checked)
                        {
                        int coupledItemIndex;

                        if (listIndex & 0x1)    /* if odd */
                            coupledItemIndex = listIndex - 1;
                        else
                            coupledItemIndex = listIndex + 1;

                        errChk( CheckListItem(dlg, alarmListBox, coupledItemIndex, FALSE));
                        }
                    changed = TRUE;
                    }
                }
            else
            if (evtCtrl == controlLoggingCheckBox)
                {
                GetCtrlVal(dlg, controlLoggingCheckBox, &logConditionally);
                SetCtrlAttribute(dlg, alarmListBox, ATTR_DIMMED, !logConditionally);
                changed = TRUE;
                }
            else
            if (evtCtrl == logModeRing)
                {
                GetCtrlVal(dlg, logModeRing, &logMode);
                SetCtrlAttribute(dlg, logTimeNumeric, ATTR_DIMMED, logMode != DATA_LOG_STOP_AT_SCAN_LIMIT);
                SetCtrlAttribute(dlg, numScansNumeric, ATTR_DIMMED, logMode == DATA_LOG_DONT_STOP);
                changed = TRUE;
                }
            else
            if (evtCtrl == numScansNumeric || promptOnOverwriteCheckBox)
                {
                changed = TRUE;
                }
            }

        errChk( GetCtrlVal(dlg, numScansNumeric, &numScans));
        if (numScans > 0)
            errChk( SetCtrlVal(dlg, logTimeNumeric, numScans / chartInfo->scanFrequency));
        }

    if (!canceled && changed)
        {
        errChk( GetCtrlVal(dlg, logModeRing, &logMode));
        errChk( GetCtrlVal(dlg, numScansNumeric, &numScans));
        errChk( GetCtrlVal(dlg, controlLoggingCheckBox, &logConditionally));
        errChk( GetCtrlVal(dlg, promptOnOverwriteCheckBox, &promptOnOverwrite));

        errChk( Internal_DAQ_Chart_SetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOG_MODE, logMode));
        errChk( Internal_DAQ_Chart_SetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_NUM_SCANS_TO_LOG, numScans));
        errChk( Internal_DAQ_Chart_SetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_LOG_CONDITIONALLY, logConditionally));
        errChk( Internal_DAQ_Chart_SetAttribute(chartInfo->panel, chartInfo->baseCtrl, ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE, promptOnOverwrite));

        errChk( ApplyAlarmLogActionSelections(chartInfo, dlg, alarmListBox));
        }

Error:
    if (dlg > 0)
        DiscardPanel(dlg);
    if (error < 0)
        return error;
    else
    if (canceled)
        return CANCELED;
    else
    if (changed)
        return CHANGE_MADE;
    else
        return NO_CHANGE_MADE;
}

/****************************************************************/

    /*  returns TRUE if the file was reOrdered, FALSE if it was already in order,
        or a negative error code.
    */
int CVIFUNC DAQ_Chart_SortLogFile(char *logFilePath, int displayProgressDialog)
{
    int error = 0;
    int response;

    errChk( response = DataLog_ReorderLogFile(logFilePath, displayProgressDialog));

Error:

    if (error < 0)
        return error;
    else
        return response;
}


